#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::AudioTagReader
#
# The main idea for this class is to handle meta-tags in audio files,
# in particular reading them.
#
# Usage example:
#
#   MultimediaParadise::AudioTagReader.new
#
# =========================================================================== #
# require 'multimedia_paradise/audio/audio_tag_reader/audio_tag_reader.rb'
# MultimediaParadise::AudioTagReader.new(ARGV)
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

class AudioTagReader < ::MultimediaParadise::Base # === MultimediaParadise::AudioTagReader

  require 'multimedia_paradise/requires/require_file_duration.rb'

  # ========================================================================= #
  # === N_PAD
  # ========================================================================= #
  N_PAD = 25

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      commandline_arguments = nil,
      run_already           = true
    )
    reset
    set_commandline_arguments(
      commandline_arguments
    )
    # ======================================================================= #
    # === Handle blocks next
    # ======================================================================= #
    if block_given?
      yielded = yield
      case yielded
      # ===================================================================== #
      # === :do_not_report_anything
      # ===================================================================== #
      when :do_not_report_anything,
           :be_quiet,
           :be_silent
        @may_we_report = false
      end
    end
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @performer
    # ======================================================================= #
    @performer = nil
    # ======================================================================= #
    # === @band
    # ======================================================================= #
    @band = nil
    # ======================================================================= #
    # === @bpm
    # ======================================================================= #
    @bpm = nil
    # ======================================================================= #
    # === @date
    # ======================================================================= #
    @date = nil
    # ======================================================================= #
    # === @disc
    # ======================================================================= #
    @disc = nil
    # ======================================================================= #
    # === @may_we_report
    #
    # Whether to output anything to the user on the commandline or not,
    # by default.
    # ======================================================================= #
    @may_we_report = true
  end

  # ========================================================================= #
  # === set_this_file
  # ========================================================================= #
  def set_this_file(i)
    @this_file = i
  end

  # ========================================================================= #
  # === report_title
  # ========================================================================= #
  def report_title(i = @title)
    e sfancy('  Title: '.ljust(N_PAD))+
      right_colour(i.to_s)
  end

  # ========================================================================= #
  # === right_colour
  # ========================================================================= #
  def right_colour(i)
    mediumpurple(i)
  end

  # ========================================================================= #
  # === report_artist
  # ========================================================================= #
  def report_artist(i = @artist)
    e sfancy('  Artist: '.ljust(N_PAD))+
      right_colour(i.to_s)
  end

  # ========================================================================= #
  # === report_album
  # ========================================================================= #
  def report_album(i = @album)
    e sfancy('  Album: '.ljust(N_PAD))+
      right_colour(i.to_s)
  end

  # ========================================================================= #
  # === report_comment
  # ========================================================================= #
  def report_comment(i = @comment)
    e sfancy('  Comment: '.ljust(N_PAD))+
      right_colour(i.to_s)
  end

  # ========================================================================= #
  # === report_genre
  #
  # To get all available genres, see:
  #
  #   https://en.wikipedia.org/wiki/ID3#Genre_list_in_ID3v1[12]
  #
  # ========================================================================= #
  def report_genre(i = @genre)
    e sfancy('  Genre: '.ljust(N_PAD))+
      right_colour(i.to_s)
  end

  # ========================================================================= #
  # === report_year
  # ========================================================================= #
  def report_year(i = @year)
    e sfancy('  Year: '.ljust(N_PAD))+
      right_colour(i.to_s)
  end

  # ========================================================================= #
  # === report_length_in_seconds
  # ========================================================================= #
  def report_length_in_seconds(
      i = @length_in_seconds
    )
    e sfancy('  Duration (in seconds): '.ljust(N_PAD))+
      right_colour(i.to_s)
  end

  # ========================================================================= #
  # === tag?
  # ========================================================================= #
  def tag?
    @tag
  end; alias tag tag? # === tag

  # ========================================================================= #
  # === title?
  # ========================================================================= #
  def title?
    @title
  end; alias title title? # === title?

  # ========================================================================= #
  # === artist?
  # ========================================================================= #
  def artist?
    @artist
  end; alias artist artist? # === artist

  # ========================================================================= #
  # === date?
  # ========================================================================= #
  def date?
    @date
  end; alias date date? # === date

  # ========================================================================= #
  # === disc?
  # ========================================================================= #
  def disc?
    @disc
  end; alias disc disc? # === disc

  # ========================================================================= #
  # === bpm?
  # ========================================================================= #
  def bpm?
    @bpm
  end; alias bpm bpm? # === bpm

  # ========================================================================= #
  # === album?
  # ========================================================================= #
  def album?
    @album
  end; alias album album? # === album

  # ========================================================================= #
  # === year?
  # ========================================================================= #
  def year?
    @year
  end; alias year year? # === year

  # ========================================================================= #
  # === band?
  # ========================================================================= #
  def band?
    @band
  end; alias band band? # === band

  # ========================================================================= #
  # === genre?
  # ========================================================================= #
  def genre?
    @genre
  end; alias genre genre? # === genre

  # ========================================================================= #
  # === performer?
  # ========================================================================= #
  def performer?
    @performer
  end; alias performer performer? # === performer

  # ========================================================================= #
  # === comment?
  # ========================================================================= #
  def comment?
    @comment
  end; alias comment comment? # === comment

  # ========================================================================= #
  # === track?
  # ========================================================================= #
  def track?
    @track
  end; alias track track? # === track

  # ========================================================================= #
  # === set_title
  # ========================================================================= #
  def set_title(i)
    @title = i.to_s
  end; alias title= set_title # === title=

  # ========================================================================= #
  # === set_album
  # ========================================================================= #
  def set_album(i)
    @album = i.to_s
  end; alias album= set_album # === album=

  # ========================================================================= #
  # === set_artist
  # ========================================================================= #
  def set_artist(i)
    @artist = i.to_s
  end; alias artist= set_artist # === artist=

  # ========================================================================= #
  # === set_comment
  # ========================================================================= #
  def set_comment(i)
    @comment = i.to_s
  end; alias comment= set_comment # === comment=

  # ========================================================================= #
  # === set_track
  # ========================================================================= #
  def set_track(i)
    @track = i.to_i # This must be an Integer.
  end; alias track= set_track # === track=

  # ========================================================================= #
  # === set_year
  # ========================================================================= #
  def set_year(i)
    @year = i.to_i # This must be an Integer.
  end; alias year= set_year # === year=

  # ========================================================================= #
  # === set_genre
  # ========================================================================= #
  def set_genre(i)
    @genre = i.to_s
  end; alias genre= set_genre # === genre=

  # ========================================================================= #
  # === set_new_year
  # ========================================================================= #
  def set_new_year(i)
    @tag.year = i.to_i # This must be an Integer.
  end

  # ========================================================================= #
  # === set_new_album
  # ========================================================================= #
  def set_new_album(i)
    @tag.album = i
  end

  # ========================================================================= #
  # === report_everything
  #
  # This method will report as much as possible about the song at hand.
  #
  # "track" would also be a tag that could be shown. Personally I don't
  # really need it, though.
  # ========================================================================= #
  def report_everything
    report_title
    report_artist
    report_album
    report_year
    report_genre
    report_comment
    report_length_in_seconds
  end; alias report report_everything # === report

  # ========================================================================= #
  # === work_on_this_file
  # ========================================================================= #
  def work_on_this_file(
      this_file = @this_file
    )
    unless Object.const_defined? :TagLib
      text = <<-EOF
Please install TagLib first via:

  gem install taglib

EOF
      e text
      return text
    end
    # ======================================================================= #
    # Load up the file via TagLib first.
    # ======================================================================= #
    TagLib::FileRef.open(this_file) { |fileref| # of class TagLib::FileRef.
      unless fileref.null?
        @tag       = fileref.tag  # Keep a reference to it here.
        @title     = @tag.title   # => "Wake Up"
        @artist    = @tag.artist  # => "DJ Punk"
        @album     = @tag.album   # => "Funeral"
        @year      = @tag.year    # => 2004
        @genre     = @tag.genre   # => "Indie Rock"
        @comment   = @tag.comment
        @track     = @tag.track
        # @band    = @tag.band # Does not exist.
        @performer = @tag.performer if @tag.respond_to? :performer
        @properties = fileref.audio_properties
        do_determine_the_length_of_the_song(this_file)
        if @may_we_report
          e
          report_everything
          e
        end
      end
    } # File is automatically closed at block end.
  end

  # ========================================================================= #
  # === this_file?
  # ========================================================================= #
  def this_file?
    @this_file
  end; alias main_file? this_file? # === main_file?

  # ========================================================================= #
  # === length_in_seconds?
  # ========================================================================= #
  def length_in_seconds?
    @length_in_seconds
  end

  # ========================================================================= #
  # === do_determine_the_length_of_the_song
  #
  # The first argument should be the name of the .mp3 file at hand.
  # ========================================================================= #
  def do_determine_the_length_of_the_song(
      this_file = main_file?
    )
    # ======================================================================= #
    # The way for the mp3-library goes like this:
    #   @length_in_seconds = @properties.length_in_seconds  # => 335 (song length)
    # But this was buggy for me, so we query ffmpeg instead. The code above
    # is retained, though, in the event that we may want to re-instate it
    # one day. (Noted this down at 26.10.2021)
    # ======================================================================= #
    @length_in_seconds = ::MultimediaParadise.length_of?(this_file)
  end

  # ========================================================================= #
  # === try_to_require_the_tablib_gem
  # ========================================================================= #
  def try_to_require_the_tablib_gem
    begin
      # ===================================================================== #
      # Require the taglib-ruby gem.
      #
      # See: https://rubygems.org/gems/taglib-ruby
      # ===================================================================== #
      require 'taglib'
    rescue LoadError; end
  end

  # ========================================================================= #
  # === update!
  #
  # This method can be used to update the tags.
  # ========================================================================= #
  def update!( 
      this_file = @this_file,
      title     = @title,
      artist    = @artist,
      genre     = @genre,
      comment   = @comment,
      track     = @track
    )
    TagLib::FileRef.open(this_file) { |fileref| # of class TagLib::FileRef.
      unless fileref.null?
        @tag       = fileref.tag # Keep a reference to it here.
        set_new_album(@album) # Use the old reference.
        set_new_year(@year)  # Use the old reference, via @year.
        @tag.title   = title.to_s
        @tag.artist  = artist.to_s
        @tag.genre   = genre.to_s
        @tag.comment = comment.to_s
        if track
          @tag.track = track.to_i # This must be an integer.
        end
        fileref.save
      end
    }
  end

  # ========================================================================= #
  # === save
  # ========================================================================= #
  def save
    @tag.save
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    try_to_require_the_tablib_gem
    # ======================================================================= #
    # We will assume that the first argument given is the file we want to
    # work on.
    # ======================================================================= #
    this_file = first_argument?.to_s.dup
    set_this_file(this_file)
    work_on_this_file(@this_file)
  end

  # ========================================================================= #
  # === MultimediaParadise::AudioTagReader[]
  # ========================================================================= #
  def self.[](i = '')
    new(i) { :be_quiet}
  end

end

# =========================================================================== #
# === MultimediaParadise.metadata_from_this_song
# =========================================================================== #
def self.metadata_from_this_song(this_song)
  MultimediaParadise::AudioTagReader[this_song]
end; self.instance_eval { alias analyse_tags_of_this_mp3_file metadata_from_this_song } # === MultimediaParadise.analyse_tags_of_this_mp3_file

end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::AudioTagReader.new(ARGV)
end # audiotagreader /home/x/songs/Youtube_KyokushinMix2012.mp3
    # audiotagreader *mp3