#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::CreateM3uPlaylist
#
# This class will generate a .m3u playlist (thus having the 'm3u' filename
# extension). The listing (aka content) for such a .m3u file should ideally
# be linear and simple, e. g. just have the filename like "foobar.mp3" per
# line, without the quotes.
#
# Usage example:
#
#   MultimediaParadise::CreateM3uPlaylist.new(ARGV)
#
# =========================================================================== #
# require 'multimedia_paradise/audio/create_m3u_playlist.rb'
# MultimediaParadise.create_m3u_playlist
# =========================================================================== #
require 'multimedia_paradise/audio/base.rb'
require 'multimedia_paradise/toplevel_methods/files_and_directories.rb'

module MultimediaParadise

class CreateM3uPlaylist < MultimediaParadise::AudioBase # === MultimediaParadise::CreateM3uPlaylist

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === initialize
  #
  # The first argument must include our dataset.
  # ========================================================================= #
  def initialize(
      dataset     = nil,
      store_here  = 'default_playlist.m3u',
      run_already = true
    )
    reset
    set_dataset(
      dataset
    )
    set_store_where(
      store_here
    )
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @debug
    #
    # If the next variable is set to true then we will show more information,
    # aka debug-related information.
    # ======================================================================= #
    @debug = false
  end

  # ========================================================================= #
  # === obtain_songs_from_the_default_directory
  #
  # This method will obtain all audio songs from the local directory
  # containing all songs.
  # ========================================================================= #
  def obtain_songs_from_the_default_directory(
      use_this_directory = directory_to_the_local_songs?
    )
    filter_for_audio_files(
      Dir["#{use_this_directory}*"]
    )
  end

  # ========================================================================= #
  # === set_store_where
  #
  # The name of the file.
  # ========================================================================= #
  def set_store_where(i)
    i << '.m3u' unless i.end_with? '.m3u'
    @store_where = i
  end

  # ========================================================================= #
  # === consider_sanitizing_dataset
  # ========================================================================= #
  def consider_sanitizing_dataset
    _ = @dataset
    if (_.is_a?(String) or _.is_a?(Array)) and _.empty?
      _ = Dir['**/**'].select {|file|
        is_audio_song?(file)
      }.map {|file|
        # =================================================================== #
        # Keep the full pathname always.
        # =================================================================== #
        unless file.include? '/'
          file = rds("#{Dir.pwd}/#{file}")
        end
        file
      }
      set_dataset(_)
    end
  end

  # ========================================================================= #
  # === dataset?
  # ========================================================================= #
  def dataset?
    @dataset
  end

  # ========================================================================= #
  # === store_where?
  # ========================================================================= #
  def store_where?
    @store_where
  end; alias name_of_the_file? store_where? # === name_of_the_file?

  # ========================================================================= #
  # === set_dataset
  #
  # The @dataset will essentially keep track of our songs.
  # ========================================================================= #
  def set_dataset(
      i = obtain_songs_from_the_default_directory
    )
    case i
    # ======================================================================= #
    # === :default_playlist
    # ======================================================================= #
    when :default_playlist,
         :default
      i = obtain_songs_from_the_default_directory
    end
    if i.is_a? Array
      if i.empty?
        # =================================================================== #
        # Try to populate the entry in the event that it is empty:
        # =================================================================== #
        i = obtain_songs_from_the_default_directory
      end
      # ===================================================================== #
      # If we have an Array, we will always sort alphabetically.
      # ===================================================================== #
      i.sort_by! {|file|
        File.basename(file)
      }
      i = i.join(N)
    # ======================================================================= #
    # Handle Hash-input next. Hashes are a bit more complicated than Arrays
    # here since we have to pick a relevant key.
    # ======================================================================= #
    elsif i.is_a? Hash
      values = i.values
      if values.first.is_a?(Hash) and values.first.has_key?('url')
        # =================================================================== #
        # This could be a complex Hash.
        # =================================================================== #
        i = values.map {|hash|
          hash['url']
        }
      end
    end
    if debug?
      opnn; ecrimson('Debugging information: We will show the '\
            'dataset that will be used.')
      e
      pp i
      e
    end
    @dataset = i
  end

  # ========================================================================= #
  # === store_file_into
  #
  # This method will create the .m3u file at hand.
  # ========================================================================= #
  def store_file_into(
      into = name_of_the_file?
    )
    what = dataset?
    if @be_verbose
      if what.is_a? Hash
        n_entries = what.keys.size
      elsif what.is_a? Array
        n_entries = what.size
      else
        n_entries = what.count(N)
      end
      into = File.absolute_path(into)
      opnn; e "Will store into `#{sfile(into)}` "\
              "(#{n_entries} entries)."
    end
    if what.is_a? Hash
      # ===================================================================== #
      # If we input a Hash then we will only use the values of
      # said Hash.
      # ===================================================================== #
      what = what.values.join(N)
    elsif what.is_a? Array
      what = what.join(N)
    end
    write_what_into(what, into)
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    consider_sanitizing_dataset
    store_file_into
  end

  # ========================================================================= #
  # === MultimediaParadise::CreateM3uPlaylist[]
  # ========================================================================= #
  def self.[](i = '')
    new(i)
  end

end

# =========================================================================== #
# === MultimediaParadise.create_m3u_playlist
# =========================================================================== #
def self.create_m3u_playlist(i = :default_playlist)
  ::MultimediaParadise::CreateM3uPlaylist.new(i)
end; self.instance_eval { alias create_default_playlist create_m3u_playlist } # === MultimediaParadise.create_default_playlist

end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::CreateM3uPlaylist.new(ARGV)
end # cplaylist