#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::Genre
#
# This class can do lots of things in regards to a yaml file (such as
# the file 'eurodance.yml'), which has to be supplied to this class.
#
# The class can then respond to the following commandline instructions:
#
#   download          # download everything into a newly created directory.
#                     # This will also rename things properly. [IMPLEMENTED]
#   generate_homepage # generate a homepage [IMPLEMENTED]
#   overview          # Show what we have in that yaml file, in a colourized
#                     # manner. [IMPLEMENTED]
#   only_names        # Display only the names of the registered songs [IMPLEMENTED]
#   open              # open the playlist via your editor [IMPLEMENTED]
#   playlist          # Generate a .m3u playlist from these results. [IMPLEMENTED]
#   validity          # Ensure that each entry is strictly and fully
#                     # correct. [NOT YET IMPLEMENTED]
#   years             # sort via years [IMPLEMENTED]
#   urls              # show only the URLs, without anything else [IMPLEMENTED]
#
# Additionally, if you supply a number, such as "eurodance 35", then we
# will download only that file at that position (starting to count at
# track 1), from a remote position.
#
# Usage example:
#
#   MultimediaParadise::Eurodance.new(ARGV)
#
# =========================================================================== #
# require 'multimedia_paradise/audio/genres/class.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

class Genre < ::MultimediaParadise::Base # === MultimediaParadise::Eurodance

  require 'multimedia_paradise/audio/create_m3u_playlist.rb'
  require 'multimedia_paradise/audio/genres/constants.rb'

  begin
    require 'html_tags'
  rescue LoadError; end

  include MultimediaParadise::Genres

  # ========================================================================= #
  # === DEFAULT_FILE
  # ========================================================================= # 
  DEFAULT_FILE = MultimediaParadise::Genres::EURODANCE_FILE

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = nil,
      run_already = true
    )
    reset
    set_yaml_file(
      DEFAULT_FILE
    )
    set_commandline_arguments(i)
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @dataset
    # ======================================================================= #
    @dataset = nil
    set_commandline_arguments
    set_namespace
  end

  # ========================================================================= #
  # === set_yaml_file
  # ========================================================================= #
  def set_yaml_file(i)
    @yaml_file = i
    # ======================================================================= #
    # === Initialize the yaml dataset
    # ======================================================================= #
    load_yaml_dataset # Must re-initialize @dataset whenever set_input() is called.
  end

  # ========================================================================= #
  # === yaml_file?
  #
  # This will always point at the main yaml file.
  # ========================================================================= #
  def yaml_file?
    @yaml_file
  end

  # ========================================================================= #
  # === show_the_location_of_the_yaml_file
  #
  # To invoke this method, try:
  #
  #   eurodance --file?
  #
  # ========================================================================= #
  def show_the_location_of_the_yaml_file
    e sfile(@yaml_file)
  end

  # ========================================================================= #
  # === set_commandline_arguments
  # ========================================================================= #
  def set_commandline_arguments(i = DEFAULT_FILE)
    i = [i] unless i.is_a? Array
    i = [DEFAULT_FILE] if i.compact.empty?
    @commandline_arguments = i
  end

  # ========================================================================= #
  # === set_namespace
  # ========================================================================= #
  def set_namespace(i = NAMESPACE)
    @namespace = i
  end

  # ========================================================================= #
  # === open_playlist_in_your_editor
  #
  # Invoke such as by doing:
  #
  #   trance --open
  #
  # ========================================================================= #
  def open_playlist_in_your_editor
    _ = yaml_file?
    _ = 'bluefish '+_
   esystem(_)
  end

  # ========================================================================= #
  # === load_yaml_dataset
  #
  # This method will also make @dataset available.
  # ========================================================================= #
  def load_yaml_dataset
    _ = yaml_file?
    begin # Must protect it because the yaml file may be buggy and raise Psych::SyntaxError.
      @dataset = YAML.load_file(_)
    rescue ArgumentError # This clause happens when syck raises an error.
      @dataset = File.read(_)
    rescue Exception => error
      opnn; e 'An error happened on line '+__LINE__.to_s+'.'
      pp error
      pp error.class
    end
    sanitize_yaml_dataset # Must sanitize it.
  end

  # ========================================================================= #
  # === sanitize_yaml_dataset
  # ========================================================================= #
  def sanitize_yaml_dataset
    if @dataset.is_a? String
      # ===================================================================== #
      # Do the manual parsing on your own if it is a String.
      # ===================================================================== #
      _ = @dataset.split(N)
      _.reject! {|entry| entry.start_with? '#' }
      _.reject! {|entry| entry.strip.empty? }
      # ===================================================================== #
      # This is presently unfinished. I may continue working on this if
      # I happen to use a plain textfile rather than a yaml file.
      # ===================================================================== #
    end
  end

  # ========================================================================= #
  # === menu                                                       (menu tag)
  # ========================================================================= #
  def menu(i = commandline_arguments?)
    if i.is_a? Array
      # ===================================================================== #
      # === Must handle input such as "play foo"
      # ===================================================================== #
      if i.last =~ /^\d+$/
        i = [i.join(' ')]
      end
      i.each {|entry| menu(entry) }
    else
      case i.delete('-').downcase # case tag
      # ===================================================================== #
      # === eurodance --download
      # ===================================================================== #
      when /^-?-?download$/,
           '--dl',
           'dl'
        download_remote_files
        exit
      # ===================================================================== #
      # === eurodance --generate-homepage
      #
      # This entry point can be used to generate a small "webpage"
      # with content of the playlist at hand.
      #
      # Invocation example:
      #
      #   trance --html
      #
      # ===================================================================== #
      when /^-?-?generate(_|-)?homepage$/i,
           'genhomepage',
           'homepage',
           'web',
           /^-?-?webpage$/i,
           /^-?-?html$/i
        generate_homepage
        exit
      # ===================================================================== #
      # === trance play 17
      #
      # This entry point allows us to open a specific file at a position
      # and "play" it, e. g. in the browser.
      #
      # Example for this:
      #
      #   trance play 3
      #
      # ===================================================================== #
      when /^play (\d+)/,
           /^open (\d+)/
        require 'open'
        url = obtain_url_from_entry_at_this_position($1.to_s.dup)
        open_in_browser(url)
        exit
      # ===================================================================== #
      # === eurodance --playlist
      # ===================================================================== #
      when '--playlist',
           'playlist',
           '--generate_playlist',
           '--generateplaylist',
           '--generate-playlist',
           'generateplaylist',
           'm3u',
           '--m3u'
        generate_m3u_playlist # trance --generate-playlist
        exit
      # ===================================================================== #
      # === eurodance --only-URLs
      # ===================================================================== #
      when /^-?-?show(_|-)?only(_|-)?the(_|-)?URLs$/i,
           /^-?-?only(_|-)?URLs$/i,
           /^-?-?URLs?$/i # === eurodance --url
        show_only_the_URLs
        exit
      # ===================================================================== #
      # === eurodance --only_names
      # ===================================================================== #
      when /^-?-?only(_|-)?names$/i
        show_only_names
        exit
      # ===================================================================== #
      # === eurodance --file?
      # ===================================================================== #
      when /^-?-?file\??/
        show_the_location_of_the_yaml_file
        exit
      # ===================================================================== #
      # === eurodance --help
      # ===================================================================== #
      when '--help','help','?'
        show_help(:then_exit)
      # ===================================================================== #
      # === eurodance --open
      # ===================================================================== #
      when 'open','OPEN','--open','op','o'
        open_playlist_in_your_editor
        exit
      # ===================================================================== #
      # === trance --reverse-years
      # ===================================================================== #
      when '--reverse-years','reverseyears','--sort-by-years-reverse',
           'sortbyyearsreverse'
        sort_playlist_via_years(:reverse)
        do_the_default_action
      # ===================================================================== #
      # === trance --years
      # ===================================================================== #
      when '--year','--years','--sort','--sort-via-year','--sort-by-years',
           'sortbyyears','years','year','sort','sortviayear'
        sort_playlist_via_years
        do_the_default_action
      # ===================================================================== #
      # === eurodance --overview
      # ===================================================================== #
      when '--overview','overview'
        show_available_songs(:then_exit)
      # ===================================================================== #
      # === eurodance 62
      # ===================================================================== #
      when /^(\d+)$/
        download_file_at_number($1.to_s.dup)
      else
        default_action
      end
    end
  end

  # ========================================================================= #
  # === do_the_default_action
  #
  # Presently, the default action is to show all available songs.
  # ========================================================================= #
  def do_the_default_action
    show_available_songs(:then_exit)
  end; alias default_action do_the_default_action # === default_action

  # ========================================================================= #
  # === show_available_songs
  # ========================================================================= #
  def show_available_songs(
      use_this_dataset = @dataset,
      shall_we_exit    = :then_exit
    )
    if use_this_dataset == :then_exit
      use_this_dataset = @dataset
      shall_we_exit = :then_exit
    end
    case shall_we_exit
    when :then_exit
      shall_we_exit = true
    end
    # ======================================================================= #
    # Work with a pointer variable.
    # ======================================================================= #
    _ = use_this_dataset # We assume that this will always be a Hash.
    n_entries = _.keys.size
    opnn; e simp(n_entries.to_s)+' entries were registered in the yaml file.'
    e
    _.each_pair.with_index {|inner_hash, index|
      key, _array = *inner_hash
      url = _array[1]
      index += 1
      group_name = _array[0]
      release_year = _array[2].to_s
      group_and_release_year = ' ('+group_name+', '+release_year+')'
      e slateblue(index.to_s.rjust(3)+') ')+simp(key)+dodgerblue(group_and_release_year)
      # ===================================================================== #
      # Next line displays the URL
      # ===================================================================== #
      left_pad = 7
      e powderblue((' ' * left_pad)+url) # Left-padded URL.
    }; e
    exit if shall_we_exit
  end

  # ========================================================================= #
  # === sort_playlist_via_years
  #
  # This method sorts the playlist via the years. We can work on the
  # @dataset variable for this to work.
  #
  # You can try this method like this:
  #   trance --sort-by-years
  # ========================================================================= #
  def sort_playlist_via_years(
      sort_mode_to_use = :normal
    )
    if sort_mode_to_use == :normal
      _ = @dataset.sort_by {|key, array|
        value = array[2]
        value = 0 if value.nil? # "Rescue".
        value
      }
    else
      _ = @dataset.sort_by {|key, array|
        -array[2]
      }
    end
    @dataset = Hash[_]
  end

  # ========================================================================= #
  # === show_only_names
  #
  # To invoke this, do:
  #
  #   eurodance --onlynames
  #
  # ========================================================================= #
  def show_only_names
    @dataset.each_pair {|key, dataset|
      e slateblue(key.to_s)
    }
  end

  # ========================================================================= #
  # === show_only_the_URLs
  #
  # To invoke this, do:
  #
  #   eurodance --only-URLs
  #
  # ========================================================================= #
  def show_only_the_URLs
    @dataset.each_pair {|key, dataset|
      url = dataset[1]
      e slateblue(url.to_s)
    }
  end

  # ========================================================================= #
  # === generate_m3u_playlist
  # ========================================================================= #
  def generate_m3u_playlist
    # ======================================================================= #
    # We only need an Array of all URLs.
    # ======================================================================= #
    dataset = return_all_remote_urls
    into = '/Depot/j/'+return_name_from_class_name
    CreateM3uPlaylist.new(dataset, into)
  end

  # ========================================================================= #
  # === dataset?
  # ========================================================================= #
  def dataset?
    @dataset
  end; alias full_dataset? dataset? # === full_dataset?

  # ========================================================================= #
  # === return_all_remote_urls
  # ========================================================================= #
  def return_all_remote_urls
    dataset = @dataset
    dataset = dataset.map {|key, inner_array|
      inner_array[1] # This is the URL.
    }
    return dataset
  end

  # ========================================================================= #
  # === return_dataset_from_one_random_song
  #
  # This is sort of a debug-method. What we do here is that we will
  # randomly return one element from the main dataset.
  # ========================================================================= #
  def return_dataset_from_one_random_song
    return dataset?.to_a.sample(1).to_h
  end

  # ========================================================================= #
  # === rename_this_file
  #
  # This will rename the file and sanitize it, then return the new file
  # name.
  # ========================================================================= #
  def rename_this_file(i)
    old_filename = i.to_s.dup
    new_filename = old_filename.delete('(').delete(')').
                   gsub(/ - /,'-').tr(' ','_').delete('!')
    mv(old_filename, new_filename)
    return new_filename
  end

  # ========================================================================= #
  # === download_remote_files
  #
  # Use this method when you want to download the remote files.
  #
  # The input to this method should be a Hash.
  #
  # Invocation example:
  #
  #   eurodance --download
  #
  # ========================================================================= #
  def download_remote_files(this_dataset = full_dataset?)
    ::MultimediaParadise.ensure_that_the_output_directory_exists
    store_here = ::MultimediaParadise.output_directory?+return_name_from_class_name+'/'
    mkdir(store_here)
    cd(store_here)
    this_dataset.each_pair {|name_of_the_song, array|
      url = array[1]
      dataset = { name_of_the_song => array }
      _ = dataset
      key = _.keys.first
      name_of_the_group = _[key][0]
      opnn; e
      e '  Now downloading the song '+dodgerblue(name_of_the_song)+
        ' (Group: '+simp(name_of_the_group)+') from the URL '+sfancy(url)
      name_of_the_song = key.to_s.delete('!')
      url = _[key][1]
      release_year = _[key][2].to_s # Need a String.
      # ======================================================================= #
      # Next, we will use youtube-dl to download this.
      # ======================================================================= #
      result = `youtube-dl #{url}`
      use_this_regex = /Merging formats into "(.+)"/
      result.scan(use_this_regex)
      e 'Working on the song '+simp(name_of_the_song)+' next.'
      new_file_name = $1.to_s.dup
      # ======================================================================= #
      # The filename has to be sanitized and renamed.
      # ======================================================================= #
      new_file_name = rename_this_file(new_file_name)
      # ======================================================================= #
      # Ok, the file is already a bit sanitized, but we wish to make it even
      # nicer. So, we rename it based on 1) the name of the group, 2) the
      # name of the song and 3) the release year.
      # ======================================================================= #
      extension_name = File.extname(new_file_name)
      _ = name_of_the_group.tr(' ','_')+'-'+name_of_the_song.tr(' ','_')+
          '_Year_'+release_year+extension_name
      mv(new_file_name, _)
      opnn; e 'The file name now is `'+sfile(_)+'`, in the '\
              'directory `'+sdir(store_here)+'`.'
    } 
  end; alias download_remote_file download_remote_files # === download_remote_file

  # ========================================================================= #
  # === download_file_at_number
  #
  # Invocation example:
  #   eurodance 62
  # ========================================================================= #
  def download_file_at_number(i)
    i = i.to_i
    _ = full_dataset?
    hash = {}
    _.each_pair.with_index {|data, index|
      if index == (i-1)
        data = Hash[*data]
        hash.update(data)
      end
    }
    download_remote_file(hash)
  end

  # ========================================================================= #
  # === return_name_from_class_name
  # ========================================================================= #
  def return_name_from_class_name(i = @namespace)
    _ = @namespace.to_s.downcase
    _ = _.split(':').last if _.include? ':'
    return _
  end

  # ========================================================================= #
  # === obtain_url_from_entry_at_this_position
  #
  # We obtain (and return) the remote URL from this entry.
  # ========================================================================= #
  def obtain_url_from_entry_at_this_position(i)
    # ======================================================================= #
    # Input must be an Integer. We also must deduct -1 because Ruby Array
    # start to count at 0.
    # ======================================================================= #
    i = i.to_i - 1
    _ = full_dataset?
    keys = _.keys
    url = _[keys[i]][1]
    return url
  end

  # ========================================================================= #
  # === show_help                                                  (help tag)
  #
  # To invoke this, try:
  #
  #   eurodance --help
  #
  # ========================================================================= #
  def show_help(
      shall_we_exit = true,
      show_opn      = false
    )
    if shall_we_exit == :then_exit
      shall_we_exit = true
    end
    e
    opnn if show_opn; e '  --download           # to download all '+name_of_the_genre?+
                        ' songs locally'
    opnn if show_opn; e '  --generate-homepage  # to generate a homepage'
    opnn if show_opn; e '  --m3u                # to generate a .m3u playlist (a file)'
    opnn if show_opn; e '  --open               # to open the respective playlist in your editor'
    opnn if show_opn; e '  --overview           # to show which entries are registered'
    opnn if show_opn; e '  --years              # to sort via years (most recent year will come last)'
    opnn if show_opn; e '  --file?              # to show where the yaml file is stored'
    opnn if show_opn; e '  --reverse-years      # to sort via years in a reverse manner '\
                        '(most recent year will come first)'
    opnn if show_opn; e '  play 17              # open the track at position 17 in the browser'
    opnn if show_opn; e '                       # so that you can play it'
    e
    exit if shall_we_exit
  end

  # ========================================================================= #
  # === guess_the_title_based_on_the_yaml_file
  #
  # This method will try to guess the filename. It may use strings such as:
  #
  #   "/home/Programs/Ruby/2.7.2/lib/ruby/site_ruby/2.7.0/multimedia_paradise/yaml/genres/trance.yml"
  #
  # ========================================================================= #
  def guess_the_title_based_on_the_yaml_file
    File.basename(@yaml_file).sub(/\.yml$/,'')
  end

  # ========================================================================= #
  # === name_of_the_genre?
  #
  # This method will return the genre in use.
  # ========================================================================= #
  def name_of_the_genre?
    _ = self.class.name.to_s
    _ = _.split(':').last if _.include? ':'
    return _.gsub(/The/,'')
  end

  # ========================================================================= #
  # === opnn
  # ========================================================================= #
  def opnn(i = NAMESPACE)
    super(i)
  end

  # ========================================================================= #
  # === generate_homepage
  #
  # This will generate a webpage/homepage, by using the links from
  # the genre.
  #
  # Invocation example:
  #
  #   eurodance --homepage
  #
  # ========================================================================= #
  def generate_homepage
    _ = '<html><body>'.dup # This will be the return string.
    _ << '<title>'+guess_the_title_based_on_the_yaml_file+
         ' ('+@dataset.keys.size.to_s+' entries)</title>'
    index = 0
    dataset?.each_pair {|name_of_the_song, array|
      index += 1
      name_of_the_group = array.first
      remote_URL = array[1]
      _ << HtmlTags.a(remote_URL,
        description: '('+index.to_s+') '+
                     name_of_the_song+' - '+
                     name_of_the_group
      )
      _ << "<br>\n"
    }
    _ << '</body></html>'
    # ======================================================================= #
    # We determine into which file we will generate. Since as of October
    # 2020 we will use the general log-directory for the
    # MultimediaParadise project for this.
    # ======================================================================= #
    into = self.class.inspect.to_s.split(':').last.downcase
    into = log_dir?+into+'.html'
    opnn; e "Now storing into the file `#{sfile(into)}`."
    write_what_into(_, into)
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    menu
  end

  # ========================================================================= #
  # === MultimediaParadise::Genre[]
  # ========================================================================= #
  def self.[](i = '')
    new(i)
  end

end; end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::Genre.new(ARGV)
end # genre