#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::LyricsFetcher
#
# This class will fetch the lyrics file, then group the lyrics into
# subsections, then return a proper match if it can.
#
# Usage example:
#
#   _ = MultimediaParadise::LyricsFetcher.new
#   _.search_for 'Luca' #
#
# =========================================================================== #
# require 'multimedia_paradise/audio/lyrics_fetcher.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

class LyricsFetcher < MultimediaParadise::Base # === MultimediaParadise::LyricsFetcher

  # ========================================================================= #
  # === SEARCH_TERM
  # ========================================================================= #
  SEARCH_TERM = 'Luca'

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      search_term = nil,
      run_already = true
    )
    reset
    set_search_term(
      search_term
    )
    run if run_already
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @hash
    # ======================================================================= #
    @hash = {}
    infer_the_namespace
  end

  # ========================================================================= #
  # === set_search_term
  # ========================================================================= #
  def set_search_term(i = SEARCH_TERM)
    if i.is_a? Array
      i = i.join(' ').strip
    end
    case i
    when nil, :default
      i = SEARCH_TERM
    end
    @search_term = i
  end

  # ========================================================================= #
  # === main_file?
  # ========================================================================= #
  def main_file?
    LYRICS_YAML_FILE
  end

  # ========================================================================= #
  # === show_help                                                  (help tag)
  # ========================================================================= #
  def show_help
    opnn; e 'These options are available:'
    e
    opnn; e '  --file?            # Feedback the full path to the .yml '\
            'file in question'
    opnn; e '  --available_titles # Feedback the available titles'
    e
  end

  # ========================================================================= #
  # === obtain_raw_dataset
  # ========================================================================= #
  def obtain_raw_dataset
    @hash[:raw_dataset] = nil
    begin
      @hash[:raw_dataset] = YAML.load_file(main_file?)
    rescue Exception => e
      pp e
    end
  end; alias raw_dataset?      obtain_raw_dataset # === raw_dataset?
       alias load_yaml_dataset obtain_raw_dataset # === load_yaml_dataset

  # ========================================================================= #
  # === open_yaml_file_in_editor
  #
  # Invocation example:
  #
  #   lfetcher open
  #
  # ========================================================================= #
  def open_yaml_file_in_editor
    esystem "bluefish #{main_file?}"
  end

  # ========================================================================= #
  # === menu
  #
  # The menu will, by default, try to find a proper title.
  # ========================================================================= #
  def menu(
      i = @search_term
    )
    case i # case tag
    # ======================================================================= #
    # === lyrics --help
    # ======================================================================= #
    when *ARRAY_HELP_OPTIONS
      show_help
      exit
    # ======================================================================= #
    # === lyrics open
    # ======================================================================= #
    when 'open'
      open_yaml_file_in_editor
      exit
    # ======================================================================= #
    # === lyrics fat
    # ======================================================================= #
    when 'feedback_available_titles','fat','--feedback',
         '--available_titles',/available-?titles/
      feedback_available_titles
      exit
    # ======================================================================= #
    # === file?
    # ======================================================================= #
    when 'file?','--file?','--location'
      show_file_path
      exit
    else
      try_to_find_this_title(i)
    end if i
  end

  # ========================================================================= #
  # === try_to_find_this_title
  #
  # This method will try to find an appropriate title.
  # ========================================================================= #
  def try_to_find_this_title(i)
    _ = raw_dataset?
    keys = _.keys.map
    possible_matches = keys.select {|entry|
      entry = entry.dup if entry.frozen?
      entry.downcase.include? i.downcase
    }
    if possible_matches.empty?
      opnn; e 'Did not find that particular key.'
    else
      @hash[:text] = _[possible_matches.first]
      report(@hash[:text])
    end
  end

  # ========================================================================= #
  # === report
  # ========================================================================= #
  def report(this = @hash[:text])
    # ======================================================================= #
    # Ok, we found it, so display it next:
    # ======================================================================= #
    cliner colour: :konsole_slateblue
    e this
    cliner colour: :konsole_slateblue
  end

  # ========================================================================= #
  # === show_file_path
  # ========================================================================= #
  def show_file_path
    e LYRICS_YAML_FILE
  end

  # ========================================================================= #
  # === lyrics?
  # ========================================================================= #
  def lyrics?
    @hash[:raw_dataset]
  end

  # ========================================================================= #
  # === fetch_subsections_from_dataset
  # ========================================================================= #
  def fetch_subsections_from_dataset
    lyrics?.each {|line|
      e line.delete(',').delete('"') if line.include? 'add_song'
    }
  end

  # ========================================================================= #
  # === feedback_available_titles
  # ========================================================================= #
  def feedback_available_titles
    _ = obtain_raw_dataset
    if _.is_a? Hash
      counter = 0
      cliner
      _.each_pair {|title, value|
        counter += 1
        e; e '('+counter.to_s.rjust(2)+') '+simp(title)
        e; e value; e
        cliner
      }
    end
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    menu
    load_yaml_dataset
    fetch_subsections_from_dataset
  end

end; end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::LyricsFetcher.new(ARGV)
end # lyrics
    # lyrics Michel Telo