#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::ReportMissingId
#
# This class will iterate over the given .mp3 files and report missing
# ids, such as a missing :genre or :comment entry.
#
# Usage example:
#
#   MultimediaParadise::ReportMissingId.new(ARGV)
#
# =========================================================================== #
# require 'multimedia_paradise/audio/report_missing_id.rb'
# MultimediaParadise.report_missing_id
# =========================================================================== #
require 'multimedia_paradise/audio/base.rb'

module MultimediaParadise

class ReportMissingId < MultimediaParadise::AudioBase # === MultimediaParadise::ReportMissingId

  require 'multimedia_paradise/audio/audio_tag_reader/audio_tag_reader.rb'

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === ARRAY_CHECK_ON_THESE_ENTRIES
  #
  # Here you can specify which mp3 tags are to be checked.
  # ========================================================================= #
  ARRAY_CHECK_ON_THESE_ENTRIES = %i(
    year
    Artist
  )

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = ARGV,
      run_already = true
    )
    reset
    set_commandline_arguments(
      i
    )
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
  end

  # ========================================================================= #
  # === work_on_these_files
  #
  # This method will work on the given audio files, and identify
  # which entry is missing.
  # ========================================================================= #
  def work_on_these_files(i)
    problem_encountered = 0
    i.each {|this_audio_file|
      audio_tag_reader = MultimediaParadise::AudioTagReader.new(this_audio_file) { :do_not_report_anything }
      # ===================================================================== #
      # Next, we will check against the main Array.
      # ===================================================================== #
      ARRAY_CHECK_ON_THESE_ENTRIES.each {|this_entry|
        result = audio_tag_reader.send(
          (this_entry.to_s.downcase+'?').to_sym
        )
        case this_entry
        when :Artist, :artist
          if result.to_s.empty?
            problem_encountered += 1
            e problem_encountered.to_s+') No artist entry for: '+sfile(this_audio_file)
          end
        when :year
          if result.to_s == '0'
            problem_encountered += 1
            e problem_encountered.to_s+') '\
              'Faulty year entry for '+sfile(this_audio_file)+'; '\
              'the value was 0.'
            e tomato(
                '  This means that no year has been registered for this audio file.'
              )
          end
        end
      }
    }
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    first = first?
    first = return_pwd if first.nil?
    _ = [first].map {|entry|
      if File.directory?(entry)
        entry = Dir[entry+'**']
      end
      entry
    }.flatten.select {|entry| is_an_audio_file?(entry) }
    work_on_these_files(_)
  end

  # ========================================================================= #
  # === MultimediaParadise::ReportMissingId[]
  # ========================================================================= #
  def self.[](i = '')
    new(i)
  end

end

# =========================================================================== #
# === MultimediaParadise.report_missing_id
# =========================================================================== #
def self.report_missing_id(i = :default_playlist)
  ::MultimediaParadise::ReportMissingId.new(i)
end; self.instance_eval { alias report_missing_ids report_missing_id } # === MultimediaParadise.report_missing_ids

end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::ReportMissingId.new(ARGV)
end # reportmissingid /home/x/songs/