#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::Audio::StreamripperWrapper
#
# This class will be used to play or rip content from streamripper
# and shoutcast things.
# =========================================================================== #
# require 'multimedia_paradise/audio/streamripper/streamripper_wrapper.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'
require 'multimedia_paradise/audio/streamripper/constants.rb'

module MultimediaParadise

module Audio

class StreamripperWrapper < ::MultimediaParadise::Base

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      play_from_where = denver?, # Denver is the default for now.
      run_already     = false
    )
    register_sigint
    reset
    if play_from_where
      set_use_this_url(play_from_where)
    end 
    run if run_already
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @string
    # ======================================================================= #
    @string = ''.dup
    @can_we_exit = true
    @use_this_url = HOUSE_MUSIC # URL_2 # This is the default.
    @run_in_simulation_mode = false # If true then we just simulate what we would do.
    @be_verbose = false    
  end

  # ========================================================================= #
  # === denver?
  # ========================================================================= #
  def denver?
    HASH_AVAILABLE_URLS['Denver area radio']
  end

  # ========================================================================= #
  # === bootliquor?
  # ========================================================================= #
  def bootliquor?
    HASH_AVAILABLE_URLS['Bootliquor']
  end; alias default_url? bootliquor? # === default_url? # This is currently the default streamripper.

  # ========================================================================= #
  # === set_use_this_url
  # ========================================================================= #
  def set_use_this_url(i = nil)
    i = i.first if i.is_a? Array
    i = denver? if i.nil?
    if i.include? '--'
      case i
      when /-?-?help/
        show_help; exit
      end
    end
    @use_this_url = i
  end

  # ========================================================================= #
  # === show_help
  # ========================================================================= #
  def show_help
    e 'Presently, this class does not allow any option.'
    e 'Perhaps in the future you may be able to set another URL '\
      'than the default one.'
  end

  # ========================================================================= #
  # === run_in_simulation_mode?
  # ========================================================================= #
  def run_in_simulation_mode?
    @run_in_simulation_mode
  end

  # ========================================================================= #
  # === do_not_exit
  # ========================================================================= #
  def do_not_exit
    @can_we_exit = false
  end

  # ========================================================================= #
  # === check_whether_streamripper_is_available
  # ========================================================================= #
  def check_whether_streamripper_is_available
    result = `streamripper 2>&1`
    if result.include? 'command not found'
      warn_and_exit 'streamripper is not installer. Please install it first.'
    else
      if be_verbose?
        efancy 'Streamripper is available. We can thus continue.'
      end
    end
  end

  # ========================================================================= #
  # === set_base_command
  # ========================================================================= #
  def set_base_command
    @string << 'streamripper '
  end

  # ========================================================================= #
  # === use_this_directory_as_base=
  # ========================================================================= #
  def use_this_directory_as_base=(i)
    Dir.chdir(i) if Dir.exist? i
  end

  # ========================================================================= #
  # === url?
  #
  # Feedback the project url here.
  # ========================================================================= #
  def url?
    e 'http://sourceforge.net/project/showfiles.php?group_id=6172'
  end

  # ========================================================================= #
  # === rip_to_single_file
  #
  # Rip to single file, default name is timestamped
  # ========================================================================= #
  def rip_to_single_file(this_file = nil)
    @string << ' -a '
    @string << this_file unless this_file.nil?
  end

  # ========================================================================= #
  # === dont_write_individual_files
  #
  # With this we deny streamripper writing individual files.
  # ========================================================================= #
  def dont_write_individual_files
    @string << ' -A'
  end

  # ========================================================================= #
  # === set_destination_directory
  #
  # This sets our destination directory
  # ========================================================================= #
  def set_destination_directory(this_dir)
    @string << ' -d '+this_dir
  end

  # ========================================================================= #
  # === set_pattern
  #
  # Write files using specified pattern. Dunno yet what this means.
  # ========================================================================= #
  def set_pattern(this_pattern)
    @string << ' -D '+this_pattern
  end

  # ========================================================================= #
  # === start_streamripper_server
  # ========================================================================= #  
  def start_streamripper_server
    unless @can_we_exit
      e 'Starting the streamripper-server now from dir '+dir?+':'
      e simp('  '+@string)
      system @string unless run_in_simulation_mode?
    end
  end

  # ========================================================================= #
  # === dir?
  #
  # Provide to us a colourized variant.
  # ========================================================================= #
  def dir?
    sdir((Dir.pwd+'/').squeeze('/'))
  end

  # ========================================================================= #
  # === set_url
  #
  # Set the URL we want to use. This should point to something that
  # can serve a .pls file.
  # ========================================================================= #
  def set_url(this_url = @use_this_url)
    if this_url.is_a? Array
      this_url = this_url.first
    end
    if this_url.nil?
      this_url = default_url?
    end
    @url = " #{this_url.strip}"
  end

  # ========================================================================= #
  # === warn_and_exit
  # ========================================================================= #
  def warn_and_exit(i)
    e i
    exit if @can_we_exit
  end

  # ========================================================================= #
  # === append
  # ========================================================================= #
  def append(i)
    @string << i
  end

  # ========================================================================= #
  # === sanitize_command
  # ========================================================================= #
  def sanitize_command
    @string.squeeze! ' '
  end

  # ========================================================================= #
  # === append_where_we_store
  # ========================================================================= #
  def append_where_we_store
    append ' -d /Depot/Temp/Audio/'
  end

  # ========================================================================= #
  # === build_command
  # ========================================================================= #
  def build_command
    append @url
    append_where_we_store
    sanitize_command
  end

  # ========================================================================= #
  # === run
  #
  # Do the actual invocation part here.
  # ========================================================================= #
  def run
    check_whether_streamripper_is_available
    set_base_command
    set_url # mandatory
    build_command
    start_streamripper_server
  end

end; end

# =========================================================================== #
# === MultimediaParadise.streamripper
# =========================================================================== #
def self.streamripper(i = nil)
  MultimediaParadise::Audio::StreamripperWrapper.new(i)
end

end

if __FILE__ == $PROGRAM_NAME
  _ = MultimediaParadise.streamripper(ARGV)
  _.do_not_exit
  _.run
end # rstreamripper