#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::Waveform::Log
#
# The purpose of this class is to log and provide benchmarks.
#
# This may be important for lengthy batch operations.
# 
# There may be other, better benchmark classes, but it is fairly trivial
# code - not mandatory to google for better solutions.
# =========================================================================== #
# require 'multimedia_paradise/audio/waveform/log.rb'
# =========================================================================== #
module MultimediaParadise

class Waveform < ::MultimediaParadise::Base

class Log # === MultimediaParadise::Waveform::Log

  attr_accessor :io

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(io = $stdout)
    @io = io
  end

  # ========================================================================= #
  # === out
  #
  # Prints the given message to the log.
  # ========================================================================= #
  def out(i)
    io.print(i) if io
  end

  # ========================================================================= #
  # === done!
  #
  # Prints the given message to the log followed by the most recent
  # benchmark (note that it calls .end! which will stop the benchmark)
  # ========================================================================= #
  def done!(message = '')
    out "#{message} (#{self.end!}s)\n"
  end

  # ========================================================================= #
  # === start!
  #
  # Starts a new benchmark clock and returns the index of the new clock.
  # 
  # If .start! is called again before .end! then the time returned will
  # be the elapsed time from the next call to start!, and calling .end!
  # again will return the time from *this* call to start! (that is, the
  # clocks are LIFO).
  # ========================================================================= #
  def start!
    (@benchmarks ||= []) << Time.now
    @current = @benchmarks.size - 1
  end

  # ========================================================================= #
  # === end!
  #
  # Returns the elapsed time from the most recently started benchmark clock
  # and ends the benchmark, so that a subsequent call to .end! will return
  # the elapsed time from the previously started benchmark clock.
  # ========================================================================= #
  def end!
    elapsed = (Time.now - @benchmarks[@current])
    @current -= 1
    elapsed
  end

  # ========================================================================= #
  # === time?
  #
  # Returns the elapsed time from the benchmark clock w/ the given index
  # (as returned from when .start! was called).
  # ========================================================================= #
  def time?(index)
    Time.now - @benchmarks[index]
  end

  # ========================================================================= #
  # === timed
  #
  # Benchmarks the given block, printing out the given message first
  # (if given).
  # ========================================================================= #
  def timed(message = nil, &block)
    start!
    out(message) if message
    yield
    done!
  end

end; end; end