#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'multimedia_paradise/base/misc.rb'
# =========================================================================== #
module MultimediaParadise

class Base # === MultimediaParadise::Base

  require 'fileutils'
  require 'multimedia_paradise/toplevel_methods/ensure_that_the_output_directory_exists.rb'
  require 'multimedia_paradise/base/commandline_arguments.rb'
  require 'multimedia_paradise/base/time.rb'

  begin
    require 'opn'
  rescue LoadError; end

  begin
    require 'save_file'
  rescue LoadError; end

  begin
    require 'cliner'
  rescue LoadError; end

  begin
    require 'roebe/classes/twentyfour_hours_notation.rb'
  rescue LoadError; end

  # ========================================================================= #
  # === internal_hash?
  # ========================================================================= #
  def internal_hash?
    @internal_hash
  end

  # ========================================================================= #
  # === reset_the_internal_hash
  #
  # The method name is a misnomer, but nonetheless it is retained, since I
  # like to use such an API.
  # ========================================================================= #
  def reset_the_internal_hash
    @internal_hash = {}
  end

  # ========================================================================= #
  # === seconds_to_time_format
  # ========================================================================= #
  def seconds_to_time_format(i)
    return Roebe::TwentyfourHoursNotation[i] # => "00:05:00.000"
  end

  # ========================================================================= #
  # === does_the_video_player_support_this_commandline?
  #
  # The following method checks whether the chosen video player at hand
  # supports a particular commandline. The check is necessary because
  # on some systems 'mpv' may be an alias to 'mplayer.
  # ========================================================================= #
  def does_the_video_player_support_this_commandline?(
      this_commandline = '--msg-color'
    )
    cmd = ::MultimediaParadise.use_which_video_player?+
          ' '+this_commandline.strip+' 2>&1'
    result = `#{cmd}`
    !result.include?('Unknown option on the command line: ')
  end

  # ========================================================================= #
  # === rename_file
  # ========================================================================= #
  def rename_file(a, b)
    FileUtils.mv(a, b)
  end

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize
    reset
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    # ======================================================================= #
    # === @use_colours
    # ======================================================================= #
    @use_colours = true
    # ======================================================================= #
    # === @be_verbose
    # ======================================================================= #
    set_be_verbose
    # ======================================================================= #
    # === @debug
    #
    # By default, debugging is disabled.
    # ======================================================================= #
    @debug       = false  
  end

  # ========================================================================= #
  # === esystem
  # ========================================================================= #
  def esystem(
      i, optional_colour_to_use = nil
    )
    MultimediaParadise.esystem(
      i, optional_colour_to_use
    )
  end

  # ========================================================================= #
  # === load_yaml
  # ========================================================================= #
  def load_yaml(i)
    MultimediaParadise.load_yaml(i)
  end

  # ========================================================================= #
  # === use_which_multimedia_player?
  # ========================================================================= #
  def use_which_multimedia_player?
    ::MultimediaParadise.use_which_multimedia_player?
  end

  # ========================================================================= #
  # === mv
  # ========================================================================= #
  def mv(old, new)
    if File.exist? old
      ::MultimediaParadise.mv(old, new) # Delegate towards FileUtils in this cae.
    else
      e 'Can not move `'+sfile(old)+'` as no such file exists.'
    end
  end

  # ========================================================================= #
  # === e
  # ========================================================================= #
  def e(i = '')
    ::MultimediaParadise.e(i)
  end

  # ========================================================================= #
  # === efancy
  # ========================================================================= #
  def efancy(i)
    e sfancy(i)
  end

  # ========================================================================= #
  # === rds
  # ========================================================================= #
  def rds(i)
    i.squeeze '/'
  end

  # ========================================================================= #
  # === is_mp3?
  # ========================================================================= #
  def is_mp3?(i)
    File.extname(i).downcase == '.mp3'
  end

  # ========================================================================= #
  # === is_mp4?
  # ========================================================================= #
  def is_mp4?(i)
    File.extname(i).downcase == '.mp4'
  end

  # ========================================================================= #
  # === is_mkv?
  # ========================================================================= #
  def is_mkv?(i)
    File.extname(i).downcase == '.mkv'
  end

  # ========================================================================= #
    # === debug?
  # ========================================================================= #
  def debug?
    @debug
  end

  # ========================================================================= #
  # === be_verbose?
  # ========================================================================= #
  def be_verbose?
    @be_verbose
  end; alias verbose? be_verbose? # === verbose?

  # ========================================================================= #
  # === enable_debug
  # ========================================================================= #
  def enable_debug
    @debug = true
  end

  # ========================================================================= #
  # === be_silent
  # ========================================================================= #
  def be_silent
    @be_verbose = false
  end; alias be_quiet be_silent # == be_quiet

  # ========================================================================= #
  # === set_be_verbose
  # ========================================================================= #
  def set_be_verbose
    @be_verbose = true
  end

  # ========================================================================= #
  # === set_be_quiet
  # ========================================================================= #
  def set_be_quiet
    @be_verbose = false
  end

  # ========================================================================= #
  # === mkdir
  # ========================================================================= #
  def mkdir(i)
    ::MultimediaParadise.mkdir(i)
  end

  # ========================================================================= #
  # === cd                                                           (cd tag)
  # ========================================================================= #
  def cd(i)
    ::MultimediaParadise.chdir(i)
  end; alias chdir cd # === chdir

  # ========================================================================= #
  # === no_file_exists_at
  #
  # This method can be consistently used to notify the user that a
  # specific file is missing.
  # ========================================================================= #
  def no_file_exists_at(this_file)
    e "No file exists at `#{sfile(this_file)}`."
  end

  # ========================================================================= #
  # === no_file_exists
  # ========================================================================= #
  def no_file_exists
    e "No file exists - may be a nil value."
  end

  # ========================================================================= #
  # === opnn
  # ========================================================================= #
  def opnn(
      namespace = { namespace: @namespace },
      &block
    )
    # ======================================================================= #
    # Turn a String into a Hash next.
    # ======================================================================= #
    if namespace.is_a? String
      namespace = { namespace: namespace } # Use the input to the method in this case.
    end
    Opn.opn(namespace, &block)
  end

  # ========================================================================= #
  # === is_audio_file?
  #
  # This method can be used in order to determine whether we are working
  # with an audio file here or whether we are not.
  # ========================================================================= #
  def is_audio_file?(i)
    ::MultimediaParadise.is_audio_file?(i)
  end; alias is_an_audio_file? is_audio_file? # === is_an_audio_file?

  # ========================================================================= #
  # === cliner_with_time_stamp
  # ========================================================================= #
  def cliner_with_time_stamp
    _ = Cliner.cliner :return_time
    if @use_colours
      _.sub!(/(\d{2}:\d{2}:\d{2})/, slateblue("\\1"))
    end
    e _
  end

  # ========================================================================= #
  # === cliner                                                   (cliner tag)
  # ========================================================================= #
  def cliner(i = nil)
    Cliner.cliner(i)
  end

  # ========================================================================= #
  # === directory_to_realvids?
  #
  # This method will return the absolute path to my local realvid
  # directory.
  #
  # It is only used in code of classes that deal with my local multimedia
  # files really.
  # ========================================================================= #
  def directory_to_realvids?
    ::MultimediaParadise.realvids_directory?
  end; alias realvids_directory? directory_to_realvids? # === realvids_directory?

  # ========================================================================= #
  # === local_audio_directory?
  # ========================================================================= #
  def local_audio_directory?
    MultimediaParadise.local_audio_directory?
  end; alias directory_containing_the_local_songs? local_audio_directory? # === directory_containing_the_local_songs?

  # ========================================================================= #
  # === file_video_collection?
  # ========================================================================= #
  def file_video_collection?
    MultimediaParadise.file_video_collection?
  end

  # ========================================================================= #
  # === cartoons_directory?
  # ========================================================================= #
  def cartoons_directory?
    MultimediaParadise.cartoons_directory?
  end; alias cartoon_directory? cartoons_directory? # === cartoon_directory

  # ========================================================================= #
  # === map_symbol_to_locally_existing_file
  # ========================================================================= #
  def map_symbol_to_locally_existing_file(i)
    case i # case tag
    # ======================================================================= #
    # === wrap :beauty
    # ======================================================================= #
    when :beauty, ':beauty',
         :american_beauty, ':american_beauty'
      i = "#{realvids_directory?}017_American_Beauty.mp4"
    # ======================================================================= #
    # === wrap :halloween
    # ======================================================================= #
    when :halloween, ':halloween'
      i = "#{realvids_directory?}215_Halloween.mp4"
    # ======================================================================= #
    # === wrap :predator
    # ======================================================================= #
    when :predator,  ':predator'
      i = "#{realvids_directory?}245_Predator.avi"
    else
      e 'Unhandled symbol: '+i.to_s
    end
    i
  end

  # ========================================================================= #
  # === remove_file                                              (remove tag)
  #
  # Use this method if you need to delete/remove a file.
  # ========================================================================= #
  def remove_file(i)
    MultimediaParadise.remove_file(i)
  end; alias remove remove_file # === remove_file
       alias delete remove_file # === delete

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
  end

  require 'multimedia_paradise/misc/milliseconds_to_long_format_converter.rb'
  # ========================================================================= #
  # === to_hh_mm_ss
  # ========================================================================= #
  def to_hh_mm_ss(i)
    MultimediaParadise.milliseconds_to_long_format(i * 1000)
  end

  # ========================================================================= #
  # === MultimediaParadise::Base[]
  # ========================================================================= #
  def self.[](i = '')
    new(i)
  end

  # ========================================================================= #
  # === time_right_now
  # ========================================================================= #
  def time_right_now
    Time.now.strftime('%H:%M:%S-%d.%m.%Y')
  end

  require 'multimedia_paradise/toplevel_methods/is_on_roebe.rb'
  # ========================================================================= #
  # === is_on_roebe?
  #
  # Determine whether we are on a roebe-system or whether we are not.
  # ========================================================================= #
  def is_on_roebe?
    ::MultimediaParadise.is_on_roebe?
  end

  # ========================================================================= #
  # === return_pwd
  #
  # This method will always return a String with a trailing '/' character.
  # ========================================================================= #
  def return_pwd
    ("#{Dir.pwd}/").squeeze('/')
  end

  # ========================================================================= #
  # === dataset_from_file_video_collection
  # ========================================================================= #
  def dataset_from_file_video_collection(
      i = ::MultimediaParadise.file_video_collection
    )
    YAML.load_file(i)
  end

  # ========================================================================= #
  # === return_all_video_files
  #
  # This method will return a Hash. The entries may look like this:
  #
  #   228=> {"title"=>"Apocalypto"
  #
  # ========================================================================= #
  def return_all_video_files
    MultimediaParadise.return_all_video_files
  end

  # ========================================================================= #
  # === select_only_video_files_from
  # ========================================================================= #
  def select_only_video_files_from(array)
    array.select {|entry| is_video_file?(entry) }
  end

  # ========================================================================= #
  # === copy_file
  # ========================================================================= #
  def copy_file(from, to = return_pwd)
    FileUtils.cp(from, to)
  end

  # ========================================================================= #
  # === report_pwd
  # ========================================================================= #
  def report_pwd
    e ("#{Dir.pwd}/").squeeze '/'
  end

  # ========================================================================= #
  # === try_to_rename_kde_konsole_tab
  # ========================================================================= #
  def try_to_rename_kde_konsole_tab(
      new_title                         = '_',
      try_to_rename_the_kde_konsole_tab = TRY_TO_RENAME_THE_KDE_KONSOLE_TAB
    )
    MultimediaParadise.try_to_rename_kde_konsole_tab(
      new_title, try_to_rename_the_kde_konsole_tab
    )
  end; alias rename_kde_konsole try_to_rename_kde_konsole_tab # === rename_kde_konsole
       alias do_rename_konsole  try_to_rename_kde_konsole_tab # === do_rename_konsole

  require 'multimedia_paradise/toplevel_methods/is_video_file.rb'
  # ========================================================================= #
  # === is_video_file?
  # ========================================================================= #
  def is_video_file?(i)
    ::MultimediaParadise.is_video_file?(i) # Delegate to the module-method here.
  end; alias is_a_video_file?      is_video_file? # === is_a_video_file?
       alias is_this_a_video_file? is_video_file? # === is_this_a_video_file?

  # ========================================================================= #
  # === filter_for_video_files
  # ========================================================================= #
  def filter_for_video_files(i)
    ::MultimediaParadise.filter_for_video_files(i)
  end

  # ========================================================================= #
  # === verbose_truth
  # ========================================================================= #
  def verbose_truth(i)
    case i.to_s
    when 't','true'
      true
    when 'f','false',''
      false
    end
  end

  # ========================================================================= #
  # === return_random_video
  # ========================================================================= #
  def return_random_video
    unless ::MultimediaParadise.respond_to? :return_random_video
      require 'multimedia_paradise/toplevel_methods/return_random_video_file_from_the_video_collection.rb'
    end
    ::MultimediaParadise.return_random_video
  end

  # ========================================================================= #
  # === save_what_into
  #
  # A simple wrapper over SaveFile.
  # ========================================================================= #
  def save_what_into(what, into)
    if Object.const_defined? :SaveFile
      SaveFile.write_what_into(what, into)
    end
  end; alias save_file       save_what_into # === save_file
       alias write_what_into save_what_into # === write_what_into

  # ========================================================================= #
  # === append_what_into
  # ========================================================================= #
  def append_what_into(what, into)
    if Object.const_defined? :SaveFile
      SaveFile.append_what_into(what, into)
    end
  end

  # ========================================================================= #
  # === project_base_directory?
  # ========================================================================= #
  def project_base_directory?
    ::MultimediaParadise.project_base_directory?
  end

  # ========================================================================= #
  # === yaml_directory?
  # ========================================================================= #
  def yaml_directory?
    "#{project_base_directory?}yaml/"
  end

  require 'multimedia_paradise/toplevel_methods/output_directory.rb'
  # ========================================================================= #
  # === log_dir?
  # ========================================================================= #
  def log_dir?
    ::MultimediaParadise.log_dir?
  end; alias log_directory? log_dir? # === log_directory?

  # ========================================================================= #
  # === ensure_that_the_output_directory_exists
  # ========================================================================= #
  def ensure_that_the_output_directory_exists(
      this_dir = log_directory?
    )
    this_dir = File.dirname(this_dir) unless this_dir.end_with? '/'
    ::MultimediaParadise.ensure_that_the_output_directory_exists
  end

  # ========================================================================= #
  # === is_multimedia_file?
  #
  # This method will return true if the given directory is either an
  # audio file or a video file.
  # ========================================================================= #
  def is_multimedia_file?(i)
    is_audio_file?(i) or is_video_file?(i)
  end; alias is_a_multimedia_file?      is_multimedia_file? # === is_a_multimedia_file?
       alias is_this_a_multimedia_file? is_multimedia_file? # === is_this_a_multimedia_file?

  # ========================================================================= #
  # === beautiful_url
  # ========================================================================= #
  def beautiful_url(i)
    unless Object.const_defined? :BeautifulUrl
      begin
        require 'beautiful_url'
      rescue LoadError; end
    end
    i = ::BeautifulUrl.new(i) if Object.const_defined? :BeautifulUrl
    i = i.first if i.is_a? Array
    return i
  end

  # ========================================================================= #
  # === register_sigint
  #
  # If an argument is passed to this method then it will be displayed.
  # ========================================================================= #
  def register_sigint(
      use_this_as_exit_message = 'Exiting now as requested.',
      &block
    )
    clear_kde_konsole_tab_on_exit = false
    # ======================================================================= #
    # === Handle blocks given to this method next
    # ======================================================================= #
    if block_given?
      yielded = yield
      case yielded
      # ===================================================================== #
      # === :clear_kde_konsole_tab_on_exit_if_is_on_roebe
      # ===================================================================== #
      when :clear_kde_konsole_tab_on_exit_if_is_on_roebe,
           :pretty
        if is_on_roebe?
          clear_kde_konsole_tab_on_exit = true
        end
      end
    end
    Signal.trap('SIGINT') {
      e use_this_as_exit_message if use_this_as_exit_message
      if clear_kde_konsole_tab_on_exit
        begin
          require 'roebe/classes/kde/kde_konsole/kde_konsole.rb'
          e Roebe.konsole_tab_title('.') # Rename the tab again.
        rescue LoadError; end
      end
      exit
    }
  end

  # ========================================================================= #
  # === opne
  # ========================================================================= #
  def opne(i = '')
    opnn; e i
  end

end; end