#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# Code in this .rb file has to do with converting multimedia formats, and
# thus will contain all kind of conversion-related aspects - but it can
# also be used for cutting, splitting etc.. of audio files, at the least
# for some of that.
#
# If you want to use a wrapper over image-to-image manipulation then
# you should use another module/class though, perhaps even the project
# called image_paradise.
# =========================================================================== #
# require 'multimedia_paradise/toplevel_methods/conversions.rb'
# MultimediaParadise.video_to_images
# =========================================================================== #
module MultimediaParadise

  begin
    require 'opn'
  rescue LoadError; end

  begin
    require 'cliner'
  rescue LoadError; end

  require 'multimedia_paradise/constants/conversions.rb'
  require 'multimedia_paradise/toplevel_methods/e.rb'
  require 'multimedia_paradise/toplevel_methods/esystem.rb'
  require 'multimedia_paradise/toplevel_methods/merge_multimedia_file.rb'
  require 'multimedia_paradise/toplevel_methods/output_directory.rb'
  require 'multimedia_paradise/toplevel_methods/rds.rb'
  require 'multimedia_paradise/toplevel_methods/run_sys_command.rb'

  # ========================================================================= #
  # === @use_this_binary_for_conversions
  #
  # Specify which program we will use by default for binary-related
  # conversions.
  # ========================================================================= #
  @use_this_binary_for_conversions = :ffmpeg # Can be :sox too.

  # ========================================================================= #
  # === MultimediaParadise.use_this_binary_for_conversions?
  # ========================================================================= #
  def self.use_this_binary_for_conversions?
    @use_this_binary_for_conversions
  end

  # ========================================================================= #
  # === cliner
  # ========================================================================= #
  def cliner
    Cliner.cliner
  end

  # ========================================================================= #
  # === MultimediaParadise.general_format_conversion
  #
  # Use ffmpeg here in a simplified manner through this method, in order
  # to convert one format to another one.
  #
  # Simply provide the input file as the first argument to this method,
  # and the desired target file as the second argument. See the following
  # usage examples for how to use this method specifically.
  #
  # Usage examples:
  #
  #   MultimediaParadise.general_format_conversion('foo.wav','foo.mp3')
  #   MultimediaParadise.convert('foobar.mp3', to: 'wav')
  #
  # ========================================================================= #
  def self.general_format_conversion(
      input_file,  # Such as 'foobar.wav'.
      target_file, # Such as 'foobar.mp3'.
      &block
    )
    if input_file.is_a? Array
      input_file.each {|entry|
        general_format_conversion(entry, target_file, &block)
      }
    else
      if target_file.is_a? Hash
        # =================================================================== #
        # === :to
        #
        # Allow support for the ":to" keyword.
        # =================================================================== #
        if target_file.has_key? :to
          target_file = target_file.delete(:to)
        end
      end
      # ===================================================================== #
      # The variable target_file may be a symbol. Check this first.
      #
      # We work on the target_file first.
      # ===================================================================== #
      case target_file.to_s
      # ===================================================================== #
      # === to_webm
      # ===================================================================== #
      when /^to(_|-)?webm$/i,
           /webm$/i
        target_file = input_file.gsub(File.extname(input_file), '')+'.webm'
      # ===================================================================== #
      # === to_ogg
      # ===================================================================== #
      when /^to(_|-)?ogg$/i,
           /ogg$/i
        target_file = input_file.gsub(File.extname(input_file), '')+'.ogg'
      # ===================================================================== #
      # === to_aac
      # ===================================================================== #
      when /^to(_|-)?aac$/i,
           /\.?aac$/i
        target_file = input_file.gsub(File.extname(input_file), '')+'.aac'
      # ===================================================================== #
      # === to_dv
      # ===================================================================== #
      when /^to(_|-)?dv$/i,
           /\.?dv$/i
        target_file = input_file.gsub(File.extname(input_file), '')+'.dv'
      # ===================================================================== #
      # === to_wav
      # ===================================================================== #
      when /^to(_|-)?wav$/i,
           /wav$/i,
           '2wav'
        target_file = input_file.gsub(File.extname(input_file), '')+'.wav'
      # ===================================================================== #
      # === to_mp3
      # ===================================================================== #
      when /^to(_|-)?mp3$/i,
           /mp3$/i
        target_file = input_file.gsub(File.extname(input_file), '')+'.mp3'
      # ===================================================================== #
      # === to_mp3
      # ===================================================================== #
      when /^to(_|-)?mp3$/i,
           /mp3$/i
        target_file = input_file.gsub(File.extname(input_file), '')+'.mp3'
      # ===================================================================== #
      # === to_opus
      # ===================================================================== #
      when /^to(_|-)?opus$/i,
           /opus$/i
        target_file = input_file.gsub(File.extname(input_file), '')+'.opus'
      # ===================================================================== #
      # === to_wv
      # ===================================================================== #
      when /^to(_|-)?wv$/i,
           /wv$/i
        target_file = input_file.gsub(File.extname(input_file), '')+'.wv'
      # ===================================================================== #
      # === to_wma
      # ===================================================================== #
      when /^to(_|-)?wma$/i,
           /wma$/i
        target_file = input_file.gsub(File.extname(input_file), '')+'.wma'
      end unless target_file.to_s.include?('/')
      # ===================================================================== #
      #          ^^^ here we assume the second argument to be like this: /Depot/j/yodel.wav
      # ===================================================================== #
      # ===================================================================== #
      # The "target_file" variable may be something such as '.wma'
      # In this case, the file name is missing. We fix this problem
      # by using the real name of the input_file as output format.
      # ===================================================================== #
      if target_file.start_with? '.'
        if input_file.include? '.'
          target_file = input_file[0, input_file.index('.')]+target_file
        end
      end if target_file.respond_to? :start_with?
      target_file = rds(target_file)
      target_file = File.absolute_path(target_file)
      if File.exist? input_file
        cliner {
          please_be_patient(
            input_file, 'to '+File.extname(target_file)+' format'
          )
          ::MultimediaParadise.display_warning(:lossy)
        }
        # =================================================================== #
        # By default we will use ffmpeg, but this can be changed to sox or
        # any other toolkit.
        # =================================================================== #
        _ = "#{::MultimediaParadise.use_this_binary_for_conversions?} -y ".dup
        _ << " -i #{input_file} "
        # =================================================================== #
        # Do some ad-hoc clean up next.
        # -ac means audio_channel
        # =================================================================== #
        if target_file.end_with?('.opus')
          _ << '" -ar 48000 -ac 2 -acodec libopus -ab 256k '
        elsif target_file.end_with?('.ogg')
          _ << ' -acodec libvorbis -aq 60 '
        end
        if block_given?
          yielded = yield
          if yielded.is_a? Hash
            # =============================================================== #
            # === :bitrate_to_use
            # =============================================================== #
            if yielded.has_key? :bitrate_to_use
              # -b:a will convert the audio bitrate to be exactly that value.
              _ << ' -b:a '+yielded[:bitrate_to_use].to_s+'k'
            end
          end
        end
        _ << " #{target_file}"
        _.squeeze!(' ')
        run_sys_command _
        # =================================================================== #
        # Since as of August 2016, we will return the path to the newly
        # generated file.
        # =================================================================== #
        return target_file.to_s
      else
        file_does_not_exist(input_file)
      end
    end
  end; self.instance_eval { alias convert          general_format_conversion } # === MultimediaParadise.convert
       self.instance_eval { alias basic_conversion general_format_conversion } # === MultimediaParadise.basic_conversion

  # ========================================================================= #
  # === MultimediaParadise.mp4_to_mp3
  # ========================================================================= #
  def self.mp4_to_mp3(a, b = '.mp3')
    ::MultimediaParadise.general_format_conversion(a, b)
  end

  # ========================================================================= #
  # === MultimediaParadise.ogg_to_mp3
  #
  # Convert .ogg files into .mp3 files.
  # ========================================================================= #
  def self.ogg_to_mp3(this_file, to = '.mp3')
    ::MultimediaParadise.general_format_conversion(this_file, to)
  end

  # ========================================================================= #
  # === MultimediaParadise.mp3_to_opus
  # ========================================================================= #
  def self.mp3_to_opus(this_file, to = '.opus')
    ::MultimediaParadise.general_format_conversion(this_file, to)
  end

  # ========================================================================= #
  # === MultimediaParadise.wav_to_midi
  #
  # Convert wav to midi file using waon.
  # ========================================================================= #
  def self.wav_to_midi(
      input_file, output_file = 'ack.mid'
    )
    cmd = 'waon -i '.dup
    cmd << input_file
    cmd << ' -o '+output_file
    run_sys_command cmd
  end 

  # ========================================================================= #
  # === MultimediaParadise.flash_to_mpg
  # ========================================================================= #
  def self.flash_to_mpg(
      input_file  = 'jokes.flv',
      output_file = 'jokes.mpg'
    )
    if File.exist? input_file
      cmd = 'ffmpeg -i '
      cmd << input_file
      cmd << ' -ab 56 -ar 22050 -b 500 -s 320×240 '
      cmd << output_file
      run_sys_command cmd
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.m4a_to_mp3
  #
  # From .m4a audio to a .mp3 file.
  # ========================================================================= #
  def self.m4a_to_mp3(this_file, to = '.mp3')
    ::MultimediaParadise.general_format_conversion(this_file, to)
  end

  # ========================================================================= #
  # === MultimediaParadise.any_to_mp3
  #
  # Wrapper method. Converts any file to mp3 format. Right now it uses 
  # ffmpeg for this. This may be superior to the other variants actually.
  # ========================================================================= #
  def self.any_to_mp3(this_file, to = '.mp3')
    ::MultimediaParadise.general_format_conversion(this_file, to)
  end

  # ========================================================================= #
  # === MultimediaParadise.decrease_volume
  #
  # This should decrease audio volume, by using sox.
  #
  # We do this by using the command --volume FACTOR.
  # ========================================================================= #
  def self.decrease_volume(
      of_this_file,
      by_how_much = '0.5'
    )
    run_sys_command 'sox --volume '+
                     by_how_much+' '+of_this_file+' AUDIO_REDUCED_'+of_this_file 
  end

  # ========================================================================= #
  # === MultimediaParadise.wav_to_opus
  #
  # Use this method to convert a .mp4 file into a .mp3 file.
  #
  # We will use ffmpeg for this conversion.
  # ========================================================================= #
  def self.wav_to_opus(this_file, to = '.opus')
    ::MultimediaParadise.general_format_conversion(this_file, to)
  end
  
  # ========================================================================= #
  # === MultimediaParadise.wav_to_mp3                 (wav_to_mp3 conversion)
  #
  # We convert a .wav file into a .mp3 file with this method here.
  #
  # The second argument can be a hash such as:
  #
  #   :output => output?
  #
  # Usage example:
  #
  #   MultimediaParadise.wav_to_mp3
  #
  # Another way would be to use ffmpeg like so (-ab is the audio bitrate):
  #
  #   ffmpeg -i audio.wav -acodec mp3 -ab 192k audio.mp3
  #
  # Usage example:
  #
  #   result = MultimediaParadise.wav_to_mp3 '/Depot/j/BOTY_2013_MorningOfOwlShowcase.mp3'
  #   MultimediaParadise.wav_to_mp3 '/Depot/jj/Youtube2008_Soundmix.wav'
  #
  # ========================================================================= #
  def self.wav_to_mp3(
      this_file, bitrate_to_use = 192, to = '.mp3'
    )
    ::MultimediaParadise.general_format_conversion(this_file, to) {{
      bitrate_to_use: bitrate_to_use
    }}
  end

  # ========================================================================= #
  # === MultimediaParadise.normalize
  #
  # Normalize audio.
  # ========================================================================= #
  def self.normalize(
      this_file
    )
    cmd  = 'normalize '
    cmd << this_file
    run_sys_command cmd
  end

  # ========================================================================= #
  # === MultimediaParadise.mp3_to_midi
  #
  # Convert mp3 to midi file using lame and waon. We need to do two
  # steps because a single step is not yet enough.
  #
  # I do not know presently (December 2021) whether ffmpeg allows 
  # mp3-to-midi conversions, but waon does, so let's use that.
  # ========================================================================= #
  def self.mp3_to_midi(
      input_file, output_file = 'ack.mid'
    )
    result = mp3_to_wav(input_file)
    cmd = 'waon -i '.dup
    cmd << result
    cmd << ' -o '+output_file
    run_sys_command cmd
  end

  # ========================================================================= #
  # === MultimediaParadise.flash_to_avi
  #
  # Convert a .flv file to .avi.
  # ========================================================================= #
  def flash_to_avi(file)
    run_sys_command "ffmpeg -i #{file} #{file}.avi"
  end

  # ========================================================================= #
  # === MultimediaParadise.flash_to_mp4
  #
  # Do flash .flv to .mp4 conversion.
  # ========================================================================= #
  def self.flash_to_mp4(input_file = 'input.flv')
    if File.exist? input_file
      cmd  = 'ffmpeg '
      cmd << '-i '+input_file+' '
      cmd << '-an output.mp4  '
      run_sys_command cmd
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.images_to_video
  #
  # Notice that %d is replaced by the image number in question.
  #
  # Usage example:
  #
  #   MultimediaParadise.images_to_video
  #
  # ========================================================================= #
  def self.images_to_video(img = 'img%d.jpg')
    frames_per_second = '5'
    img = img.join(' ') if img.is_a? Array
    cmd = 'ffmpeg -r '+frames_per_second+' -b '+DEFAULT_VIDEO_BITRATE.to_s+' -i '+img+' output.mp4'
    run_sys_command cmd
  end

  # ========================================================================= #
  # === MultimediaParadise.avi_to_m4v
  # ========================================================================= #
  def self.avi_to_m4v(
      output_file = 'output.m4v', *input_files
    )
    output_file = 'output.m4v' if output_file.nil?
    output_file = 'output.'+output_file unless output_file.include? '.'
    _  = 'ffmpeg '.dup
    _ << '-i '+input_files.join(' ')+' '
    _ << '-r 12.5 '
    _ << '-b 80 '
    _ << '-vcodec libxvid '
    _ << '-maxrate 80 '
    _ << '-bufsize 5 '
    _ << output_file
    run_sys_command _
  end

  # ========================================================================= #
  # === MultimediaParadise.to_vob
  #
  # This method converts input to a .vob file, which is a file that is
  # used for DVD-playback.
  #
  # Usage example:
  #
  #   require 'multimedia_paradise/autoinclude'
  #
  #   to_vob 'Simpsons_Halloween_DasHaus.avi'
  #
  # ========================================================================= #
  def self.to_vob(*i)
    _ = 'ffmpeg -y -i '+i.join(' ')+' MERGED.vob'
    e _
    run_sys_command _
  end

  # ========================================================================= #
  # === MultimediaParadise.demultiplex
  #
  # Demultiplex using tcdemux.
  #
  # Ffmpeg should also allow it - this is an example for it:
  #
  #   ffmpeg -i source -map 0:a -c:a copy audioout -map 0:v -c:v copy videoout
  #
  # Usage example:
  #
  #   MultimediaParadise.demultiplex('foo.mpg')
  #
  # ========================================================================= #
  def self.demultiplex(this_file)
    cmd  = 'tcdemux -i '.dup
    cmd << this_file
    cmd << ' -x ac3 > bla.ac3'
    run_sys_command cmd
  end

  # ========================================================================= #
  # === MultimediaParadise.stretch_audio
  #
  # More than 1.0 means longer duration, less then 1.0 will shorten the song.
  # ========================================================================= #
  def self.stretch_audio(
      this_audio_file,
      by_this_factor = 1.2
    )
    if this_audio_file.is_a? Array
      this_audio_file.each {|entry| stretch_audio(entry) }
    else
      cmd = 'sox '+this_audio_file+' '+this_audio_file+' stretch '+by_this_factor.to_s
      run_sys_command cmd
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.trim_audio
  # ========================================================================= #
  def self.trim_audio(of_this_file)
    e 'Start at how many seconds:'
    start_pos = $stdin.gets.chomp
    e 'Capture how many seconds?'
    duration = $stdin.gets.chomp
    cmd = 'sox '+of_this_file+' trim '+start_pos+' '+duration
    e cmd
  end

  # ========================================================================= #
  # === MultimediaParadise.avi_to_flash
  #
  # Convert an .avi file to flash format, i.e. to .flv.
  # Right now we will simply use ffmpeg for this.
  # ========================================================================= #
  def self.avi_to_flash(file)
    run_sys_command 'ffmpeg -i '+file+' '+file+'.flv'
  end

  # ========================================================================= #
  # === MultimediaParadise.au_to_wav
  #
  # convert .au file to .wav file.
  # ========================================================================= #
  def self.au_to_wav(
      this_file, to = 'wav'
    )
    ::MultimediaParadise.general_format_conversion(this_file, to: ot)
  end

  # ========================================================================= #
  # === MultimediaParadise.video_to_images
  #
  # This method can, by leveraging "ffmpeg", convert a video-file into
  # a set of image-files.
  #
  # The first argument to this method should be the filename of the
  # video-file, such as "foobar.mp4" or "cool_movie.mkv".
  #
  # Right now the default target format for the images will be .jpg,
  # but you could also use .png or any other image-format. This is
  # the second argument to this method.
  #
  # If you need a variant that works for bash (the shell), try
  # this:
  #
  #   ffmpeg -i ack.mkv extracted_images_from_file_ack_%d.jpg
  #
  # Or via the bin/video_to_images file:
  #
  #   video_to_images ack.mkv
  #   videotoimages *.avi
  #
  # ========================================================================= #
  def self.video_to_images(
      which_video      = 'video.mpg',
      format_to_use    = :default,
      framerate_to_use = 25
    )
    if format_to_use.frozen?
      format_to_use = format_to_use.dup
    end
    case format_to_use
    # ======================================================================= #
    # === :default
    # ======================================================================= #
    when :default
      format_to_use = '.jpg'
    end
    unless format_to_use.start_with? '.'
      format_to_use.prepend '.'
    end
    [which_video].flatten.compact.each {|this_video_file|
      cliner
      cmd = 'ffmpeg '+
            '-i '+this_video_file.to_s+' '+
            '-r '+framerate_to_use.to_s+
            ' extracted_images_from_the_movie_%d'+format_to_use
      run_sys_command cmd
      cliner
    }
  end; self.instance_eval { alias to_images video_to_images } # === MultimediaParadise.to_images

  # ========================================================================= #
  # === MultimediaParadise.audio_info
  #
  # As of November 2013 this method makes use of ffprobe, from ffmpeg
  # instead.
  # ========================================================================= #
  def self.audio_info(i)
    if i.is_a? Array
      i.each {|entry| audio_info(entry) }
    else
      # cmd = 'sox '+i+' -e stat'
      cmd = 'ffprobe '+i
      run_sys_command cmd
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.file_does_not_exist
  #
  # Use this method when reporting that a file does not exist (or was not
  # found).
  # ========================================================================= #
  def self.file_does_not_exist(file)
    e "No file exists at `#{sfile(file)}`."
  end; self.instance_eval { alias no_file_exists_at file_does_not_exist } # === MultimediaParadise.no_file_exists_at

  # ========================================================================= #
  # === MultimediaParadise.for_ipod
  #
  # Convert to ipod.
  # ========================================================================= #
  def self.for_ipod
    cmd = 'mencoder -ovc raw -ofps 15 -oac pcm '.dup
    cmd << '-vf scale=176:-2,expand=176:132,format=bgr16 input.file '
    cmd << '-o output.avi'
  end

  # ========================================================================= #
  # === MultimediaParadise.wav_to_ogg
  #
  # Perform a .wav to .ogg conversion through this method here.
  #
  # Usage examples:
  #
  #   wav_to_ogg 'test.wav'
  #   MultimediaParadise.wav_to_ogg('/Depot/j/foobar.wav','/Depot/j/yodel.ogg')
  #
  # ========================================================================= #
  def self.wav_to_ogg(
      i  = '/Depot/j/foobar.wav',
      to = 'ogg'
    )
    MultimediaParadise.convert(i, to: to)
  end; self.instance_eval { alias wav2ogg wav_to_ogg } # === MultimediaParadise.wav2ogg

  # ========================================================================= #
  # === MultimediaParadise.mp3_to_wav
  #
  # This method will attempt to convert a .mp3 file to a .wav file.
  #
  # Returns result as well.
  #
  # Rather than use ffmpeg we could also use lame, such as via:
  #
  #   'lame --decode "'+i+'" "'+output_file+'"'
  #
  # Usage example:
  #
  #   MultimediaParadise.mp3_to_wav('/Depot/j/foobar.mp3','/Depot/j/yodel.wav')
  #
  # ========================================================================= #
  def self.mp3_to_wav(
      i  = '/Depot/j/foobar.mp3',
      to = 'wav'
    )
    MultimediaParadise.convert(i, to: to)
  end; self.instance_eval { alias mp3towav mp3_to_wav } # === MultimediaParadise.mp3towav
       self.instance_eval { alias to_wav   mp3_to_wav } # === MultimediaParadise.to_wav

  # ========================================================================= #
  # === MultimediaParadise.please_be_patient
  #
  # Simply display a little message so the user can remain patient.
  #
  # The second argument is an optional target-format message.
  #
  # Usage example:
  #
  #   please_be_patient(file, 'to mp3 format')
  #
  # ========================================================================= #
  def self.please_be_patient(
      file, optional_target_format = ''
    )
    optional_target_format = optional_target_format.to_s
    optional_target_format << ' ' if optional_target_format.empty?
    opn; e 'Converting `'+::MultimediaParadise.sfile(file)+'` '+
            optional_target_format.strip+
           ' next, please be patient.'
  end

  # ========================================================================= #
  # === MultimediaParadise.extract_audio_from_until
  #
  # Usage example:
  #
  #  MultimediaParadise.extract_audio_from_until('/Depot/j/foobar.mp4', 25, 35)
  #
  # ========================================================================= #
  def self.extract_audio_from_until(
      this_file,
      from_start_pos = 30, 
      duration       = 30
    )
    this_file = File.absolute_path(this_file)
    base_dir = File.dirname(this_file)
    output_file = 'CLIPPED_'+File.basename(this_file)
    output_file = base_dir+'/'+output_file
    this_command = 'ffmpeg -y'+ # -y for "automatically overwrite".
                   ' -i '+this_file+' -ss '+from_start_pos.to_s+
                   ' -t '+duration.to_s+' '+output_file
    run_sys_command this_command
  end

  # ========================================================================= #
  # === MultimediaParadise.increase_volume
  #
  # Increase the volume of an audio file using sox with this method.
  #
  # We do this by using the command --volume FACTOR.
  #
  # A value of 0.9 would mean a new audio file with 90% volume of the
  # original file. 2.0 would mean 200% of the volume.
  #
  # Note that this method depends on sox, which in turn has limitations.
  # For instance, sox can not handle .mp4 files. In the long run I may
  # have to switch to ffmpeg, but for now this method will stay as
  # it is. (In the future perhaps an additional argument has to
  # denote which toolkit we will use.)
  #
  # Usage examples:
  #
  #   MultimediaParadise.increase_volume 'Mittwoch_14.10.2015.mp3'
  #   MultimediaParadise.increase_volume '/Depot/j/CLIPPED_foobar.mp4', '1.5'
  #   MultimediaParadise.increase_volume '/Depot/j/ack.mp3', '1.5'
  #
  # ========================================================================= #
  def self.increase_volume(
      of_this_file,
      by_how_much = '2.0'
    )
    of_this_file = File.absolute_path(of_this_file)
    unless by_how_much.is_a? String
      by_how_much = by_how_much.to_f.to_s
    end
    extname = File.extname(of_this_file) # This will result in e. g. ".mp4".
    run_sys_command 'sox --volume '+
                     by_how_much+' '+
                     of_this_file+' '+
                     of_this_file.delete_suffix(extname)+
                     'AUDIO_GAINED_'+extname 
  end

  # ========================================================================= #
  # === MultimediaParadise.capture_screen
  #
  # This captures a screen by using ffmpeg.
  # ========================================================================= #
  def self.capture_screen(size_angabe = '1024x800')
    begin
      require 'capture_screen'
    rescue LoadError; end
    img_file = '$MY_TEMP/GRABBED_X11_screencast.mov'
    _ = CaptureScreen.new :dont_run_yet
    _.store_here = img_file
    _.size = size_angabe
    return _.main_string?
  end

  # ========================================================================= #
  # === MultimediaParadise.remove_noise
  #
  # Use sox to remove noise stuff.
  #
  # Usage example:
  #
  #   require 'multimedia_paradise'; MultimediaConversions.remove_noise('*.mp3')
  #
  # ========================================================================= #
  def self.remove_noise(
      of_this_file = '*.mp3', this_rate = '3400'
    )
    of_this_file = glob(of_this_file) if of_this_file.include? '*'
    if of_this_file.is_a? Array
      of_this_file.each {|entry|
        remove_noise(entry, this_rate)
      }
    else
      esystem 'sox '+of_this_file+' removed_noise_'+of_this_file+' lowp '+this_rate
    end 
  end

  # ========================================================================= #
  # === MultimediaParadise.extract_video
  #
  # Extract the video portion using mencoder.
  #
  # The return value will be the new filepath in question, or nil otherwise.
  # ========================================================================= #
  def self.extract_video(this_file)
    new_file_location = nil
    if this_file.is_a? Array
      this_file = this_file.first # For now, we only use the first member.
    end
    name_of_the_extension = File.extname(this_file.delete('.'))
    output_file_name = 'video_extracted_from_'+
                       File.basename(this_file.sub(/\.avi$/,'')).
                       sub(/#{name_of_the_extension}$/,'')+'.avi'
    output_file_name = output_file_name.squeeze('.')
    _ = 'mencoder -idx -ovc copy -nosound '+this_file+' -o '+output_file_name
    run_sys_command _ if _
    new_file_location = output_file_name
    return new_file_location
  end

  # ========================================================================= #
  # === MultimediaParadise.display_warning
  #
  # Use this method when you wish to display a warning.
  #
  # Usage example:
  #
  #   display_warning :lossy
  #
  # ========================================================================= #
  def self.display_warning(what_kind_of_warning)
    case what_kind_of_warning
    when :lossy
      e 'Warning: Converting from one lossy format to another '\
        'lossy format.'
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.mp3_to_aac
  #
  # Convert from .mp3 to .aac format, using ffmpeg.
  #
  # Usage example:
  #
  #   MultimediaParadise.to_aac('/Depot/jjj/foobar.mp3')
  #
  # ========================================================================= #
  def self.mp3_to_aac(
      this_file,
      to = '.aac'
    )
    ::MultimediaParadise.general_format_conversion(this_file, to)
  end; self.instance_eval { alias to_aac mp3_to_aac } # === MultimediaParadise.to_aac

  # ========================================================================= #
  # === MultimediaParadise.mp3_to_ogg
  #
  # Convert from .mp3 to .ogg format.
  #
  # Usage example:
  #
  #   mp3_to_ogg 'foo.mp3'
  #   MultimediaParadise.mp3_to_ogg('/Depot/j/output.mp3')
  #   MultimediaParadise.to_ogg('/Depot/jjj/foobar.mp3')
  #
  # ========================================================================= #
  def self.mp3_to_ogg(
      this_file,
      to = '.ogg'
    )
    ::MultimediaParadise.general_format_conversion(this_file, to)
  end; self.instance_eval { alias mp3toogg mp3_to_ogg } # === MultimediaParadise.mp3toogg
       self.instance_eval { alias mp32ogg  mp3_to_ogg } # === MultimediaParadise.mp32ogg
       self.instance_eval { alias to_ogg   mp3_to_ogg } # === MultimediaParadise.to_ogg

  # ========================================================================= #
  # === MultimediaParadise.webm_to_mp3
  #
  # Usage example:
  #
  #   MultimediaParadise.webm_to_mp3 'Intermission_Six_Days-Jg9P3Z7ZCcE.f171.webm', '/Depot/j/yo.mp3'
  #
  # ========================================================================= #
  def self.webm_to_mp3(a, b = '.mp3')
    ::MultimediaParadise.general_format_conversion(a, b)
  end

  # ========================================================================= #
  # === MultimediaParadise.to_webm
  #
  # Usage example:
  #
  #   MultimediaParadise.to_webm 'foobar.mp4'
  #
  # ========================================================================= #
  def self.to_webm(a, b = '.webm')
    ::MultimediaParadise.general_format_conversion(a, b)
  end

  # ========================================================================= #
  # === MultimediaParadise.to_mkv
  #
  # Usage example:
  #
  #   MultimediaParadise.to_mkv 'foobar.mp4'
  #
  # ========================================================================= #
  def self.to_mkv(a, b = '.mkv')
    ::MultimediaParadise.general_format_conversion(a, b)
  end

  # ========================================================================= #
  # === MultimediaParadise.to_dv
  #
  # This entry point will convert into "Digital Video" format.
  #
  # Usage example:
  #
  #   MultimediaParadise.to_dv 'foobar.mp4'
  #
  # ========================================================================= #
  def self.to_dv(a, b = '.dv')
    ::MultimediaParadise.general_format_conversion(a, b)
  end

end

if __FILE__ == $PROGRAM_NAME
  # MultimediaParadise.extract_audio_from_until('/Depot/j/foobar.mp4')
  # MultimediaParadise.increase_volume '/Depot/j/CLIPPED_foobar.mp4', '3'
  # MultimediaParadise.increase_volume '/Depot/j/ack.mp3', '1.5'
  # MultimediaParadise.convert('/Depot/j/foobar.mp3', to: 'wav')
  # MultimediaParadise.mp3_to_wav('/Depot/j/foobar.mp3','/Depot/j/yodel.wav')
  # MultimediaParadise.webm_to_mp3 'Intermission_Six_Days-Jg9P3Z7ZCcE.f171.webm', '/Depot/j/yo.mp3'
  # MultimediaParadise.wav_to_ogg('/Depot/j/foobar.wav','/Depot/j/yodel.ogg')
  MultimediaParadise.ogg_to_mp3('/Depot/j/yodel.ogg', '/Depot/j/this_came_from_ogg.mp3')
end