#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'multimedia_paradise/ffmpeg/ffmpeg.rb'
# =========================================================================== #
module MultimediaParadise

  require 'multimedia_paradise/toplevel_methods/e.rb'
  require 'multimedia_paradise/toplevel_methods/esystem.rb'
  require 'multimedia_paradise/toplevel_methods/return_screen_resolution.rb'

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_to_mp4
  #
  # This method can be used to convert any video to a .mp4 format, by
  # making use of ffmpeg.
  #
  # The input to this method should be the files that you wish to
  # convert into the .mp4 format.
  # ========================================================================= #
  def self.ffmpeg_to_mp4(*these_files)
    output_file = 'foobar.mp4'
    _ = 'ffmpeg -i '+[these_files].flatten.compact.join(' ')+
        ' -vcodec copy -acodec copy '+
        output_file
    esystem _
    return output_file
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_return_image_from_this_video_file
  #
  # Invocation example:
  #
  #   MultimediaParadise.return_image_from_this_video_file()
  #
  # ========================================================================= #
  def self.ffmpeg_return_image_from_this_video_file(
      input_file = 'input.mp4',
      starting_position = 5, # in n seconds
      width_and_height_for_the_image = '480x270'
    )
    esystem 'ffmpeg -i '+input_file+' -vframes 1 -an -s '+
            width_and_height_for_the_image.to_s+' '+
            '-y -ss '+
            starting_position.to_s+' output.jpg'
  end; self.instance_eval {
    alias return_image_from_this_video_file ffmpeg_return_image_from_this_video_file # === MultimediaParadise.return_image_from_this_video_file
  }

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_chroma_subsampling
  #
  # Chroma subsampling is the practice of encoding images by implementing
  # less resolution for chroma information (i.e. color information) than
  # for luma information (brightness information), taking advantage of
  # the human visual system's lower acuity for color differences than
  # for luminance.
  #
  # Note that yuv420p denotes 4:2:0 chroma subsampling.
  # ========================================================================= #
  def self.ffmpeg_chroma_subsampling(this_video)
    esystem 'ffmpeg -i '+this_video+' -vf format=yuv420p output.mp4'
  end; self.instance_eval { alias chroma_subsampling ffmpeg_chroma_subsampling } # === MultimediaParadise.chroma_subsampling

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_deinterlace_this_video
  #
  # Deinterlacing is done generally on older videos or videos that had
  # been used for broadcasting purposes to minimize the interlacing
  # effect.
  #
  # Yadif ("yet another deinterlacing filter") is the name of the
  # filter used by ffmpeg. 
  # ========================================================================= #
  def self.ffmpeg_deinterlace_this_video(this_video)
    esystem 'ffmpeg -i '+this_video+' -vf yadif output.mp4'
  end; self.instance_eval { alias deinterlace_this_video ffmpeg_deinterlace_this_video } # === MultimediaParadise.deinterlace_this_video

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_scale_this_video
  #
  # This method can be used to scale (resize) a video.
  #
  # The first argument should be the path to a local video file. If an
  # Array is passed then this method will attempt to batch-process
  # that given input (and assume that this input is an array of
  # locally existing files).
  #
  # Invocation examples:
  #
  #   MultimediaParadise.ffmpeg_scale_this_video('input.mp4', height: 320)
  #   MultimediaParadise.ffmpeg_scale_this_video('input.mp4', height: 320, height: -1)
  #   MultimediaParadise.resize('yo.mp4', height: 320)
  #
  # ========================================================================= #
  def self.ffmpeg_scale_this_video(
      this_video, optional_hash = {}
    )
    if this_video.is_a? Array
      this_video.each {|entry| ffmpeg_scale_this_video(entry) }
    else
      use_this_as_output_file = 'output.mp4'
      cmd = ''.dup
      cmd << 'ffmpeg -i'
      cmd << " #{this_video}"
      # ===================================================================== #
      # Height and width must be determined before the next step. Set them to
      # nil first.
      # ===================================================================== #
      width = nil
      height = nil
      if optional_hash.is_a? Hash
        # ======================================================================= #
        # === :height
        # ======================================================================= #
        if optional_hash.has_key? :height
          height = optional_hash.delete(:height)
        end
        # ======================================================================= #
        # === :width
        # ======================================================================= #
        if optional_hash.has_key? :width
          width = optional_hash.delete(:width)
        end
      end
      if width.nil?
        width = 1080 # Currently hardcoded on purpose. Used to be 640 in the past.
      end
      if height.nil?
        height = -1
      end
      cmd << " -vf scale=#{width}:#{height}"
      if File.exist? use_this_as_output_file
        random_number = rand(100)+1 # Pick a "random" name for the file.
        use_this_as_output_file = "output#{random_number}.mp4"
      end
      # ===================================================================== #
      # Add the name of the output file next:
      # ===================================================================== #
      cmd << " #{use_this_as_output_file}"
      e
      esystem cmd
      e
    end
  end; self.instance_eval { alias ffmpeg_resize_this_video ffmpeg_scale_this_video } # === MultimediaParadise.ffmpeg_resize_this_video
       self.instance_eval { alias resize                   ffmpeg_scale_this_video } # === MultimediaParadise.resize

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_query_the_audio_codec_of_this_file
  #
  # This method will return a String - the audio codec in use.
  #
  # We will use ffmpeg for this and scan for an entry such as:
  #
  #   "    Stream #0:1(eng): Audio: aac (LC), 48000 Hz, stereo, fltp (default)"
  #
  # Many .mp4 files typically use the "aac" codec. .avi files on the
  # other hand, are often using the mp3 codec for their audio data.
  # ========================================================================= #
  def self.ffmpeg_query_the_audio_codec_of_this_file(i)
    # ======================================================================= #
    # See: https://rubular.com/r/8KrR1NTUq9hQU8
    # ======================================================================= #
    use_this_regex = /Audio: ([a-zA-Z0-9]{0,10})/
    result = []
    i = [i].flatten.map
    i.each {|this_file|
      _ = `ffmpeg -i #{this_file} 2>&1`.split("\n")
      _.select! {|line|
        line.include? ': Audio: '
      }
      _.map! {|entry|
        entry =~ use_this_regex
        $1.to_s.dup
      }
      result << _.first
    }
    result
  end; self.instance_eval { alias ffmpeg_audio_codec?          ffmpeg_query_the_audio_codec_of_this_file } # === MultimediaParadise.ffmpeg_audio_codec?
       self.instance_eval { alias ffmpeg_query_the_audio_codec ffmpeg_query_the_audio_codec_of_this_file } # === MultimediaParadise.ffmpeg_query_the_audio_codec

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_vertical_flip
  #
  # This method will make use of the video-filter called "vflip". A
  # vertical flip will flip the video from head-to-bottom, so people
  # will suddenly appear as if they are hanging from the ceiling. :)
  #
  # Invocation example:
  #
  #   MultimediaParadise.ffmpeg_vertical_flip '/Depot/Video/Reallife/O.J.Simpson_TV_News_etwa_1995.avi'
  #
  # ========================================================================= #
  def self.ffmpeg_vertical_flip(this_video_file)
    if this_video_file.is_a? Array
      this_video_file.each {|entry| ffmpeg_vertical_flip(entry) }
    else
      extname = File.extname(this_video_file)
      base_dir = File.dirname(this_video_file)
      filename = File.basename(this_video_file)
      output_name = base_dir+'/vertical_flip_'+filename.sub(/#{extname}/,'')+extname
      _ = 'ffmpeg -i '+this_video_file+' -vf "vflip" '+output_name
      esystem _
      output_name # <- Also return the output name here.
    end
  end; self.instance_eval { alias ffmpeg_flip_upside_down ffmpeg_vertical_flip } # === MultimediaParadise.flip_upside_down

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_horizontal_flip
  #
  # This method will make use of the video-filter called "hflip".
  # ========================================================================= #
  def self.ffmpeg_horizontal_flip(this_video_file)
    if this_video_file.is_a? Array
      this_video_file.each {|entry| horizontal_flip(entry) }
    else
      extname = File.extname(this_video_file)
      output_name = 'horizontal_flip_'+this_video_file.sub(/#{extname}/,'')+extname
      _ = 'ffmpeg -i '+this_video_file+' -vf "hflip" '+output_name
      esystem _
      output_name # <- Also return the output filename here.
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_rotate_this_video_file_by_90_degrees
  #
  # This method can be used to rotate a video file by 90°. Sometimes
  # smartphones create tilted videos, so this method can correct
  # this distortion.
  #
  # This can be fixed by using FFMPEG's transpose filter. The
  # transpose value will work in multiples of 90, so "1"
  # means 90°.
  #
  # Invocation example:
  #
  #   MultimediaParadise.ffmpeg_rotate_this_video_file_by_90_degrees('foo.mpg')
  #
  # ========================================================================= #
  def self.ffmpeg_rotate_this_video_file_by_90_degrees(
      this_video_file
    )
    output_file = 'output_'+this_video_file+'.mp4' 
    _ = 'ffmpeg -i '+this_video_file+' -vf transpose=1 '+output_file
    esystem _
    output_file # <- And return the output file too.
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_increase_volume_of_this_audio_file
  # ========================================================================= #
  def self.ffmpeg_increase_volume_of_this_audio_file(
      input_file,
      modify_volume_by_this_value = '5.0' # or 3.0 or 4.0 and so forth.
    )
    [input_file].flatten.compact.each {|this_file|
      _ = 'ffmpeg -i '+this_file+' -vcodec copy -af "volume='+modify_volume_by_this_value.to_s+
         '" output_'+this_file
      esystem _
    }
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_start_screencast
  #
  # This method can be used to "record" the desktop, by making use of
  # x11grab and the huffyuv codec.
  #
  # If more flexibility is required then this may be added at a later time.
  #
  # Keep in mind that this will create HUGE files depending on the video
  # codec in use - like 20 seconds may occupy 2 gigabytes for the
  # lossless huffyuv codec.
  #
  # If you wish to use different video or audio codecs, you can use
  # these commandline flags:
  #
  #   -vcodec: video encoding codec
  #   -acodec: audio encoding codec
  #
  # To record the desktop only, one could use a commandline flag such as
  # this one here:
  #
  #   ffmpeg -f x11grab -s 1366x768 -i :0.0 -r 25 -vcodec libx264 output.mkv
  #   ffmpeg -f x11grab -r 25 -s 1920x1080 -i :0.0 -vcodec huffyuv screencast.avi
  #
  # For some useful links (should these persist), have a look at:
  #
  #  https://www.pcsuggest.com/linux-screen-recorder-ffmpeg/
  #
  # ========================================================================= #
  def self.ffmpeg_start_screencast(
      use_this_screen_resolution = return_screen_resolution, # <- We will use the default screen-resolution here.
      output_filename            = 'screencast',
      use_this_framerate         = '25',
      use_this_video_codec       = 'huffyuv' # <- or libx264, for instance.
    )
    # ======================================================================= #
    # Must use a proper output-format. By default this will be .mkv,
    # although it could also be .avi.
    # ======================================================================= #
    unless output_filename.include? '.'
      output_filename = output_filename.dup if output_filename.frozen?
      output_filename << '.mkv'
    end
    _ = 'ffmpeg -f x11grab -r '+
        use_this_framerate.to_s+' -s '
    _ << use_this_screen_resolution+' '
    _ << ' -i :0.0 '
    _ << '-vcodec '+use_this_video_codec+' '
    #_ << '-f alsa -i hw:0 '
    _ << ' '+output_filename
    esystem _
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_chop_into_segments_of_n_seconds_size
  #
  # Use this (audio) method to chop up an audio file into segments
  # of n size. You could also use this on a video file, but the
  # primary use of it is to chop up a large audio file into
  # smaller audio files.
  #
  # The first argument is the name (path) of the file that you wish
  # to modify.
  #
  # If only one argument is provided (the filename), then we will default
  # to a duration of 60 seconds.
  # ========================================================================= #
  def self.ffmpeg_chop_into_segments_of_n_seconds_size(
      this_file, # <- Your target file that will be chopped up.
      n_seconds = 60
    )
    if this_file
      this_file = this_file.dup if this_file.frozen?
      extname = File.extname(this_file)
      filename = this_file.sub(/#{extname}/,'')
      _ = 'ffmpeg -i '+this_file+' -f segment -segment_time '+n_seconds.to_s+' -c copy out_'+filename+'_%03d'+extname
      esystem(_)
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_denoise
  #
  # This method attempts to remove noise from a multimedia file.
  #
  # Note that this uses ffmpeg. Sox can also be used to filter
  # away noise. The first argument is for the lowpass value;
  # the second argument is for the highpass value.
  #
  # The FFMPEG highpass filter is explained here:
  #
  #   https://ffmpeg.org/ffmpeg-filters.html#highpass
  #
  # ========================================================================= #
  def self.ffmpeg_denoise(
      input_file,
      lowpass_value  = 3000,
      highpass_value =  200
    )
    if input_file.is_a? Array
      if input_file.empty?
        opn; e "Please provide at the least one file name to the method "\
               "#{sfancy(__method__.to_s+'()')}"
      else
        input_file.each {|entry|
          ffmpeg_denoise(entry, lowpass_value, highpass_value)
        }
      end
    else
      _ = 'ffmpeg -i '+input_file+' '
      _ << '-af "'
      _ << 'lowpass=f='+lowpass_value.to_s+','
      _ << 'highpass=f='+highpass_value.to_s
      _ << '" '
      _ << 'denoised_output_'+input_file.to_s
      esystem _
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_use_lame_codec
  #
  # Use this method if you quickly want to replace one audio codec with
  # another - in this context the lame (liblame) audio codec.
  #
  # The input should be a locally existing video file. The -acodec value
  # will be libmp3lame.
  #
  # The input to this method should be an existing video file.
  # ========================================================================= #
  def self.ffmpeg_use_lame_codec(
      this_video_file
    )
    this_video_file = [this_video_file].flatten.compact
    this_video_file.each {|video_file|
      output_file = "output_#{File.basename(video_file)}"
      _ = 'ffmpeg -i "'+video_file+
          '" -vcodec copy -acodec libmp3lame "'+output_file+'"'
      esystem _
    }
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_delay_audio
  #
  # This method will delay the audio of a video-file, by n-seconds.
  #
  # The second argument to this method, `by`, handles that delay.
  # ========================================================================= #
  def self.ffmpeg_delay_audio(
      this_file = ARGV,
      by        = '3.85 seconds' # <- Input your delay here.
    )
    if this_file.is_a? Array
      this_file.each {|entry|
        ffmpeg_delay_audio(entry, by)
      }
    else
      by = by.to_s # Need a String past this point here.
      by.sub!(/ seconds/,'') if by.include? ' seconds' # Strip away ' seconds'.
      _ = 'ffmpeg -i "'+this_file+'"'.dup
      _ << ' -itsoffset '+by
      _ << ' -map 0:v -map 1:a -c copy'
      # ===================================================================== #
      # Next, determine the output-file.
      # ===================================================================== #
      _ << ' "OUTPUT_WITH_DELAYED_AUDIO_'+this_file+'"'
      esystem _
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_analyze_audio_stream
  #
  # This method will use ffmpeg in order to analyze the audio stream.
  #
  # A Hash will be returned by this method, which will keep the
  # decibel-score value.
  #
  # An example result can be seen next:
  #
  #   [Parsed_volumedetect_0 @ 0x1f32d80] n_samples: 20468176
  #   [Parsed_volumedetect_0 @ 0x1f32d80] mean_volume: -15.9 dB
  #   [Parsed_volumedetect_0 @ 0x1f32d80] max_volume: -0.9 dB
  #   [Parsed_volumedetect_0 @ 0x1f32d80] histogram_0db: 4
  #   [Parsed_volumedetect_0 @ 0x1f32d80] histogram_1db: 91
  #   [Parsed_volumedetect_0 @ 0x1f32d80] histogram_2db: 546
  #   [Parsed_volumedetect_0 @ 0x1f32d80] histogram_3db: 1906
  #   [Parsed_volumedetect_0 @ 0x1f32d80] histogram_4db: 6486
  #   [Parsed_volumedetect_0 @ 0x1f32d80] histogram_5db: 36079
  #
  # The Hash may then be like this:
  #
  #   {"/Depot/Audio/Sumo_Knockouts_2018.mp3"=> {:maximum_volume=>-0.9, :mean_volume=>-15.9}}
  #
  # ========================================================================= #
  def self.ffmpeg_analyze_audio_stream(
      of_this_file
    )
    hash = {}
    of_this_file = [of_this_file].flatten.compact
    of_this_file.each {|this_file|
      hash[this_file] = {}
      _ = 'ffmpeg -i '+this_file+' -af "volumedetect" -f null /dev/null 2>&1'
      e
      result = `#{_}`
      max_volume_is = result.scan(/max_volume: (.+)$/).flatten.first.sub(/ dB/,'').to_f
      hash[this_file][:maximum_volume] = max_volume_is # in decibel, dB.
      mean_volume_is = result.scan(/mean_volume: (.+)$/).flatten.first.sub(/ dB/,'').to_f
      hash[this_file][:mean_volume] = mean_volume_is
    }
    hash
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_slow_down_this_video_file
  #
  # This method can be used to slow down a video file.
  #
  # FFmpeg will be doing this slow down action.
  #
  # This uses the "setpts" video filter of ffmpeg, for the video stream
  # at hand. The video stream can be changed without tampering with
  # the audio stream. This is also the default modus operandi for this
  # method here.
  # ========================================================================= #
  def self.ffmpeg_slow_down_this_video_file(
      this_video_file  = ARGV,
      pts_value_to_use = 2.0
    )
    if this_video_file.is_a? Array
      this_video_file.each {|entry|
        ffmpeg_slow_down_this_video_file(entry, pts_value_to_use)
      }
    else
      # ===================================================================== #
      # To slow down a video, a  multiplier greater than 1 has to be used: 
      # ===================================================================== #
      _ = "ffmpeg -i #{this_video_file} -filter:v \"setpts=#{pts_value_to_use}*PTS\" "\
          "OUTPUT_SLOWED_DOWN_VIDEO_#{this_video_file}"
      esystem _
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_flip_video_to_the_right
  #
  # This toplevel-method can be used to flip a video to the right,
  # that is, to rotate it by +90 degrees.
  # ========================================================================= #
  def self.ffmpeg_flip_video_to_the_right(i)
    _ = "ffmpeg -i #{i} -vf transpose=1 flipped_to_the_right_#{i}"
    esystem _
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_flip_video_to_the_left
  #
  # This toplevel-method can be used to flip a video to the left,
  # that is, to rotate it by -90 degrees.
  #
  # Strangely enough the number 2 denotes left-hand flipping (or,
  # to remain in ffmpeg-parlance, this means "90°CounterClockwise"
  # rotation).
  # ========================================================================= #
  def self.ffmpeg_flip_video_to_the_left(i)
    _ = "ffmpeg -i #{i} -vf transpose=2 flipped_to_the_left_#{i}"
    esystem _
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_extract_audio
  #
  # This method will specifically use ffmpeg in order to extract the
  # audio from a video or audio file.
  # ========================================================================= # 
  def self.ffmpeg_extract_audio(
      i = input.mkv, # The input filename.
      o = :ogg       # The output filename.
    )
    extname = File.extname(i)
    case o
    when :ogg
      use_this_name = i.
                      sub(/#{extname}$/,'')
      o = use_this_name+'.ogg'
    end
    esystem 'ffmpeg -i '+i+' -vn '+o
  end

  # ========================================================================= #
  # === MultimediaParadise.ffmpeg_convert_this_multimedia_file
  #
  # This method can be used to convert a file from A to B, through
  # ffmpeg.
  #
  # Remember that the command ffmpeg -codecs will print every codec
  # FFmpeg knows about.
  # ========================================================================= #
  def self.ffmpeg_convert_this_multimedia_file(
      input_file  = 'input.mp3',
      output_file = :mkv #'output.ogg'
    )
    yielded = nil
    if block_given?
      yielded = yield
    end
    if File.exist? input_file
      case output_file
      when :mkv
        extname = File.extname(input_file)
        use_this_name = input_file.
                        sub(/#{extname}$/,'')
        output_file = use_this_name+'.mkv'
      end
      _ = "ffmpeg -i #{input_file}"
      case yielded
      # ===================================================================== #
      # === :copy_audio
      # ===================================================================== #
      when :copy_audio
        _ << ' -c:a copy' 
      # ===================================================================== #
      # === :libvorbis_audiostream
      # ===================================================================== #
      when :libvorbis_audiostream
        _ << ' -c:a libvorbis'
      # ===================================================================== #
      # === :vp9_codec
      # ===================================================================== #
      when :vp9_codec
        _ << ' -c:v vp9' 
      else
        if yielded.is_a? Hash
          # ================================================================= #
          # === :framerate
          # ================================================================= #
          if yielded.has_key? :framerate
            _ << " -r #{yielded.delete(:framerate)}"
          end
        end
      end
      _ << " #{output_file}"
      e _
      system _
    else
      e 'No file exists at '+input_file+'.'
    end
  end

end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise.ffmpeg_convert_this_multimedia_file
end # ffmpeg_convert_this_multimedia_file