#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::GUI::Gtk::YoutubeDownloaderModule
#
# The purpose of this class is to be able to download a remote youtube
# file, either via youtube-dl or some other way, and to automatically
# modify this downloaded video file.
#
# Usage example:
#
#   require 'multimedia_paradise/youtube_downloader.rb'
#   MultimediaParadise::GUI::Gtk::YoutubeDownloader.new
#
# =========================================================================== #
# require 'multimedia_paradise/gui/shared_code/youtube_downloader/youtube_downloader_module.rb'
# include ::MultimediaParadise::GUI::Gtk::YoutubeDownloaderModule
# =========================================================================== #
module MultimediaParadise

module GUI

module Gtk

module YoutubeDownloaderModule

  begin
    require 'wget'
  rescue LoadError; end

  require 'gtk_paradise/requires/require_the_base_module.rb'
  include ::Gtk::BaseModule

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === WORK_DIRECTORY
  # ========================================================================= #
  WORK_DIRECTORY = '/home/Temp/MultimediaParadise/'

  # ========================================================================= #
  # === TITLE
  # ========================================================================= #
  TITLE = 'Youtube Downloader'

  # ========================================================================= #
  # === WIDTH
  # ========================================================================= #
  WIDTH = '90%'

  # ========================================================================= #
  # === HEIGHT
  # ========================================================================= #
  HEIGHT = 200

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = nil,
      run_already = true
    )
    super(:vertical)
    reset
    set_input(i)
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    reset_the_internal_variables
    # ======================================================================= #
    # === @configuration
    # ======================================================================= #
    @configuration = [true, __dir__, NAMESPACE]
    set_file_name
    set_use_this_font(:dejavu_condensed_22)
    # ======================================================================= #
    # === @title
    # ======================================================================= #
    @title  = TITLE
    # ======================================================================= #
    # === @width
    # ======================================================================= #
    set_width(WIDTH)
    # ======================================================================= #
    # === @height
    # ======================================================================= #
    set_height(HEIGHT)
    use_gtk_paradise_project_css_file
    append_project_css_file
    increase_font_size
  end

  # ========================================================================= #
  # === padding?
  # ========================================================================= #
  def padding?
    2
  end

  # ========================================================================= #
  # === border_size?
  # ========================================================================= #
  def border_size?
    2
  end

  # ========================================================================= #
  # === connect_skeleton                                        (connect tag)
  # ========================================================================= #
  def connect_skeleton
    abort_on_exception
    add(@box_xyz)
  end

  # ========================================================================= #
  # === set_input
  # ========================================================================= #
  def set_input(i = '')
    i = i.first if i.is_a? Array
    i = i.to_s.dup
    @input = i
  end

  # ========================================================================= #
  # === input?
  # ========================================================================= #
  def input?
    @input
  end

  # ========================================================================= #
  # === url?
  # ========================================================================= #
  def url?
    @url_input.text
  end

  # ========================================================================= #
  # === enter_working_directory
  # ========================================================================= #
  def enter_working_directory
    ::MultimediaParadise.cd(work_directory?)
  end

  # ========================================================================= #
  # === ensure_that_the_work_directory_exists
  # ========================================================================= #
  def ensure_that_the_work_directory_exists
    ::MultimediaParadise.mkdir(WORK_DIRECTORY) unless File.directory? WORK_DIRECTORY
  end

  # ========================================================================= #
  # === report_what_directory_will_be_the_working_directory
  # ========================================================================= #
  def report_what_directory_will_be_the_working_directory
    opnn; e 'The working directory will be `'+::Colours.sdir(working_directory?)+'`.'
  end

  # ========================================================================= #
  # === working_directory?
  # ========================================================================= #
  def working_directory?
    WORK_DIRECTORY
  end; alias work_directory? working_directory? # === work_directory?

  # ========================================================================= #
  # === opnn
  # ========================================================================= #
  def opnn
    Opn.opn(namespace: NAMESPACE)
  end

  # ========================================================================= #
  # === MultimediaParadise::YoutubeDownloader.title?
  # ========================================================================= #
  def self.title?
    TITLE
  end

  # ========================================================================= #
  # === set_file_name
  # ========================================================================= #
  def set_file_name(i = nil)
    i = i.dup if i
    @file_name = i
  end

  # ========================================================================= #
  # === file_name?
  # ========================================================================= #
  def file_name?
    @file_name
  end

  # ========================================================================= #
  # === rename_main_file
  # ========================================================================= #
  def rename_main_file
    _ = file_name?
    if file_name?.include? ' '
      _.tr!(' ','_')
    end
    _.squeeze!('_') if _.include? '__'
    _.delete!('ó') if _.include? 'ó'
    _.delete!('(') if _.include? '('
    _.delete!(')') if _.include? ')'
    unless _ == file_name?
      mv(file_name?, _)
      set_file_name(_)
    end
  end

  # ========================================================================= #
  # === set_default_gui_font_size
  # ========================================================================= #
  def set_default_gui_font_size(
      use_this_size = 22
    )
    set_font('Sans '+use_this_size.to_s)
  end

  # ========================================================================= #
  # === add_to_widget
  # ========================================================================= #
  def add_to_widget(i)
    # i = GLib.convert(i, 'iso-8859-1', 'utf-8') # Do a conversion here.
    @textarea.buffer.append(i+N)
  end

  # ========================================================================= #
  # === create_the_textarea
  # ========================================================================= #
  def create_the_textarea
    # ======================================================================= #
    # === @textarea
    # ======================================================================= #
    @textarea = gtk_textarea
    @textarea.bblack1
    @textarea.set_size_request(300, 50)
  end

  # ========================================================================= #
  # === create_skeleton
  # ========================================================================= #
  def create_skeleton
    create_the_textarea
    @url_input = gtk_input
    @url_input.set_size_request(220, 40)
    _ = 'https://youtube.com/watch?v=...'
    @url_input.set_placeholder_text(_)
    @url_input.hint = 
      'Provide the URL here, which may start like <b>'+_+'</b>.'
    @button_do_download = gtk_button('Download remote URL')
    @button_do_download.lightblue
    @button_do_download.clear_background
    @button_do_download.bblack1
    @button_do_download.on_hover(:mintcream)
    @button_do_download.hint = 
      '<b>Clicking</b> on this button will download the remote URL '\
      'given. Provide the remote URL in the input-field to '\
      'the <b>left side</b> of this button.'
    @button_do_download.on_clicked {
      do_download
    }
    hbox = gtk_hbox
    hbox.maximal(@url_input, 2)
    hbox.maximal(@button_do_download, 2)
    minimal(hbox)
    minimal(@textarea)
  end

  # ========================================================================= #
  # === do_download
  #
  # For now we will use youtube-dl for the actual download.
  # ========================================================================= #
  def do_download(
      _ = url?
    )
    opnn; e "Downloading `#{sfancy(_)}` using youtube-dl."
    _ = 'youtube-dl '+_
    opnn; e _
    result = `#{_}`
    # ======================================================================= #
    # Next, we will search for [ffmpeg] entries.
    # ======================================================================= #
    _ = result.split(N).select {|line|
      line.include? '[ffmpeg] Merging formats'
    }
    _ = _.first
    regex = /\[ffmpeg\] Merging formats into "(.+)"/ # See: http://rubular.com/r/YYVCi91eTb
    _.match(regex)
    set_file_name($1)
    rename_main_file
    opnn; e 'The file name is: `'+sfile(file_name?)+'`.'
    add_to_widget(file_name?)
    opnn; e 'All done!'
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    ensure_that_the_work_directory_exists
    report_what_directory_will_be_the_working_directory
    enter_working_directory
    create_skeleton
    set_default_gui_font_size
  end

  # ========================================================================= #
  # === MultimediaParadise::GUI::Gtk::YoutubeDownloader[]
  # ========================================================================= #
  def self.[](i = '')
    new
  end

  # ========================================================================= #
  # === MultimediaParadise::GUI::Gtk::YoutubeDownloaderModule.run
  # ========================================================================= #
  def self.run
    require 'gtk_paradise/run'
    _ = ::MultimediaParadise::GUI::Gtk::YoutubeDownloader.new(ARGV)
    r = ::Gtk.run
    r << _
    r.set_size_request(_.width?,_.height?)
    r.automatic_title
    r.top_left_then_run
  end

end; end; end; end