#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# This class shall be representative for all Chord objects.
# =========================================================================== #
module MultimediaParadise

class Chord # === MultimediaParadise::Chord

  attr_reader :notes

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(notes)
    if notes.size < 2
      raise ArgumentError, 'Chords must have at least two notes'
    end
    @notes = Set.new(notes) { |note|
      if note.is_a?(Note)
        note
      else
        Note.new(note)
      end
    }
  end

  # ========================================================================= #
  # === eql?
  # ========================================================================= #
  def eql?(other_chord)
    @notes == other_chord.notes
  end

  # ========================================================================= #
  # === hash
  # ========================================================================= #
  def hash
    @notes.hash
  end

  # ========================================================================= #
  # === ==
  # ========================================================================= #
  def ==(other_chord)
    self.eql?(other_chord)
  end

  # ========================================================================= #
  # === note_strings
  #
  # Spec and implement.
  # ========================================================================= #
  def note_strings
    Set.new(@notes.map(&:note_string))
  end

  # ========================================================================= #
  # === describe
  #
  # Describe the various minor and major chords.
  # ========================================================================= #
  def describe
    note_array = @notes.to_a.sort
    distances = (1...note_array.size).map { |i|
      note_array[0].distance_to(note_array[i])
    }
    quality = case distances
    when [4, 7]
      :major
    when [3, 7]
      :minor
    when [3, 6]
      :diminished
    when [4, 8]
      :augmented
    when [4, 7, 11]
      :major_7
    when [3, 7, 10]
      :minor_7
    when [3, 6, 9]
      :diminished_7
    when [3, 6, 10]
      :half_diminished_7
    when [4, 8, 10]
      :augmented_7
    end
    [note_array.first.letter, quality]
  end

  # ========================================================================= #
  # === to_s
  # ========================================================================= #
  def to_s
    @notes.to_a.sort.map(&:to_s).join(' / ')
  end

  # ========================================================================= #
  # === inversion
  #
  # Give the n-th inversion of the chord which simply adjusts the lowest
  # N notes up by one octave.
  #
  # @returns [Chord]: the specified inversion of chord
  # ========================================================================= #
  def inversion(amount)
    if amount < 1
      raise ArgumentError,
        'Inversion amount must be greater than or equal to 1'
    end
    if amount >= @notes.size
      raise ArgumentError,
        'Not enough notes in chord for a proper inversion.'
    end
    note_array = @notes.to_a.sort
    notes = (0 .. .amount).map { note_array.shift.adjust_by_semitones(12) }
    Chord.new(notes + note_array)
  end

  # ========================================================================= #
  # === first_inversion
  #
  # Calls inversion(1)
  # ========================================================================= #
  def first_inversion
    self.inversion(1)
  end

  # ========================================================================= #
  # === second_inversion
  #
  # Calls inversion(2)
  # ========================================================================= #
  def second_inversion
    self.inversion(2)
  end

  # ========================================================================= #
  # === third_inversion
  #
  # Calls inversion(3)
  # ========================================================================= #
  def third_inversion
    self.inversion(3)
  end

  # ========================================================================= #
  # === MultimediaParadise::Chord.parse_chord_string
  # ========================================================================= #
  def self.parse_chord_string(
      chord_string,
      assumed_octave = nil
    )
    if note_string_match = chord_string.match(/^([A-Ga-g])([#b]?)([^\d]*)(\d*)$/)
      full_string, note, accidental, interval, octave = note_string_match.to_a
      if note.empty? && assumed_octave.nil?
        raise ArgumentError, 'No octave found and no octave assumed.'
      end
      # ===================================================================== #
      # Build a new Note here, which will also be returned.
      # full_string is irrelevant though.
      # ===================================================================== #
      Note.new(
        note + accidental + octave, assumed_octave
      ).chord(interval) { full_string }
    end
  end

end; end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::Chord.new
end