#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::::ConvertAudioToVideoWithImage
#
# This class can take an audio file, usually a .mp3 file, and combine
# it with a still image, before outputting an .avi file.
#
# The file here was created in July 2018, in order to convert some .mp3
# files into video files that could then be uploaded.
#
# =========================================================================== #
# Usage examples:
#
#   require 'multimedia_paradise/multimedia/convert_audio_to_video_with_image.rb'
#
#   MultimediaParadise::ConvertAudioToVideoWithImage.new
#   MultimediaParadise.to_mp4_with_this_image(ARGV[0], ARGV[1])
# 
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

class ConvertAudioToVideoWithImage < ::MultimediaParadise::Base # === MultimediaParadise::ConvertAudioToVideoWithImage.new

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      audio_file  = nil,
      image_file  = nil,
      run_already = true
    )
    reset
    set_use_this_audio_file(audio_file)
    set_use_this_image_file(image_file)
    if block_given?
      yielded = yield
      case yielded
      # === :to_mp4
      when :to_mp4
        do_store_as_mp4_file
      end
    end
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @use_this_as_file_extension_for_the_output
    # ======================================================================= #
    @use_this_as_file_extension_for_the_output = 'avi'
  end

  # ========================================================================= #
  # === do_store_as_mp4_file
  # ========================================================================= #
  def do_store_as_mp4_file
    @use_this_as_file_extension_for_the_output = 'mp4'
  end

  # ========================================================================= #
  # === set_use_this_audio_file
  # ========================================================================= #
  def set_use_this_audio_file(i)
    i = i.first if i.is_a? Array
    @use_this_audio_file = i
  end

  # ========================================================================= #
  # === audio_file?
  # ========================================================================= #
  def audio_file?
    @use_this_audio_file
  end

  # ========================================================================= #
  # === set_use_this_image_file
  # ========================================================================= #
  def set_use_this_image_file(i)
    i = i.first if i.is_a? Array
    @use_this_image_file = i
  end

  # ========================================================================= #
  # === image_file?
  # ========================================================================= #
  def image_file?
    @use_this_image_file
  end

  # ========================================================================= #
  # === do_the_conversion
  #
  # This is the method that does the actual conversion-step.
  # ========================================================================= #
  def do_the_conversion
    image_file = image_file?
    e "The image `#{sfile(image_file)}` will be used."
    output_file_name = audio_file?.dup.
                       sub(/\.mp3$/,'')+
                       '.'+
                       @use_this_as_file_extension_for_the_output
    # ======================================================================= #
    # _ = 'ffmpeg -loop 1 -r 1 -i '+image_file+
    #    ' -i '+audio_file?+' '\
    #    '-c:a copy -shortest '+
    #    output_file_name
    # ======================================================================= #
    # The next one was added as of 17.05.2020.
    #
    # Example:
    #
    #   ffmpeg -i pic-0.png -i slide1_audio.mp3 -acodec copy -vcodec libx264 final.mp4
    #
    # ======================================================================= #
    _ = 'ffmpeg'\
        ' -i '+image_file+
        ' -i '+audio_file?+
        ' -acodec copy'+
        ' -vcodec libx264'+
        ' '+output_file_name
    esystem _
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    do_the_conversion
  end

  # ========================================================================= #
  # === MultimediaParadise::ConvertAudioToVideoWithImage[]
  # ========================================================================= #
  def self.[](i = '')
    new(i)
  end

end

# =========================================================================== #
# === MultimediaParadise.to_avi_with_this_image
#
# Easier toplevel-method to add a static image file to an audio file.
# =========================================================================== #
def self.to_avi_with_this_image(audio_file, image_file)
  MultimediaParadise::ConvertAudioToVideoWithImage.new(audio_file, image_file)
end

# =========================================================================== #
# === MultimediaParadise.to_mp4_with_this_image
#
# Easier toplevel-method to add a static image file to an audio file.
# =========================================================================== #
def self.to_mp4_with_this_image(audio_file = 'slide1_audio.mp3', image_file = 'image.png')
  MultimediaParadise::ConvertAudioToVideoWithImage.new(audio_file, image_file) { :to_mp4 }
end

end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::ConvertAudioToVideoWithImage.new(ARGV[0], ARGV[1])
end # convert_audio_to_video_with_image
    # convert_audio_to_video_with_image /Depot/jjj/Infektionsbio.mp3 /Depot/jjj/A_CAT.jpg