#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::PlayFromThisList
#
# This class can play multimedia-files that are input via a file or
# a list. This allows us to batch-play files based on the given
# input, which do not need to be in a strict format (e. g. input
# being 100% equal to the filename at hand).
#
# Usage example:
#
#   MultimediaParadise::PlayFromThisList.new(ARGV)
#
# =========================================================================== #
# require 'multimedia_paradise/multimedia/play_from_this_list.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

class PlayFromThisList < ::MultimediaParadise::Base # === MultimediaParadise::PlayFromThisList

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      commandline_arguments = nil,
      run_already           = true
    )
    reset
    set_commandline_arguments(
      commandline_arguments
    )
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @dataset
    # ======================================================================= #
    @dataset = nil
  end

  # ========================================================================= #
  # === set_input
  # ========================================================================= #
  def set_commandline_arguments(i = '')
    i = [i].flatten.compact
    @commandline_arguments = i
  end

  # ========================================================================= #
  # === commandline_arguments?
  # ========================================================================= #
  def commandline_arguments?
    @commandline_arguments
  end

  # ========================================================================= #
  # === determine_from_which_list_to_play
  # ========================================================================= #
  def determine_from_which_list_to_play
    first = @commandline_arguments.first
    if first
      if first.is_a? String
        # =================================================================== #
        # In this case assume the user did input a list.
        # =================================================================== #
        @dataset = first
      elsif File.exist? first
        @dataset = File.readlines(first)
      end
      try_to_sanitize_the_dataset
    end
  end

  # ========================================================================= #
  # === try_to_sanitize_the_dataset
  # ========================================================================= #
  def try_to_sanitize_the_dataset
    if @dataset.is_a? String
      @dataset = @dataset.split(N).reject {|line|
        line.strip.empty?
      }
    end
    # ======================================================================= #
    # Past this point we assume to only have an Array. Sanitize this
    # Array further:
    # ======================================================================= #
    @dataset.map! {|entry|
      entry.strip!
      if entry.start_with? '- '
        entry[0,2] = ''
      end
      entry.strip
    }
  end

  # ========================================================================= #
  # === play_the_files_in_the_dataset
  # ========================================================================= #
  def play_the_files_in_the_dataset
    if @dataset.nil? or @dataset.empty?
      opnn; e 'No valid input has been given to this class.'
    else
      @dataset.each {|entry| play_this_file(entry) }
    end
  end

  # ========================================================================= #
  # === play_this_file
  # ========================================================================= #
  def play_this_file(i)
    if i.include? "'"
      i = '"'+i+'"'
    end
    esystem "#{MultimediaParadise.use_this_player} #{i}"
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    determine_from_which_list_to_play
    play_the_files_in_the_dataset
  end

  # ========================================================================= #
  # === MultimediaParadise::PlayFromThisList[]
  # ========================================================================= #
  def self.[](i = '')
    new(i)
  end

end

# =========================================================================== #
# === MultimediaParadise.play_from_this_list
# =========================================================================== #
def self.play_from_this_list(i = ARGV)
  ::MultimediaParadise::PlayFromThisList.new(i)
end; self.instance_eval { alias play_this_list play_from_this_list } # === MultimediaParadise.play_this_list

end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::PlayFromThisList.new(ARGV)
end # play_from_this_list