#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::ReadMetaTags
#
# This class can read the meta tags of e. g. .mp3 files and so forth,
# and display them. It requires TagLib for this to work, presently.
#
# Usage example:
#
#   MultimediaParadise::ReadMetaTags.new
#
# =========================================================================== #
# require 'multimedia_paradise/multimedia/read_meta_tags.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

class ReadMetaTags < ::MultimediaParadise::Base # === MultimediaParadise::ReadMetaTags.new

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === N_LJUST
  # ========================================================================= #
  N_LJUST = 18

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = nil,
      run_already = true
    )
    reset
    set_input(i)
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @title
    # ======================================================================= #
    @title = nil
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
  end

  # ========================================================================= #
  # === set_title
  # ========================================================================= #
  def set_title(i = nil)
    @title = i
  end

  # ========================================================================= #
  # === set_input
  # ========================================================================= #
  def set_input(i = '')
    i = [i].flatten.compact
    @input = i # Must be an Array.
  end

  # ========================================================================= #
  # === input?
  # ========================================================================= #
  def input?
    @input
  end

  # ========================================================================= #
  # === display_the_title
  # ========================================================================= #
  def display_the_title
    e '  Title: '.ljust(N_LJUST)+
      orange(@title.to_s)
  end

  # ========================================================================= #
  # === read_meta_tags_of_this_file
  # ========================================================================= #
  def read_meta_tags_of_this_file(this_file)
    # ===================================================================== #
    # Load the file if it exists.
    # ===================================================================== #
    if File.exist? this_file
      n_ljust = N_LJUST
      TagLib::FileRef.open(this_file) { |fileref|
        unless fileref.null?
          tag = fileref.tag
          cliner
          set_title(tag.title)
          display_the_title
          e '  Artist: '.ljust(n_ljust)+
            slateblue(tag.artist.to_s)
          e '  Album: '.ljust(n_ljust)+
            lightgreen(tag.album.to_s)
          e '  Year: '.ljust(n_ljust)+
            lightgreen(tag.year.to_s)
          # e '  Track number: '.ljust(n_ljust)+tag.track.to_s
          if tag.genre and !tag.genre.empty?
            e '  Genre: '.ljust(n_ljust)+
              tag.genre.to_s+::Colours.rev
          end
          e '  Comment: '.ljust(n_ljust)+
            crimson(tag.comment.to_s)
          properties = fileref.audio_properties
          e '  Song length: '.ljust(n_ljust)+
            mediumslateblue(properties.length.to_s+' seconds')
          cliner
          e
        end
      }
    else
      opnn; no_file_exists_at(this_file)
    end
  end

  # ========================================================================= #
  # === try_to_load_the_taglib_library
  # ========================================================================= #
  def try_to_load_the_taglib_library
    begin
      require 'taglib-ruby'
    rescue LoadError
      e 'The taglib-ruby gem is not installed. Consider installing '\
        'it ( gem install taglib ).'
      exit
    end
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    try_to_load_the_taglib_library
    @input.each {|this_file|
      read_meta_tags_of_this_file(this_file)
    }
  end

  # ========================================================================= #
  # === []
  # ========================================================================= #
  def self.[](i = '')
    new(i)
  end

end

# =========================================================================== #
# === MultimediaParadise.read_meta_tags_of_this_file
# =========================================================================== #
def self.read_meta_tags_of_this_file(i)
  MultimediaParadise::ReadMetaTags.new(i)
end
 
end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::ReadMetaTags.new(ARGV)
end # read_meta_tags *mp3