#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::StartLengthDuration
#
# This simple class holds length, start and duration information.
#
# ALL those values MUST be absolute (positive) values, all the time!
#
# Every time you set @length, @start or @duration, you must invoke
# ensure_proper_values()
#
# Span gives the total duration (from start + duration)
# =========================================================================== #
# require 'multimedia_paradise/multimedia/start_length_duration.rb'
# =========================================================================== #
module MultimediaParadise

class StartLengthDuration

  begin
    require 'colours'
  rescue LoadError; end

  begin
    require 'cliner'
  rescue LoadError; end

  attr_reader :length   # The length, in seconds.
  attr_reader :start    # Specify where to start. In seconds.
  attr_reader :duration # Keeps the right values for the duration. In seconds.

  # ========================================================================= #
  # === DEBUG
  # ========================================================================= #
  DEBUG = false

  include Colours

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      optional_input_duration = nil
    )
    set_duration(optional_input_duration) if optional_input_duration
    reset # set them to 0.
    ensure_proper_values
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    @length   = 0
    @start    = 0
    @duration = 0
    @span     = 0
    @debug    = DEBUG
  end; alias initialize_variables reset # reset tag

  # ========================================================================= #
  # === set_end
  # ========================================================================= #
  def set_end(i) # a bit cheating here. Depends on @start.
    set_duration(i.to_i - @start)
  end

  # ========================================================================= #
  # === set_length
  # ========================================================================= #
  def set_length(i)
    @length = i.to_f
    ensure_proper_values
  end

  # ========================================================================= #
  # === length=
  # ========================================================================= #
  def length=(i) # an alias to set_length
    set_length(i)
  end

  # ========================================================================= #
  # === calculate_span
  # ========================================================================= #
  def calculate_span # the @span is start to duration. Like: '30-50'
    @span = @start.to_s+'-'+(@start + @duration).to_s
  end

  # ========================================================================= #
  # === feedback
  # ========================================================================= #
  def feedback
    cliner {
      e 'Length (Total length of this video file, in seconds): '+
         simp(@length.to_s)
      e 'Start: '+simp(@start.to_s)
      e 'Duration: '+simp(@duration.to_s)
      e 'Span: '+simp(@span)
    }
  end; alias debug feedback # === debug

  # ========================================================================= #
  # === be_verbose
  # ========================================================================= #
  def be_verbose
    @debug = true
  end

  # ========================================================================= #
  # === span?
  #
  # Give the total duration. Formula: start + duration. In String format.
  # ========================================================================= #
  def span?
    @span
  end; alias span span? # === span

  # ========================================================================= #
  # === start=
  # ========================================================================= #
  def start=(i) # wrapper to the above.
    set_start(i)
  end

  # ========================================================================= #
  # === set_start
  # ========================================================================= #
  def set_start(i) # Use only this method to modify @start.
    @start = i.to_f
    ensure_proper_values
  end

  # ========================================================================= #
  # === ensure_proper_values
  # ========================================================================= #
  def ensure_proper_values # We must keep them as integers and as positives.
    @length   = @length.to_f.abs
    @start    = @start.to_f.abs
    @duration = @duration.to_f.abs
    @span     = @span.to_s # is a string.
  end

  # ========================================================================= #
  # === debug?
  # ========================================================================= #
  def debug?
    @debug
  end

  # ========================================================================= #
  # === set_duration
  #
  # Use this method when you want to set the @duration variable. This is 
  # the actual powerhorse of this script. Since Oct 2012 we will use the
  # absolute value of any given input.
  # ========================================================================= #
  def set_duration(i)
    i = i.to_s
    e "The Input to duration was: `"+sfancy(i)+'`' if debug?
    count = i.count('%') # How many of these do we have?
    if i.include? '%' # Ok, it could include one % or two %
      if count > 1  # has at least two.
        i = i.gsub(/%/,'') # get rid of the % first.
        # Now we assume that the two numbers given work in %
        start_pos, end_pos = i.split('-') # in percent.
        one_percent_is = length / 100.0
        start_pos = one_percent_is * start_pos.to_i
        end_pos   = (one_percent_is * end_pos.to_i) - start_pos
        set_start(start_pos)
        i = end_pos
      else
        i = i.gsub(/%/,'') # get rid of the % first.
        one_percent_is = length / 100.0
        i = one_percent_is * i.to_i
      end
    end
    if i.to_s.include? '-' # Assume negative number means from right end.
      set_start(@length - i.to_i.abs) if count < 2
    end
    @duration = i # Set @duration finally.
    ensure_proper_values
    calculate_span # This is a dependent variable.
  end; alias set_duration= set_duration # === set_duration=

end; end

if __FILE__ == $PROGRAM_NAME
  _ = MultimediaParadise::StartLengthDuration.new
  _.be_verbose
  _.length = 525
  _.set_start 50.3
  # _.set_duration '30'
  # _.feedback
  # _.set_duration '30%'
  # _.feedback  
  # _.set_duration '-30'
  # _.feedback
  # _.set_duration '-30%'
  # _.feedback
  _.set_duration '30%-80%'
  _.feedback
end # $RUBY_TIME/start_length_duration.rb