#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'multimedia_paradise/toplevel_methods/chop_off_first_n_seconds.rb'
# =========================================================================== #
module MultimediaParadise

  require 'multimedia_paradise/toplevel_methods/return_duration_of_this_multimedia_file.rb'
  require 'multimedia_paradise/toplevel_methods/e.rb'
  require 'multimedia_paradise/toplevel_methods/esystem.rb'
  require 'multimedia_paradise/colours/colours.rb'

  # ========================================================================= #
  # === FIVE_MINUTES
  # ========================================================================= #
  FIVE_MINUTES = 60 * 5

  # ========================================================================= #
  # === TEN_MINUTES
  # ========================================================================= #
  TEN_MINUTES = 60 * 10

  # ========================================================================= #
  # === MultimediaParadise.chop_off_first_n_seconds
  #
  # Use this method to chop off the first n seconds from a multimedia file.
  #
  # By default, we will chop off the first 10 minutes.
  #
  # This just wraps ffmpeg, such as:
  #
  #   ffmpeg -ss 0 -t 30 -i file.mp3 file.wav
  #
  # Options explained:
  #   -ss 0 - Start at 0 seconds
  #   -t 30 - Capture 30 seconds (from 0, so 0:00 - 0:30). If you want 1 minute of audio, use -t 60.
  #   file.mp3 - Input file
  #   file.wav - output file
  #
  # Specific usage example:
  #   MultimediaParadise.chop_off_first_n_seconds(:default, '/Users/x/STUDIUM/UNI_WIEN/321015_Pathophysiologie/audio/01-Pathophysiologie-30.05.2016_Vormittag_Überlebenszeit_Apoptose.mp3')
  # ========================================================================= #
  def self.chop_off_first_n_seconds(
      n_seconds                         = TEN_MINUTES, # 10 minutes.
      of_this_multimedia_file           = ARGV,
      also_do_create_complementary_file = false
    )
    if of_this_multimedia_file.is_a? Array
      of_this_multimedia_file.each {|entry|
        chop_off_first_n_seconds(n_seconds, entry, also_do_create_complementary_file)
      }
    else
      case n_seconds
      when '10 minutes'
        n_seconds = TEN_MINUTES
      when '5 minutes'
        n_seconds = FIVE_MINUTES
      when :default
        n_seconds = TEN_MINUTES
      end
      case also_do_create_complementary_file
      when :do_create_complementary_file
        also_do_create_complementary_file = true
      end
      ffmpeg_command = 'ffmpeg -ss 0 -t '+n_seconds.to_s+' -i '+of_this_multimedia_file
      extname        = File.extname(of_this_multimedia_file)
      output_file    = of_this_multimedia_file.sub(/(#{extname})/, '_CHOPPED_AT_THE_START\\1')
      ffmpeg_command << ' '+output_file
      esystem ffmpeg_command
      # ===================================================================== #
      # Also create the "missing" part of the multimedia file, so that
      # we will end up with two files as a result.
      # ===================================================================== #
      if also_do_create_complementary_file
        e 'Now also creating the complementary file.'
        duration       = ::MultimediaParadise.return_duration_of_this_multimedia_file(of_this_multimedia_file) - n_seconds.to_f
        ffmpeg_command = 'ffmpeg -ss '+n_seconds.to_s+' -t '+duration.to_s+' -i '+of_this_multimedia_file
        extname        = File.extname(of_this_multimedia_file)
        output_file    = of_this_multimedia_file.sub(/(#{extname})/, '_CHOPPED_AT_THE_END\\1')
        ffmpeg_command << ' '+output_file
        e orange(ffmpeg_command)
        system ffmpeg_command
      end
    end
  end

end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise.chop_off_first_n_seconds(:default, ARGV, :do_create_complementary_file)
end # chop_off_first_n_seconds 01-Pathophysiologie-30.05.2016_Vormittag_Überlebenszeit_Apoptose.mp3