#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'multimedia_paradise/toplevel_methods/cut_from_to.rb'
# =========================================================================== #
module MultimediaParadise

  require 'multimedia_paradise/colours/colours.rb'
  require 'multimedia_paradise/toplevel_methods/e.rb'
  require 'multimedia_paradise/toplevel_methods/esystem.rb'

  # ========================================================================= #
  # === MultimediaParadise.video_duration
  #
  # This method is a wrapper over the method called
  # MultimediaParadise.cut_from_to(). It specifically 
  # uses the reversed arrangement of arguments, compared
  # to that other method.
  #
  # Specific usage example:
  #
  #   MultimediaParadise.video_duration(of_this_file, '00:30:00-00:35:00')
  #   MultimediaParadise.video_duration('/home/x/studium/UNI_WIEN/321005_Grundlagen_der_Chemie/videos/DONE_Einheit_01_Grundlagen.mp4', '00:04:00-20:35:00')
  #
  # ========================================================================= #
  def self.video_duration(
      of_this_file,
      time_range = '00:30:00-00:35:00'
    )
    MultimediaParadise.cut_from_to(time_range, of_this_file)
  end

  # ========================================================================= #
  # === MultimediaParadise.cut_from_to
  #
  # This method will make use of ffmpeg to cut out segments from a video
  # file.
  #
  # ffmpeg can respond to the -ss option to specify a start timestamp. The
  # -t option can be used to specify the encoding duration.
  #
  # One constraint is that the timestamps need to be in HH:MM:SS.xxx format;
  # or in seconds.
  #
  # The following example in plain ffmpeg skips the first 30 seconds, and
  # then clips for at the least 10 seconds after that:
  #
  #   ffmpeg -ss 00:00:30.0 -i input.wmv -c copy -t 00:00:10.0 output.wmv
  #
  # Input to this method should be in a format like this:
  #
  #   MultimediaParadise.cut_from_to('00:00:01-00:25:10',   this_file: 'Tales_from_the_crypt_S05E10_Came_the_Dawn_1993.mp4')
  #   MultimediaParadise.cut_from_to('00:02:23 / 00:28:04', this_file: 'Tales_from_the_crypt_S05E10_Came_the_Dawn_1993.mp4')
  #
  # ========================================================================= #
  def self.cut_from_to(
      from_to_position = '',
      hash             = {}
    )
    use_this_file = nil
    start_position = '00:00:00.0' # This is the starting position.
    duration       = nil
    if hash.is_a? String
      hash = { this_file: hash }
    end
    if from_to_position.is_a?(Hash) and hash.is_a?(Hash)
      hash.update(from_to_position)
      from_to_position = '' # And re-set to the old default again.
    end
    if hash.is_a?(Hash) and hash.has_key?(:this_file)
      # ===================================================================== #
      # Designate the file that is to be used next:
      # ===================================================================== #
      use_this_file = hash.delete(:this_file)
    end
    case use_this_file
    # ======================================================================= #
    # === :default
    # ======================================================================= #
    when :default
      use_this_file =
        'TALES_FROM_THE_CRYPT/Tales_from_the_crypt_S06E11_Surprise_Party.mkv'
    end
    if use_this_file.start_with? 'TALES_FROM_THE_CRYPT/'
      use_this_file = use_this_file.dup if use_this_file.frozen?
      use_this_file.sub!(
        Regexp.quote('TALES_FROM_THE_CRYPT/'),
        "#{HOME_DIRECTORY_OF_USER_X}VIDEO/Tales_from_the_crypt/"
      )
    end
    if hash.is_a? Hash
      if hash.has_key?(:to)
        duration = hash[:to]
      end
    end
    if from_to_position.is_a?(String) and from_to_position.frozen?
      from_to_position = from_to_position.dup
    end
    if use_this_file
      # ===================================================================== #
      # The input contains a valid file.
      # ===================================================================== #
      if File.exist? use_this_file
        _ = "ffmpeg -i #{use_this_file} -c copy".dup
        if from_to_position.is_a?(String)
          from_to_position.delete!(' ')
        end
        seek_this_token = '-' # <- This can be '-' or '/'.
        if from_to_position.is_a? String
          if from_to_position.include? '/'
            seek_this_token = '/'
          end
          if from_to_position.include?(seek_this_token)
            splitted = from_to_position.split(seek_this_token)
            start_position = splitted.first
            duration = splitted.last
          end
        end
        _ << " -ss #{start_position}"
        _ << " -t #{duration}"
        dirname = "#{File.dirname(use_this_file)}/".squeeze '/'
        output_file = dirname+'output_'+File.basename(use_this_file)
        _ << " #{output_file}"
        esystem _
        if File.exist? output_file
          e "Created file at `#{sfile(output_file)}`."
          return File.absolute_path(output_file) # And return it here, just to make this method more useful.
        end
      else
        e "No file exists at #{use_this_file}."
      end
    else
      e 'Please provide a valid file, via a key called '\
        ':this_file as part of the second argument '\
        '(which is a Hash).'
    end
  end; self.instance_eval { alias cut cut_from_to } # === MultimediaParadise.cut

end

if __FILE__ == $PROGRAM_NAME
  # MultimediaParadise.cut_from_to
  # MultimediaParadise.cut_from_to(
  #   '00:00:01-00:25:10',
  #   this_file: '/home/x/VIDEO/Tales_from_the_crypt/Tales_from_the_crypt_S05E10_Came_the_Dawn_1993.mp4'
  # )
  MultimediaParadise.cut(
    to: '00:22:50',
    # ======================================================================= #
    # Hardcode your file here - or use :default to harcode it in the
    # above method definitions instead.
    # ======================================================================= #
    this_file: :default
  )
end # cut_from_to