#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'multimedia_paradise/toplevel_methods/denoise.rb'
# =========================================================================== #
require 'multimedia_paradise/ffmpeg/ffmpeg.rb'

module MultimediaParadise

  # ========================================================================= #
  # === MultimediaParadise.denoise
  #
  # This method attempts to remove noise from a multimedia file.
  #
  # Note that this uses ffmpeg. Sox can also be used to filter
  # away noise. The first argument is for the lowpass value;
  # the second argument is for the highpass value.
  #
  # The FFMPEG highpass filter is explained here:
  #
  #   https://ffmpeg.org/ffmpeg-filters.html#highpass
  #
  # ========================================================================= #
  def self.denoise(
      input_file,
      lowpass_value  = 3000,
      highpass_value =  200
    )
    ffmpeg_denoise(input_file, lowpass_value, highpass_value)
  end

  # ========================================================================= #
  # === MultimediaParadise.sox_denoise
  #
  # This is similar to MultimediaParadise.denoise() but it uses sox
  # in addition to ffmpeg.
  #
  # See the following link:
  #
  #   http://www.zoharbabin.com/how-to-do-noise-reduction-using-ffmpeg-and-sox/
  #
  # ========================================================================= #
  def self.sox_denoise(
      input_file = 'input.mp4',
      start_time = '00:00:00'
    )
    input_file = input_file.join(' ').strip if input_file.is_a? Array
    # ======================================================================= #
    # (1) Split audio and video streams into two separate files:
    # ======================================================================= #
    _ = "ffmpeg -i #{input_file} -qscale 0 -an output_file_#{input_file}" # Video stream.
    esystem _
    temp_audio_wav = 'output_file_'+input_file.sub(/\.mp4$/,'')+'.wav'
    _ = ' ffmpeg -i '+input_file+' -qscale 0 '+temp_audio_wav # Audio stream.
    esystem _
    # ======================================================================= #
    # (2) Generate a sample of noise from the audio of the file. Choose a
    # segment of the audio where there is no speech and only noise, e. g.
    # where the speaker was silent for a second.
    # ======================================================================= #
    output_wav_file = 'output_noiseaudio.wav'
    _ = 'ffmpeg -i '+input_file+' -vn -ss '+start_time+' -t 00:00:01 '+output_wav_file
    esystem _
    # ======================================================================= #
    # (3) Generate a noise profile in sox:
    # ======================================================================= #
    _ = 'sox '+output_wav_file+' -n noiseprof noise.prof'
    esystem _
    # ======================================================================= #
    # (4) Clean the noise samples from the audio stream:
    # ======================================================================= #
    _ = 'sox '+temp_audio_wav+' '+temp_audio_wav.sub(/\.wav$/,'')+'_tmpaud-clean.wav noisered noise.prof 0.21'
    esystem _
    # ======================================================================= #
    # (5) Change the 0.21 value above ^^^ to adjust the level of sensitivity
    # in the sampling rates (0.2-0.3 often provides the best result).
    # ======================================================================= #
    # (6) Merge audio and video streams back together:
    # ======================================================================= #
    _ = 'ffmpeg -i tmpaud-clean.wav -i tmpvid.mp4 -qscale 0 vid.mp4'
    esystem _
  end

end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise.denoise(ARGV)  
  # MultimediaParadise.sox_denoise(ARGV)
end # denoise