#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'multimedia_paradise/toplevel_methods/merge_multimedia_file.rb'
# =========================================================================== #
# MultimediaParadise.merge_mp3(:default, '*mp3')
# MultimediaParadise.merge_avi_files
# =========================================================================== #
module MultimediaParadise

  require 'multimedia_paradise/colours/colours.rb'
  require 'multimedia_paradise/toplevel_methods/e.rb'
  require 'multimedia_paradise/toplevel_methods/merge_these_videos.rb'

  # ========================================================================= #
  # === MultimediaParadise.merge_multimedia_file                  (merge tag)
  #
  # This method combines several multimedia-related actions, into one
  # bundled method.
  #
  # Do note that the input files come last in the various merge methods,
  # but they do come first for the input to this method here.
  #
  # The reason is that we can pass a variable amount to a method in
  # ruby via the * syntax.
  #
  # Usage example:
  #
  #   MultimediaParadise.merge_multimedia_file(ARGV)
  #
  # ========================================================================= #
  def self.merge_multimedia_file(
      input_files = nil, # This should be an Array.
      output_file = nil, # This should be a String.
      extname     = nil
    )

    require 'multimedia_paradise/conversions/conversions.rb'

    if input_files.is_a?(String) and input_files.include? '*'
      input_files = Dir[input_files]
    end
    input_files.flatten!
    # ======================================================================= #
    # Determine the extname to use next.
    # ======================================================================= #
    if extname.nil? and input_files.is_a?(Array) and !input_files.empty? 
      extname = File.extname(input_files.first)
    end

    extname = extname.to_s.delete('.')

    case extname # This will be a String, as determined above.
    # ======================================================================= #
    # === mp4
    # ======================================================================= #
    when 'mp4'
      eimportant(
        output_file = merge_mp4(output_file, input_files)
      )
    # ======================================================================= #
    # === mp3
    # ======================================================================= #
    when 'mp3' # Act on mp3 files here.
      eimportant(
        output_file = merge_mp3(output_file, input_files)
      )
    # ======================================================================= #
    # === ogg
    # ======================================================================= #
    when 'ogg'
      eimportant(
        output_file = merge_ogg(output_file, input_files)
      )
    # ======================================================================= #
    # === flv
    # ======================================================================= #
    when 'flv'
      eimportant(
        output_file = merge_flv(output_file, input_files)
      )
    # ======================================================================= #
    # === m4v
    # ======================================================================= #
    when 'm4v'
      eimportant(
        output_file = merge_m4v(output_file, input_files)
      )
    else
      output_file = ::MultimediaParadise.merge_these_videos(input_files)
    end
    return output_file
  end # ========================================================================= #
      # === MultimediaParadise.merge()
      #
      # Add an alias.
      # ========================================================================= #
      self.instance_eval { alias merge merge_multimedia_file }

  # ========================================================================= #
  # === MultimediaParadise.merge_mov
  #
  # Use this to merge some .mov files. Hmm. For some reason mencoder
  # makes this an .avi file.
  # ========================================================================= #
  def self.merge_mov(
      output_file = 'output.mov', *input_files
    )
    output_file << '.mov' unless output_file.end_with? '.mov'
    _  = 'mencoder -idx -forceidx -oac mp3lame -ovc xvid -xvidencopts fixed_quant=4'.dup
    _ << ' -o "'+output_file+'"'
    _ << ' '+input_files.join(' ')
    run_sys_command _
  end

  # ========================================================================= #
  # === MultimediaParadise.merge_avi_files
  #
  # This method can be specifically used to merge several .avi files
  # together. The input to this method should come in the form of
  # .avi files.
  # ========================================================================= #
  def self.merge_avi_files(
      *these_avi_files
    )
    these_avi_files = [these_avi_files].flatten.compact
    if these_avi_files.empty?
      these_avi_files = Dir['*.avi'] # Try to find .avi files in this case.
    end
    cmd = 'ffmpeg -i "concat:'+
           these_avi_files.join('|')+
           '" -c copy output.avi'
    e
    e cmd
    e
    system cmd
  end

  # ========================================================================= #
  # === MultimediaParadise.merge_m4v
  # ========================================================================= #
  def self.merge_m4v(
      output_file = nil,
      input_files = []
    )
    output_file = 'output.m4v' # The input is .m4v, so the output should also be .mp4.
    we_created_these_temporary_video_files = ''
    input_files.each_with_index {|input_file, index|
      temporary_video_file = 'temp'+index.to_s+'.ts'
      # ======================================================================= #
      # Next get rid of prior-existing temporary video files, in the
      # event that these exist.
      # ======================================================================= #
      File.delete(temporary_video_file) if File.exist? temporary_video_file
      we_created_these_temporary_video_files << temporary_video_file+'|'
      _ = 'ffmpeg -i '+input_file+' -c copy -bsf:v h264_mp4toannexb '+temporary_video_file
      esystem _, :colourize_output
    }
    if we_created_these_temporary_video_files.end_with? '|'
      we_created_these_temporary_video_files.chop!
    end
    _ = 'ffmpeg -i "concat:'+we_created_these_temporary_video_files+
        '" -c copy -bsf:a aac_adtstoasc '+output_file.to_s
    esystem _, :colourize_output
    if File.exist? output_file
      e 'Stored into the file '+sfile(output_file)+'.'
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.merge_ogg                        (merge ogg files)
  #
  # Use this method if you wish to merge some .ogg files together.
  #
  # We presently use ffmpeg for this.
  # ========================================================================= #
  def self.merge_ogg(
      output_file = 'output.ogg', *input_files
    )
    input_files.flatten! # Flatten them.
    input_files = input_files.join('\|') if input_files.is_a? Array
    _  = 'ffmpeg -i concat:'+input_files # input_files are always an Array.
    _ << ' -acodec copy'
    _ << ' '+output_file
    run_sys_command _
  end

  # ========================================================================= #
  # === MultimediaParadise.merge_mp4
  #
  # Merge mp4 files. This is currently done using ffmpeg's concat protocol.
  #
  # Unfortunately, concatenating .mp4 files is not trivial and may require
  # an intermediate step such as remuxing them to ts (mpegts) before then
  # putting them together into a .mp4 file.
  #
  # The concat protocol follows this rule:
  #
  #   ffmpeg -i "concat:input1.mpg|input2.mpg|input3.mpg" -c copy output.mpg
  #
  # ========================================================================= #
  def self.merge_mp4(
      output_file = 'output.mp4',
      *input_files
    )
    input_files.flatten!
    # ======================================================================= #
    # Pad all input files next, to avoid "foo'bar.mp3" problems.
    # ======================================================================= #
    input_files.map! {|entry|
      entry = '"'+entry+'"' unless entry.start_with?('"')
      entry
    }

    MultimediaParadise.ensure_that_the_output_directory_exists # Our log-base directory must exist.
    output_file = 'output.mp4' if output_file.nil?
    output_file = 'output_'+output_file unless output_file.include? '.'
    output_file << '.mp4' unless output_file.end_with? 'mp4'
    output_file = '"'+output_file+'"' unless output_file.start_with?('"')
    File.delete(output_file) if File.exist? output_file # Don't want any backups.
    # ======================================================================= #
    # Ok now, the first step is to produce intermediate .ts files. Since
    # they are intermediate, we don't have to store them persistently.
    # We do this by iterating over our input_files Array first.
    # ======================================================================= #
    input_files.each_with_index {|the_input_mp4_file, index|
      index += 1
      _ = "ffmpeg -i #{the_input_mp4_file} -c copy -bsf:v h264_mp4toannexb -f mpegts"
      name_of_the_ts_file = MultimediaParadise.store_in_this_directory+'intermediate'+index.to_s+'.ts'
      # ======================================================================= #
      # We must kill old .ts files, just in case that they exist. This is 
      # acceptable since it is a temporary directory anyway.
      # ======================================================================= #
      File.delete(name_of_the_ts_file) if File.exist? name_of_the_ts_file
      _ << ' '+name_of_the_ts_file
      e _
      system _
    }
    # ======================================================================= #
    # Ok, now we created the intermediate .ts files, so now we can apply
    # the concat protocol.
    # ======================================================================= #
    input_files = input_files.flatten.map.with_index {|entry, index|
      index += 1
      entry = MultimediaParadise.store_in_this_directory+'intermediate'+index.to_s+'.ts'
      entry
    }
    _  = 'ffmpeg -i "concat:'+input_files.join('|')+'" -c copy '
    _ << '  -bsf:a aac_adtstoasc '
    _ << output_file
    _.strip!
    esystem _
    if File.exist? output_file
      e "The file should now exist at `#{sfile(output_file)}`."
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.merge_mp3
  #
  # This method will merge a variable number of .mp3 files.
  #
  # Could use an alternative too:
  #   mp3wrap output.mp3 *mp3
  #
  # In "pure" ruby however, try this:
  #
  #   require 'multimedia_multimedia'
  #
  #   MultimediaParadise.merge_mp3(:default, '*mp3')
  #
  # ========================================================================= #
  def self.merge_mp3(
      output_file = 'output.mp3', *input_files
    )
    case output_file
    # This entry point is for the default name.
    when nil, :default
      output_file = 'output.mp3'
    end
    output_file = 'output_'+output_file unless output_file.include? '.'
    output_file << '.mp3' unless output_file.end_with? 'mp3'
    input_files = Dir['*.mp3'] if input_files.first == '*mp3' # This is a shorthand notation for "get all mp3 files".
    if input_files.any? {|entry| entry.include? ' ' }
      input_files.map! {|entry| '"'+entry+'"'}
    end
    # ======================================================================= #
    # We could use either mpgtx or ffmpeg.
    # _  = 'mpgtx -j '
    # _ << '-o '+output_file
    # _ << ' '+input_files.join(' ')
    # ======================================================================= #
    _ = 'ffmpeg -i concat:'+input_files.join('\|')
    _ << " -acodec copy #{output_file}"
    run_sys_command _
    return output_file # And return this output file here as well.
  end

  # ========================================================================= #
  # === MultimediaParadise.vobmerge
  #
  # You must pass the vob files to merge to this method.
  # ========================================================================= #
  def self.vobmerge(*input_files)
    _ = 'cat '.dup
    input_files = input_files.flatten
    input_files.each {|file| _ << "#{file} " }
    _ << ' > merged_'+File.basename(input_files.first) #.gsub(/\.vob/,'')
    run_sys_command _
  end

  # ========================================================================= #
  # === MultimediaParadise.merge_flv
  #
  # Merge .flv files. Uses mencoder right now.
  # ========================================================================= #
  def self.merge_flv(
      output_file = 'output.flv', *input_files
    )
    _  = 'mencoder -forceidx -oac mp3lame -ovc copy'.dup
    _ << ' -o '+output_file
    _ << ' '+input_files.join(' ')
    run_sys_command _
  end

end