#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'multimedia_paradise/toplevel_methods/merge_these_videos.rb'
# =========================================================================== #
module MultimediaParadise

  require 'multimedia_paradise/colours/colours.rb'
  require 'multimedia_paradise/toplevel_methods/esystem.rb'
  require 'multimedia_paradise/toplevel_methods/output_directory.rb'

  # ========================================================================= #
  # === MultimediaParadise.merge_these_videos
  #
  # This method can be used to merge multiple videos together. The
  # "concat" option of ffmpeg will be used for this merge-action.
  #
  # Ideally the video-files that are given as input to this method
  # should all use the same resolution and the  same frame rate.
  #
  # There are several ways how to go about this. A simple one is to
  # simply create a text-file that can then tell ffmpeg which files
  # are to be merged together. This is the strategy that this
  # method will presently (2018) pursue.
  #
  # The ffmpeg-concact options are explained here:
  #
  #   https://trac.ffmpeg.org/wiki/Concatenate
  #
  # Invocation examples:
  #
  #   MultimediaParadise.merge_these_videos('foo.mp4','bar.mp4')
  #   MultimediaParadise.merge('foo.mp4','bar.mp4')
  #
  # ========================================================================= #
  def self.merge_these_videos(
      *i
    )
    i = [i].flatten.map(&:strip)
    e ::Colours.rev+
      ::Colours.steelblue(i.size.to_s)+
      ' video files will be merged next.'
    new_file_extension = nil
    # ======================================================================= #
    # We will store into the following text-file:
    # ======================================================================= #
    text_file = "#{log_directory?}file_that_keeps_track_of_the_videos.md"
    # ======================================================================= #
    # Make sure that the base directory exists:
    # ======================================================================= #
    unless File.directory?(File.dirname(text_file))
      mkdir(File.dirname(text_file))
    end
    File.delete(text_file) if File.exist? text_file
    name_of_the_output_file = 'merged_files_'.dup
    e 'The video files will be analyzed. A log file will be stored at:'
    e
    e "  #{sfile(text_file)}"
    e
    i.each {|this_video_file|
      # ===================================================================== #
      # We have to obtain the absolute file path:
      # ===================================================================== #
      this_video_file = File.absolute_path(this_video_file)
      File.open(text_file, 'a+') { |file|
        file.write("file #{this_video_file}\n")
      }
      raw_name = this_video_file.sub(/#{File.extname(this_video_file)}$/,'').dup
      name_of_the_output_file << File.basename(raw_name).delete('/')
      new_file_extension = File.extname(this_video_file).delete('.')
    }
    # ======================================================================= #
    # Ensure that the file name is not too long, thus truncating.
    # ======================================================================= #
    if name_of_the_output_file.size > 200
      name_of_the_output_file = name_of_the_output_file[0 .. 200]
    end
    name_of_the_output_file << '.'+new_file_extension
    name_of_the_output_file = '"'+name_of_the_output_file.
                              delete("'")+'"'
    # ======================================================================= #
    # Build up the string for ffmpeg next. The option -y is used to
    # overwrite automatically.
    # ======================================================================= #
    _ = 'ffmpeg '\
        '-y '\
        '-safe 0 '\
        '-f concat '\
        '-i '+text_file+
        ' -c copy '+name_of_the_output_file
    esystem _
    e
    e 'Ouput stored at:'
    e
    e "  #{sfile(name_of_the_output_file)}"
    e
    return name_of_the_output_file # <- And also return it here.
  end; self.instance_eval { alias merge merge_these_videos } # === MultimediaParadise.merge

end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise.merge_these_videos(ARGV)
end # merge_these_videos foo.mp4 bar.mp4