#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'multimedia_paradise/toplevel_methods/misc.rb'
# MultimediaParadise.bitrate_of_this_song
# MultimediaParadise.verbose_analyse_this_mp3_file_for_id3_tags
# MultimediaParadise.return_video_codec_of_this_file
# =========================================================================== #
module MultimediaParadise

  require 'multimedia_paradise/constants/misc.rb'
  require 'multimedia_paradise/toplevel_methods/esystem.rb'
  require 'multimedia_paradise/ffmpeg/ffmpeg.rb'

  # ========================================================================= #
  # === MultimediaParadise.loop_this_video
  #
  # This method can be used to loop a video, via ffmpeg.
  #
  # A commandline variant for this would look as follows:
  #
  #  ffmpeg -stream_loop 3 -i  Tata.webm   -c copy output.webm
  #
  # ========================================================================= #
  def self.loop_this_video(this_video = 'foobar.webm', n_times = '5x')
    result = []
    [this_video].flatten.compact.each {|work_on_this_video_file|
      n_times = n_times.to_s.delete('x').sub(/ times/,'')
      output_file = "output_#{work_on_this_video_file}"
      result << output_file
      cmd = "ffmpeg -stream_loop #{n_times} -i #{work_on_this_video_file} -c copy #{output_file}"
      esystem cmd
    }
    return result
  end

  # ========================================================================= #
  # === MultimediaParadise.bin_dir?
  # ========================================================================= #
  def self.bin_dir?
    RbConfig::CONFIG['bindir'].to_s+'/' # Ensure trailing '/'.
  end; self.instance_eval { alias bindir? bin_dir? } # === MultimediaParadise.bindir?

  # ========================================================================= #
  # === MultimediaParadise.copy_bin_directory
  #
  # This method can be used to copy the content of the bin/ directory
  # onto the current working directory.
  #
  # Usage example from within ruby code:
  #
  #   require 'multimedia_paradise'; MultimediaParadise.copy_bin_directory
  #
  # ========================================================================= #
  def self.copy_bin_directory
    file_installed_binaries = "#{MultimediaParadise.project_base_directory?}"\
                              "yaml/internal/installed_binaries.yml"
    mkdir('bin/')
    cd('bin/')
    if File.exist? file_installed_binaries
      e 'Making use of the file '+sfile(file_installed_binaries)+'.'
      install_these_binaries = YAML.load_file(file_installed_binaries)
      installation_directory = `gem env`.split("\n").select {|line| # Check for INSTALLATION DIRECTORY: /root/.gem
        line.include? 'INSTALLATION DIRECTORY:'
      }.first.
      scan(/INSTALLATION DIRECTORY: (.+)$/).flatten.first
      install_these_binaries.each {|entry|
        target = installation_directory+'/bin/'+entry
        if File.exist? target
          e "Copying the file #{target} into the current working "\
            "directory next."
          copy_file(target, Dir.pwd)
        else
          # ================================================================= #
          # Since as of January 2023 first try the bin_dir, before giving up.
          # ================================================================= #
          target = MultimediaParadise.bin_dir?+File.basename(entry)
          if File.exist? target
            e "Copying the file #{target} into the current working "\
              "directory next."
            copy_file(target, Dir.pwd)
          else
            no_file_exists_at(target)
          end
        end
      }
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.return_video_codec_of_this_file
  #
  # This method will use ffmpeg's ffprobe and return the video codec of the
  # given file at hand.
  # ========================================================================= #
  def self.return_video_codec_of_this_file(i = ARGV)
    if i
      if i.is_a? Array
        i = i.first
      end
      cmd = "ffprobe -v error -select_streams v:0 -show_entries stream=codec_name -of default=noprint_wrappers=1:nokey=1 #{i}"
      result = `#{cmd}`
      return result
    end
  end; self.instance_eval { alias video_codec  return_video_codec_of_this_file } # === MultimediaParadise.video_codec
       self.instance_eval { alias video_codec? return_video_codec_of_this_file } # === MultimediaParadise.video_codec?

  # ========================================================================= #
  # === MultimediaParadise.to_gif
  #
  # This method can be used to quickly turn an .avi file or a .mp4
  # file into an animated .gif.
  #
  # The quality is pretty low, so this is not ideal, but if you
  # have a use case for animated .gif files then this is one
  # way to do so.
  #
  # Usage example:
  #
  #   MultimediaParadise.to_gif('foobar.mp4')
  #
  # ========================================================================= #
  def self.to_gif(i)
    [i].flatten.compact.each {|this_video_file|
      _ = File.basename(this_video_file).
               delete_suffix(
                 File.extname(this_video_file)
               )
      name_of_the_output_gif_file = "output_#{_}.gif"
      cmd = "ffmpeg -i #{this_video_file} -pix_fmt rgb8 #{name_of_the_output_gif_file}" # Or use 0rgb.
      esystem(cmd)
      return name_of_the_output_gif_file
    }
  end
  
  # ========================================================================= #
  # === MultimediaParadise.create_noise_profile
  #
  # Note that this method depends on the external program called "sox".
  #
  # Usage example:
  #
  #   MultimediaParadise.create_noise_profile('foobar.mp3')
  #
  # ========================================================================= #
  def self.create_noise_profile(
      of_this_mp3_file
    )
    array_results = []
    [of_this_mp3_file].flatten.compact.each {|work_on_this_mp3_file|
      work_on_this_mp3_file = File.absolute_path(work_on_this_mp3_file)
      filename_without_extension = work_on_this_mp3_file.delete_suffix('.mp3')
      this_profile_file_is_to_be_created = "noise_profile_based_on_#{File.basename(filename_without_extension)}"
      if File.exist?(this_profile_file_is_to_be_created) # Always remove old profiles.
        File.delete(this_profile_file_is_to_be_created)
      end
      cmd_to_run = "sox #{work_on_this_mp3_file} -n noiseprof #{this_profile_file_is_to_be_created}.profile"
      e cmd_to_run
      `#{cmd_to_run}`
      array_results << this_profile_file_is_to_be_created
    }
    return array_results
  end

  # ========================================================================= #
  # === MultimediaParadise.shrink_quality_of_these_mp3_files
  #
  # This method can be used to batch-shrink several .mp3 files in one go.
  #
  # Usage example:
  #
  #   MultimediaParadise.shrink_quality_of_these_mp3_files(Dir['*.mp3'])
  #
  # ========================================================================= #
  def self.shrink_quality_of_these_mp3_files(*array)
    array.flatten.each {|this_mp3_file|
      esystem "lame --mp3input -b 64 #{this_mp3_file}"
    }
  end

  # ========================================================================= #
  # === MultimediaParadise.report_how_many_real_videos_are_stored
  #
  # This method will report how many "real" videos are stored in the
  # file video_collection.yml, which is distributed as part of
  # this gem. You can, however had, also specify your own .yml
  # file, as first argument to this method.
  #
  # To tap into this method, do issue:
  #
  #   mpa --n-videos?
  # 
  # ========================================================================= #
  def self.report_how_many_real_videos_are_stored(
      in_this_yaml_file = file_video_collection?
    )
    if File.exist? in_this_yaml_file
      dataset = YAML.load_file(in_this_yaml_file)
      e rev+
        'There are '+sfancy(dataset.keys.size)+' registered videos '+
        'in the file'
      e '`'+sfile(in_this_yaml_file)+'`.'
    else
      e 'Please supply an existing yaml file (.yml) to this method.'
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.return_the_mpeg_layer_from_this_mp3_file
  # ========================================================================= #
  def self.return_the_mpeg_layer_from_this_mp3_file(i)
    cmd = "mpg123 -t #{i} 2>&1"
    result = `#{cmd}`
    splitted = result.split("\n")
    selected = splitted.select {|line|
      line.start_with? 'MPEG ' # Obtain the MPEG layer here.
    }.uniq
    result = selected.first # Pick the first result here.
    # ======================================================================= #
    # We may have to clean up that result a little bit still.
    # ======================================================================= #
    if result and result.include?(' ') and result.include?('III')
      result = result[0 .. (result.index('III')+2)]
    end
    return result
  end

  # ========================================================================= #
  # === MultimediaParadise.overlay_two_audio_files
  #
  # This method can merge in two audio files, usually .mp3 files.
  #
  # The functionality depends on "sox".
  # ========================================================================= # 
  def self.overlay_two_audio_files(
      first,
      second
    )
    output_file = 'output_file.mp3'
    esystem 'sox --combine mix '+
            first.to_s+
            ' '+
            second.to_s+
            ' '+
            output_file
  end

  # ========================================================================= #
  # === MultimediaParadise.samplerate_of_this_song
  #
  # Usage example:
  #
  #   MultimediaParadise.samplerate_of_this_song('foobar.mp3') # => 44100
  #
  # ========================================================================= #
  def self.samplerate_of_this_song(this_song)
    if this_song.is_a? Array
      this_song = this_song.first
    end
    result = `ffprobe -select_streams a -show_streams #{this_song} 2>&1`
    if result
      result = result.split("\n").select {|entry|
        entry.start_with? 'sample_rate'
      }.first.to_s
    end
    if result and result.include?('=')
      result = result.split('=').last
    end
    return result
  end

  # ========================================================================= #
  # === MultimediaParadise.bitrate_of_this_song
  #
  # This method will make use of ffmpeg's ffprobe to query the bitrate of
  # e. g. a .mp3 song. The return value is a String such as "192 kb/s".
  #
  # The commandline variant would be the following:
  #
  #   ffprobe foobar.mp3 2>&1 | grep Stream
  #
  # Result:
  #
  #   Stream #0:0: Audio: mp3, 11025 Hz, mono, s16p, 32 kb/s
  #
  # Usage example:
  #
  #   MultimediaParadise.bitrate_of_this_song('foobar.mp3') # => "192 kb"
  #
  # ========================================================================= #
  def self.bitrate_of_this_song(this_song)
    use_this_regex = /(\d{1,3} ?kb)\/s/ # See: https://rubular.com/r/CfmHoK5w0goYpU
    if this_song.is_a? Array
      this_song = this_song.first
    end
    result = `ffprobe #{this_song} 2>&1 | grep Stream`.strip
    result = result.scan(use_this_regex).flatten.first.to_s
    return result
  end

  # ========================================================================= #
  # === MultimediaParadise.reverse_audio
  #
  # Reverse an audio file.
  # ========================================================================= #
  def self.reverse_audio(
      this_file = 'output.wav'
    )
    output_file = 'reversed_'+this_file
    cmd = 'sox -n '+this_file+' '+output_file+' reverse'
    run_sys_command cmd
    return output_file
  end

  # ========================================================================= #
  # === MultimediaParadise.generate_sine_tone
  #
  # Generate a n-second sine tone.
  # ========================================================================= #
  def self.generate_sine_tone(
      n_seconds = 5
    )
    output_file = 'output.wav'
    esystem 'sox -n '+output_file+' synth '+n_seconds.to_s+' sine 440'
    return output_file
  end

  # ========================================================================= #
  # === MultimediaParadise.generate_sine_tone_sweep
  #
  # Generate a n-second sine tone sweep.
  # ========================================================================= #
  def self.generate_sine_tone_sweep(
      n_seconds = 5
    )
    output_file = 'output.wav'
    esystem 'sox -n '+output_file+' synth '+n_seconds.to_s+' sine 0:800'
    return output_file
  end

  # ========================================================================= #
  # === MultimediaParadise.repeat_this_song_n_times
  #
  # This will make use of sox and repeat a given song n times.
  # ========================================================================= #
  def self.repeat_this_song_n_times(
      this_song = ARGV,
      n_times   = 3
    )
    [this_song].flatten.compact.each {|a_song|
      if File.exist? a_song
        esystem 'sox '+a_song+' repeated_song_'+a_song+' repeat '+n_times.to_s
      else
        no_file_exists_at(a_song)
      end
    }
  end

  # ========================================================================= #
  # === MultimediaParadise.append_silence_to_this_song
  #
  # Note that this method currently depends on sox. Perhaps someone knows
  # how to do so via ffmpeg; I don't right now.
  #
  # The first argument to this method is the delay, in n seconds. You can
  # use a fraction, such as 0.25 here as well.
  #
  # The second argument is the song, such as "bla.mp3". This filename should
  # exist.
  #
  # The method will then simply add "silence" to the song at hand - it
  # will append "silence".
  #
  # Usage examples:
  #
  #   MultimediaParadise.append_silence_to_this_song(5, 'foobar.mp3')
  #   MultimediaParadise.append_silence_to_this_song(5, 'ack.mp3')
  #   MultimediaParadise.append_silence_to_this_song(.027, 'foobar.mp3')
  #
  # ========================================================================= #
  def self.append_silence_to_this_song(
      add_this_delay      = 5,
      work_on_these_songs = ['foobar.mp3']
    )
    [work_on_these_songs].flatten.compact.each {|this_song|
      if File.exist? this_song
        cmd = "sox #{this_song} padded_#{this_song} pad 0 #{add_this_delay.to_f}"
        esystem cmd
      else
        e 'No file exists at '+this_song.to_s+'.'
      end
    }
  end

  # ========================================================================= #
  # === MultimediaParadise.overlay_this_image_onto_that_video
  #
  # This method can be used to overlay an image over an existing local
  # video file.
  #
  # The options in use are as follows:
  #
  #   overlay=25:25: The image will be positioned 25px to the right and
  #                  25px down, originating from the top left corner (0:0).
  #
  #   enable='between(t,0,20)': The overlay image will be shown from
  #                             00:00:00 to 00:00:20
  #
  # ========================================================================= #
  def self.overlay_this_image_onto_that_video(
      this_image,
      that_video,
      duration_in_n_seconds = 3
    )
    _ = 'ffmpeg -y -i '+that_video+' '\
        '-i '+this_image+' -filter_complex "[0:v][1:v] '\
        'overlay=80:80:enable=\'between(t,0,'+
        duration_in_n_seconds.to_s+')\'" '\
        '-pix_fmt yuv420p -c:a copy '\
        'output.mp4'
    e
    esystem _
    e
  end

  # ========================================================================= #
  # === MultimediaParadise.split_into_chunks_of_n_minutes
  # ========================================================================= #
  def self.split_into_chunks_of_n_minutes(
      this_video = '*.mp4',
      n_minutes  = 5
    )
    this_video = this_video.first if this_video.is_a? Array
    _ = 'ffmpeg -i '+this_video+' -c copy -map 0 -segment_time '\
        '00:0'+n_minutes.to_s+':00 -f segment output%03d.mp4'
    e
    esystem _
    e
  end

  # ========================================================================= #
  # === MultimediaParadise.split_into_chunks_of_5_minutes
  # ========================================================================= #
  def self.split_into_chunks_of_5_minutes(
      i = '*.mp4'
    )
    MultimediaParadise.split_into_chunks_of_n_minutes(i, 5)
  end

  # ========================================================================= #
  # === MultimediaParadise.return_ss_mm_hh
  #
  # This may return a string such as "16:35:07".
  # ========================================================================= #
  def self.return_ss_mm_hh
    ::Time.now.strftime('%H:%M:%S')
  end

  # ========================================================================= #
  # === MultimediaParadise.to_ts
  #
  # This will batch-convert all .mp4 files into .ts files.
  # ========================================================================= #
  def self.to_ts
    files = Dir['*.mp4']
    files.each {|this_file|
      cmd = 'ffmpeg -i '+this_file+' -c copy '\
            '-bsf:v h264_mp4toannexb -f mpegts '+
            this_file.sub(/#{File.extname(this_file)}$/,'')+
            '.ts'

      esystem cmd
    }
  end

  # ========================================================================= #
  # === MultimediaParadise.set_the_aspect_ratio_of_this_video
  #
  # This method will set the aspect-ratio of a given video file, via
  # ffmpeg. The commandline-flag for this is called -aspect.
  #
  # Original example:
  #
  #   ffmpeg -i foo.mp4 -aspect 16:9 bar.mp4
  #
  # Example for the MultimediaParadise namespace:
  #
  #   MultimediaParadise.set_the_aspect_ratio_of_this_video('foo.mpg')
  #
  # ========================================================================= #
  def self.set_the_aspect_ratio_of_this_video(
      this_video, use_this_aspect_ratio = '16:9'
    )
    extension_to_use = File.extname(this_video)
    cmd_to_use = 'ffmpeg -i '+this_video+' -aspect '+
                 use_this_aspect_ratio+' output_'+
                 this_video.sub(/#{extension_to_use}$/,'')+extension_to_use
    e
    e cmd_to_use
    e
    system cmd_to_use
  end

  # ========================================================================= #
  # === MultimediaParadise.extract_images_from_this_video_file
  #
  # This variant uses the default rate, which is 25 images per second.
  #
  # Note that the method is named incorrectly; in the long run I may
  # rename it (written in January 2020).
  #
  # Usage example:
  #
  #   MultimediaParadise.extract_images_from_this_video_file('foobar.mp4')
  #
  # ========================================================================= #
  def self.extract_images_from_this_video_file(*i)
    i.flatten.each {|entry|
      _extension_to_use = File.extname(entry) # No longer in use.
      _ = 'ffmpeg '\
          '-i '+entry+
          ' %08d.png' #OUTPUT_'+entry.sub(/#{extension_to_use}$/,'')+
          # extension_to_use+
      e
      esystem _
      e
    }
  end

  # ========================================================================= #
  # === MultimediaParadise.extract_one_image_per_second
  #
  # This method will extract one-image-per-second from a given
  # input video file.
  #
  # The first argument to this method should be the video file you
  # are working with, such as 'foobar.mp4'.
  # ========================================================================= #
  def self.extract_one_image_per_second(
      from_this_video_file = 'my_video.mpeg',
      frame_size_to_use    = '75x75'
    )
    extension_to_use = File.extname(from_this_video_file)
    _ = 'ffmpeg -i '+from_this_video_file+
        ' -r 1 '+ # ← Specify the video-frame here, aka "1 image per second".
        '-s '+frame_size_to_use+
        ' OUTPUT_'+from_this_video_file.sub(/#{extension_to_use}$/,'')+
        extension_to_use
    e
    esystem _
    e
  end

  # ========================================================================= #
  # === MultimediaParadise.convert_these_images_into_a_video
  #
  # This method can be used to "assemble" images into a video.
  # ========================================================================= #
  def self.convert_these_images_into_a_video(*images)
    _ = 'ffmpeg -f image2 -i '+images.join(' ').strip+' video.mpg'
    esystem _
  end

  # ========================================================================= #
  # === MultimediaParadise.high_quality_transcoding_of_this_video_file
  #
  # This method can be used for high-quality transcoding of a given
  # input video file.
  #
  # - C:V is the video codec of choice
  # - preset is the compression preset (in this case slow for higher
  #                                    quality compression)
  # - CRF is the Constant Rate Factor, which preserves an overall
  #   level of quality throughout the file by adjusting each frame's
  #   bitrate based on the given quality level. The higher the CRF,
  #   the lower the overall quality level.
  # - The video filter flag (-vf) is used to call FFMPEG's pre-bundled
  #   video filters.
  # - yadif (Yet Another Deinterlacing Filter) deinterlaces an interlaced
  #   input, as progressive video is not only easier to compress and most
  #   current computer monitors and televisions are progressive scan.
  # - When encoding with H.264/MPEG-4 AVC, the audio format used is AAC
  #   (Advanced Audio Coding); in order to enable FFMPEG's experimental,
  #   native AAC encoder.
  # - -strict -2 needs to be added to the command.
  # - An external library such as libfaac can also be used, and -strict
  #   -2 can be omitted.
  #
  # ========================================================================= #
  def self.high_quality_transcoding_of_this_video_file(this_video_file)
    output_file = 'OUTPUT_'+this_video_file.
                  sub(/#{File.extname(this_video_file)}$/,'')+'.mp4'
    _ = 'ffmpeg -i '+this_video_file+' -c:v libx264 -preset slow '\
        '-crf 18 -vf yadif -strict -2 '+
        output_file
    e
    esystem _
    e
  end; self.instance_eval { alias high_quality_transcoding high_quality_transcoding_of_this_video_file } # === MultimediaParadise.high_quality_transcoding

  # ========================================================================= #
  # === MultimediaParadise.identify_video
  #
  # Use this to identify a given video or audio file.
  #
  # Another way would be this:
  #
  #   mplayer -vo null -ao null -frames 0 -identify POP_America_TheLastUnicorn.mp3
  #
  # You will have to parse this lateron.
  # ========================================================================= #
  def self.identify_video(this_file)
    cmd = "ffmpeg -i #{this_file}"
    e
    e cmd
    e
    result = `{cmd}`
    return result
  end

  # ========================================================================= #
  # === MultimediaParadise.show_available_audio_formats
  #
  # Use this method if you want to show the available (registered)
  # audio formats.
  # ========================================================================= #
  def self.show_available_audio_formats
    e; AUDIO_FORMATS.each {|entry|
      e "  - #{entry}" # It is ok to prepend the '-' since this is for display purposes.
    }; e
  end

  # ========================================================================= #
  # === MultimediaParadise.show_artist
  #
  # Use this for .mp3 files, via ffmpeg. It will show the artist.
  #
  # Input should be a .mp3 file.
  #
  # Usage example from within ruby code:
  #
  #   MultimediaParadise.show_artist('/Depot/Audio/YoutubeMix_VivaldisFourSeasons.mp3')
  #   MultimediaParadise.show_artist('/Depot/Audio/Vivaldi_Concerto8.mp3')
  #
  # ========================================================================= #
  def self.show_artist(
      i, report_classname = true
    )
    result = MultimediaParadise.return_artist(i)
    if result.empty?
      opnn if report_classname
      e 'This file does not have any artist-tag.'
    else
      opnn if report_classname
      e "The artist is: #{simp(result)}"
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.return_artist
  # ========================================================================= #
  def self.return_artist(
      i,
      report_classname = true
    )
    i = i.to_s
    cmd =
      'ffprobe -loglevel error -show_entries format_tags=artist '\
      '-of default=noprint_wrappers=1:nokey=1 '+i
    result = `#{cmd}`.chomp
    return result
  end

  # ========================================================================= #
  # === MultimediaParadise.load_yaml
  # ========================================================================= #
  def self.load_yaml(i)
    YAML.load_file(i)
  end

  # ========================================================================= #
  # === MultimediaParadise.open_yaml_file
  #
  # This will open my video collection.
  # ========================================================================= #
  def self.open_yaml_file(
      i = FILE_VIDEO_COLLECTION
    )
    esystem "bluefish #{i}"
  end

  # ========================================================================= #
  # === MultimediaParadise.flip_video_to_the_left
  #
  # This toplevel-method can be used to flip a video to the left,
  # that is, to rotate it by -90 degrees.
  #
  # It will tap into ffmpeg for this task.
  # ========================================================================= #
  def self.flip_video_to_the_left(i)
    ffmpeg_flip_video_to_the_left(i)
  end

  # ========================================================================= #
  # === MultimediaParadise.flip_video_to_the_right
  #
  # This toplevel-method can be used to flip a video to the right,
  # that is, to rotate it by +90 degrees.
  # ========================================================================= #
  def self.flip_video_to_the_right(i)
    ffmpeg_flip_video_to_the_right(i)
  end

  # ========================================================================= #
  # === glob
  # ========================================================================= #
  def glob(i)
    Dir[i]
  end

  # ========================================================================= #
  # === video_collection?
  # ========================================================================= #
  def video_collection?
    ::MultimediaParadise.file_video_collection
  end

  # ========================================================================= #
  # === MultimediaParadise::PLAY_ZOOMED
  #
  # If this constant is set to true, then we will play in a zoomed way.
  # ========================================================================= #
  PLAY_ZOOMED = YAML.load_file(
    "#{PROJECT_BASE_DIRECTORY}configuration/play_zoomed.yml"
  )

  # ========================================================================= #
  # === MultimediaParadise.verbose_analyse_this_mp3_file_for_id3_tags
  # ========================================================================= #
  def self.verbose_analyse_this_mp3_file_for_id3_tags(this_mp3_file)
    require 'id3lib'
    if this_mp3_file.is_a? Array
      this_mp3_file.each {|entry|
        verbose_analyse_this_mp3_file_for_id3_tags(entry)
      }
    else
      dataset = ID3Lib::Tag.new(this_mp3_file)
      e "#{::Colours.rev}"\
        "The .mp3 file is called: #{sfancy(this_mp3_file)}"
      if dataset.empty?
        e 'Has no id3-tags!'
      else
        # e 'Has some id3-tags.' # <- Dont report this in this case, though.
        # ===================================================================== #
        # Iterate through these entries next.
        # ===================================================================== #
        dataset.each_entry {|inner_hash|
          value = inner_hash[:text].to_s
          case inner_hash[:id]
          # =================================================================== #
          # === :TIT2
          # =================================================================== #
          when :TIT2
            e 'The title is: '+
               Colours.steelblue(value) unless value.empty?
          # =================================================================== #
          # === :TCON
          #
          # The genre.
          # =================================================================== #
          when :TCON
            e 'The genre is: '+
               Colours.steelblue(value) unless value.empty?
          # =================================================================== #
          # === :TYER
          #
          # The year.
          # =================================================================== #
          when :TYER
            e 'The year is: '+
               Colours.steelblue(value) unless value.empty?
          # =================================================================== #
          # === :TPE1
          #
          # The artist.
          # =================================================================== #
          when :TPE1
            e 'The artist is: '+
               Colours.steelblue(value) unless value.empty?
          end
        }
        # pp dataset # <- for debugging.
      end
    end
  end

  # ========================================================================= #
  # === MultimediaParadise.analyze_this_directory_for_id3_tags
  # ========================================================================= #
  def self.analyze_this_directory_for_id3_tags(i = Dir.pwd)
    require 'id3lib'
    all_files = Dir["#{i}/*.mp3"]
    all_files.each {|this_mp3_file|
      verbose_analyse_this_mp3_file_for_id3_tags(this_mp3_file)
    }
  end

  # ========================================================================= #
  # === MultimediaParadise.compress_this_ogg_file
  #
  # This method will use a sampling rate of 22050 Hz. This should be
  # perfectly adequate for speech-related audio, but I also found it
  # to work well when compressing a tetris.ogg file (for the game
  # tetris).
  # ========================================================================= #
  def self.compress_this_ogg_file(i = 'foobar.ogg')
    _ = 'ffmpeg -i '+entry+' '\
        '-c:a libvorbis -ab 32k '\
        '-ar 22050'\
        ' OUTPUT_'+i
    e
    esystem _
    e
  end

  # ========================================================================= #
  # === MultimediaParadise.try_to_rename_kde_konsole_tab
  # ========================================================================= #
  def self.try_to_rename_kde_konsole_tab(
      new_title                         = '_',
      try_to_rename_the_kde_konsole_tab = TRY_TO_RENAME_THE_KDE_KONSOLE_TAB
    )
    if try_to_rename_the_kde_konsole_tab
      begin
        require 'roebe/requires/require_kde_konsole.rb'
      rescue LoadError
        e 'roebe/requires/require_kde_konsole.rb is not available'
      end
      if Object.const_defined?(:Roebe) and
         Roebe.respond_to?(:rename_konsole)
        Roebe.rename_konsole(new_title)
      end
    end
  end; self.instance_eval { alias rename_kde_konsole     try_to_rename_kde_konsole_tab } # === MultimediaParadise.rename_kde_konsole
       self.instance_eval { alias do_rename_konsole      try_to_rename_kde_konsole_tab } # === MultimediaParadise.do_rename_konsole
       self.instance_eval { alias rename_kde_konsole_tab try_to_rename_kde_konsole_tab } # === MultimediaParadise.rename_kde_konsole_tab

  # ========================================================================= #
  # === MultimediaParadise.crop_this_video
  #
  # This method can be used to crop a particular video file, via
  # ffmpeg.
  #
  # The raw ffmpeg-command would go like something like this:
  #
  #   ffmpeg -i input.mp4 -filter:v "crop=out_w:out_h:x:y" output.mp4
  #
  # These variables are:
  #
  #   → out_w is the width  of the output rectangle
  #   → out_h is the height of the output rectangle
  #   → x and y specify the top left corner of the output rectangle
  #   → output.mp4 is the output file
  #
  # ========================================================================= #
  def self.crop_this_video(
      i           = 'input.mp4',
      output_file = 'output.mp4'
    )
    [i].flatten.compact.each {|this_video_file|
      esystem 'ffmpeg -i '+
              this_video_file.to_s+
              ' -filter:v "crop=w:h:x:y" '+
              output_file.to_s
    }
  end

end

if __FILE__ == $PROGRAM_NAME
  include MultimediaParadise
  map_number_to_videofile
  puts 'The version of the MultimediaParadise project is: '+version?  
end # rb shared.rb