#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::CaptureScreen
#
# This class will capture a screen and store the output somewhere.
#
# Usage example:
#   MultimediaParadise::CaptureScreen.new
# =========================================================================== #
# require 'multimedia_paradise/video/capture_screen.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

class CaptureScreen < ::MultimediaParadise::Base # === MultimediaParadise::CaptureScreen.new

  # ========================================================================= #
  # === DEFAULT_INPUT
  # ========================================================================= #
  DEFAULT_INPUT = ' -i :0.0'

  # ========================================================================= #
  # === DEFAULT_STORE_IN_THIS_DIRECTORY
  #
  # This is currently hardcoded to the current working directory,
  # but you could also use another target directory.
  # ========================================================================= #
  DEFAULT_STORE_IN_THIS_DIRECTORY = :pwd

  # ========================================================================= #
  # === USE_THIS_SCREEN_RESOLUTION
  # ========================================================================= #
  USE_THIS_SCREEN_RESOLUTION = '1920x1080'

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = nil, # This variable is not in use, for now.
      run_already = true
    )
    register_sigint
    reset
    case i # Handle special instructions here. (case i)
    when :dont_run_yet
      run_already = false
      i = nil
    end
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @run_this_command
    # ======================================================================= #
    @run_this_command = 'ffmpeg' # This is the base command.
    set_framerate 25 # These are used for buildup.
    # add ' g -30' # What is this option? I have no idea.
    set_store_where DEFAULT_STORE_IN_THIS_DIRECTORY # Where to store the captured dataset.
  end

  # ========================================================================= #
  # === run_main_string
  # ========================================================================= #
  def run_main_string
    sanitize_main_string
    # ======================================================================= #
    # Next, show the command to the user, with some "padding".
    # ======================================================================= #
    e
    e main?
    e
    system main? # And run it finally.
  end

  # ========================================================================= #
  # === main_string?
  # ========================================================================= #
  def main_string?
    @run_this_command
  end; alias main? main_string? # === main?

  # ========================================================================= #
  # === size=                                                      (size tag)
  #
  # This is -video_size which is equal to -s option in ffmpeg.
  # ========================================================================= #
  def size=(i = '1024x768')
    add('-video_size '+i)
  end; alias video_size  size= # === video_size
       alias video_size= size= # === video_size=
       alias size        size= # === size

  # ========================================================================= #
  # === no_threads
  #
  # Don't use any threads.
  # ========================================================================= #
  def no_threads
    add '-threads 0'
  end

  # ========================================================================= #
  # === store_here=
  #
  # This determines where to store the captured video.
  # ========================================================================= #
  def store_here=(i)
    case i
    # ======================================================================= #
    # === :pwd
    # ======================================================================= #
    when :pwd
      i = return_pwd
    end
    if i.end_with? '/'
      i = i.dup if i.frozen?
      i << 'output.mp4' # Create a .mp4 file by default.
    end
    @store_where = i
  end; alias set_store_where store_here= # === set_store_where

  # ========================================================================= #
  # === store_where?
  # ========================================================================= #
  def store_where?
    @store_where
  end

  # ========================================================================= #
  # === add_audio_support
  #
  # Use this method if you wish quick support for audio.
  # ========================================================================= #
  def add_audio_support
    add ' -f alsa -ac 2 -i pulse'
  end

  # ========================================================================= #
  # === set_duration
  # ========================================================================= #
  def set_duration(i)
    add '-t '+i.to_s
  end

  # ========================================================================= #
  # === sameq
  # ========================================================================= #
  def sameq
    '' # This option never existed, hence we won't use it. We retain this for backwards compatibility however.
  end

  # ========================================================================= #
  # === add                                                         (add tag)
  #
  # This method will simply add to the main system command that will be
  # run.
  # ========================================================================= #
  def add(i)
    i = i.to_s unless i.is_a? String
    i = i.dup if i.frozen?
    i.strip! # We don't want leading or trailing ' '.
    # ======================================================================= #
    # Always prepend a ' ', unless it is already there.
    # ======================================================================= #
    i.prepend ' ' unless i.start_with? ' '
    @run_this_command = @run_this_command.dup if @run_this_command.frozen?
    @run_this_command << i
  end; alias append add # === append

  # ========================================================================= #
  # === resolution?
  # ========================================================================= #
  def resolution?
    USE_THIS_SCREEN_RESOLUTION
  end
  
  # ========================================================================= #
  # === video_codec
  #
  # Popular video codecs for ffmpeg are:
  #
  #   libx264
  #   ffv1
  #
  # ========================================================================= #
  def video_codec=(i = 'libx264')
    add " -vcodec #{i}"
  end; alias vcodec video_codec= # === vcodec

  # ========================================================================= #
  # === add_default_input
  # ========================================================================= #
  def add_default_input
    add DEFAULT_INPUT # This is usually ' -input :0.0'.
  end

  # ========================================================================= #
  # === add_default_format
  # ========================================================================= #
  def add_default_format
    add ' -f x11grab'
  end

  # ========================================================================= #
  # === add_default_framerate
  # ========================================================================= #
  def add_default_framerate
    add ' -framerate '+@framerate
  end

  # ========================================================================= #
  # === add_default_video_size
  # ========================================================================= #
  def add_default_video_size
    add ' -video_size '+resolution?
  end

  # ========================================================================= #
  # === set_framerate
  #
  # In ffmpeg, "-r" equals the "-framerate" option.
  # ========================================================================= #
  def set_framerate(i)
    i = i.to_s
    @framerate = i
  end; alias framerate= set_framerate # === framerate=
       alias framerate  set_framerate # === framerate

  # ========================================================================= #
  # === sanitize_main_string                                   (sanitize tag)
  # ========================================================================= #
  def sanitize_main_string
    @run_this_command.squeeze!(' ') if @run_this_command.include? '  '
    unless main?.include? store_where?
      append store_where?
    end
  end

  # ========================================================================= #
  # === overwrite_if_the_file_already_exists
  # ========================================================================= #
  def overwrite_if_the_file_already_exists
    add ' -y' # Yes, overwrite the file if it already exists.
  end

  # ========================================================================= #
  # === build_up_main_string
  # ========================================================================= #
  def build_up_main_string
    add_default_video_size # Add the default screen resolution here.
    add_default_framerate
    add_default_format
    overwrite_if_the_file_already_exists
    add_default_input
    add_default_audio
  end

  # ========================================================================= #
  # === add_default_audio
  # ========================================================================= #
  def add_default_audio
    add ' -f alsa -ac 2 -i hw:0'
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    build_up_main_string
    run_main_string
  end; alias start run # === start

end; end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::CaptureScreen.new(ARGV, :dont_run_yet)
end # capturescreen