#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::Video::CopyMissingVideoFiles
#
# This class will copy missing video files.
#
# Right now, the default target directory is hardcoded, but one
# day it should be more flexible.
#
# Usage examples:
#
#   MultimediaParadise::CopyMissingVideoFiles.new(ARGV)
#
# =========================================================================== #
# require 'multimedia_paradise/video/copy_missing_video_files.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

module Video

class CopyMissingVideoFiles < ::MultimediaParadise::Base # CopyMissingVideoFiles.new

  begin
    require 'mountpoints'
  rescue LoadError; end

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === LJUST
  # ========================================================================= #
  LJUST = 28

  # ========================================================================= #
  # === DEFAULT_TARGET_DIRECTORY
  # ========================================================================= #
  SUBDIR = '/Video/Realvids/' # This part will be appended.

  # ========================================================================= #
  # === DEFAULT_TARGET_DIRECTORY
  # ========================================================================= #
  DEFAULT_TARGET_DIRECTORY = ENV['USB1'].to_s+SUBDIR

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      target_directory = DEFAULT_TARGET_DIRECTORY,
      run_already      = true
    )
    register_sigint
    reset
    set_start_directory
    set_target_directory(
      target_directory
    )
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @array_copied_these_files
    # ======================================================================= #
    @array_copied_these_files = []
  end

  # ========================================================================= #
  # === set_start_directory
  #
  # Our starting directory.
  # ========================================================================= #
  def set_start_directory(i = return_pwd) # This comes before set_target_directory
    i << '/' unless i.end_with? '/'
    @start_directory = rds(i)
  end

  # ========================================================================= #
  # === set_target_directory
  #
  # We will copy the video files to this target directory.
  # ========================================================================= #
  def set_target_directory(
      i = obtain_target_directory
    )
    i = i.first if i.is_a? Array
    i = DEFAULT_TARGET_DIRECTORY if i.nil?
    i = rds(i)
    # ======================================================================= #
    # We must find an alternative if the current directory is the
    # same as the start-directory.
    # ======================================================================= #
    i = find_alternative_to(@start_directory) if i == @start_directory
    @target_directory = rds(i)
    if i.start_with? '--'
      menu(i)
    end
  end

  # ========================================================================= #
  # === find_alternative_to
  # ========================================================================= #
  def find_alternative_to(i)
    if i.include? 'USB2'
      i = i.gsub(/USB2/, 'USB1')
    elsif i.include? 'USB1'
      i = i.gsub(/USB1/, 'USB2')
    end
    return i
  end

  # ========================================================================= #
  # === menu                                                       (menu tag)
  #
  # This is the commandline-menu interface for this class.
  # ========================================================================= #
  def menu(i)
    if i.is_a? Array
      i.each {|entry| menu(entry) }
    else
      case i
      # ===================================================================== #
      # === copy_missing_video_files --help
      # ===================================================================== #
      when /^-?-?help$/i
        show_help(:then_exit)
      end
    end
  end

  # ========================================================================= #
  # === obtain_target_directory
  # ========================================================================= #
  def obtain_target_directory
    if Object.const_defined? :Mountpoints
      mounted_devices = Mountpoints[].map {|entry| entry + SUBDIR }
      mounted_devices.select! {|entry| Dir.exist? entry }
      result = mounted_devices.first # Should be fine.
    else
      result = DEFAULT_TARGET_DIRECTORY
    end
    result
  end

  # ========================================================================= #
  # === target?
  # ========================================================================= #
  def target?
    @target_directory
  end

  # ========================================================================= #
  # === from?
  # ========================================================================= #
  def from?
    @start_directory
  end

  # ========================================================================= #
  # === reject_non_videos
  #
  # Here we will get rid of .srt files.
  # ========================================================================= #
  def reject_non_videos
    @files_from_start_directory.reject!  {|entry| entry.include? '.srt'}
    @files_from_target_directory.reject! {|entry| entry.include? '.srt'}
  end

  # ========================================================================= #
  # === compare_the_two_directories
  # ========================================================================= #
  def compare_the_two_directories
    @files_from_target_directory.each {|entry|
      result = @files_from_start_directory.map {|e|
        File.basename(e)
      }.include? File.basename(entry)
      if result # is included
      else
        opnn { :no_colon }
        e
        e
        e swarn('The following video file is NOT '+
                'included: '+sfile(File.basename(entry)))
        opnn
        e "It will thus be copied to the current directory, "\
          "from the"
        opnn; e "location `#{sfile(entry)}`."
        copy_video_file(entry)
      end
    }
  end

  # ========================================================================= #
  # === copy_video_file
  # ========================================================================= #
  def copy_video_file(i)
    @array_copied_these_files << i
    if File.exist? i
      FileUtils.cp(i, Dir.pwd)
    else
      opnn; e 'Can not copy '+sfile(i)+' as it does not exist.'
    end
  end

  # ========================================================================= #
  # === report_finished_tasks
  # ========================================================================= #
  def report_finished_tasks
    if @array_copied_these_files.empty?
      opnn; e 'No file was copied. This could mean that everything is '
      opnn; e 'already existing on the target directory. '+
               steelblue('\o/')
    else
      opnn; e 'Finished!'
      e
      opnn; e 'The following files were copied:'
      e
      @array_copied_these_files.sort.each {|file|
        e "  #{sfile(File.basename(file))}"
      }
      e
    end
  end

  # ========================================================================= #
  # === show_help                                                  (help tag)
  # ========================================================================= #
  def show_help(optional_argument = :do_nothing)
    e
    e 'This class can copy video files from another directory to the'
    e 'target directory.'
    e
    case optional_argument
    when :then_exit
      exit
    end
  end

  # ========================================================================= #
  # === get_available_video_files
  # ========================================================================= #
  def get_available_video_files
    cliner
    opnn; e 'Now obtaining available video files.'
    opnn; e 'We will use these two locations:'
    e
    e '  - '+sfancy(from?.ljust(LJUST))+' # Start Location.'
    e '  - '+sfancy(target?.ljust(LJUST))+' # Target Location.'
    e
    opnn; e 'Note that we will simply copy missing files from <- -> to.'
    cliner
    @files_from_start_directory  = Dir["#{from?}*"].sort.select {|entry| File.file?(entry) }
    @files_from_target_directory = Dir["#{target?}*"].sort.select {|entry| File.file?(entry) }
    reject_non_videos
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    get_available_video_files
    compare_the_two_directories
    report_finished_tasks
  end

end; end

# =========================================================================== #
# === MultimediaParadise.copy_missing_video_files
# =========================================================================== #
def self.copy_missing_video_files(i = ARGV.first)
  MultimediaParadise::Video::CopyMissingVideoFiles.new(i)
end

end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::Video::CopyMissingVideoFiles.new(ARGV)
end # cmissing /home/x/video/Realvids/