#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::FindVideo
#
# Use this class to find a video from my yaml collection.
#
# The input can be a String such as 'gladiator'.
#
# To do this from the commandline, try:
#
#   findvid aff
#   findvid glad
#
# =========================================================================== #
# require 'multimedia_paradise/video/find_video/find_video.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

class FindVideo < Base # === MultimediaParadise::FindVideo

  require 'multimedia_paradise/video/find_video/constants.rb'

  # ========================================================================= #
  # === initialize
  #
  # The first argument is the search term.
  # ========================================================================= #
  def initialize(
      search_term             = '*',
      be_silent_or_be_verbose = :be_verbose
    )
    reset
    set_search_term(
      search_term
    )
    set_verbosity_level(
      be_silent_or_be_verbose
    )
    # ======================================================================= #
    # === Handle blocks
    # ======================================================================= #
    if block_given?
      yielded = yield
      if yielded.is_a? Symbol
        case yielded
        when :be_quiet
          @be_verbose = false
        end
      end
    end
    check_on_search_term
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @results
    # ======================================================================= #
    @results = nil
    # ======================================================================= #
    # === @debug
    # ======================================================================= #
    @debug = false
    # ======================================================================= #
    # === @hash_video_collection_name
    # ======================================================================= #
    @hash_video_collection_name = {}
    # ======================================================================= #
    # === @video_collection
    # ======================================================================= #
    @video_collection = VIDEO_COLLECTION
    # ======================================================================= #
    # === @candidates
    # ======================================================================= #
    @candidates = []
  end

  # ========================================================================= #
  # === candidates?
  # ========================================================================= #
  def candidates?
    @candidates
  end; alias candidates candidates? # === candidates

  # ========================================================================= #
  # === set_search_term
  #
  # We will assign the search term here in this method.
  # ========================================================================= #
  def set_search_term(i)
    i = i.first if i.is_a? Array
    i = i.to_s
    i = i.strip
    @search_term = i
  end

  # ========================================================================= #
  # === set_debug
  # ========================================================================= #
  def set_debug(i = false)
    @debug = i
  end; alias enable_debug set_debug # === enable_debug

  # ========================================================================= #
  # === search_term?
  # ========================================================================= #
  def search_term?
    @search_term
  end

  # ========================================================================= #
  # === results
  #
  # Returns the positional values in Array format. This can then be used to
  # find the corresponding match when playing a video.
  # 
  # Example:
  #
  #   [242, 250, 266]
  #
  # ========================================================================= #
  def results?
    @results
  end; alias results results? # === results
    
  # ========================================================================= #
  # === reset_candidates
  #
  # Reset the candidates here again.
  # ========================================================================= #
  def reset_candidates
    @candidates = []
  end

  # ========================================================================= #
  # === load_video_yaml_file
  # ========================================================================= #
  def load_video_yaml_file(
      i = video_collection?
    )
    dataset = load_yaml(i)
    dataset.each { |t|
      new_number = t[0] # This is the number.
      new_title  = t[1]['title']
      if new_title.nil?
        opnn; e "The position at #{new_number} is nil."
        opnn; e 'The assumed dataset was:'
        pp t
      end
      # We must be careful - there may be entries with the same title.
      if @hash_video_collection_name.has_key? new_title
        warn_about_this_old_title(new_title)
      else
        @hash_video_collection_name[new_title] = new_number
      end
    }
  end

  # ========================================================================= #
  # === warn_about_this_old_title
  # ========================================================================= #
  def warn_about_this_old_title(new_title)
    opnn; e 'Warning! We already have an old title called '+
            sfancy(new_title.to_s)+'.'
    pp @hash_video_collection_name.invert[new_title]
  end

  # ========================================================================= #
  # === ef
  # ========================================================================= #
  def ef(i)
    e i, false
  end

  # ========================================================================= #
  # === raise_and_exit
  # ========================================================================= #
  def raise_and_exit(i)
    e i; exit
  end

  # ========================================================================= #
  # === video_collection?
  # ========================================================================= #
  def video_collection?
    @video_collection
  end

  # ========================================================================= #
  # === set_results
  # ========================================================================= #
  def set_results(i = nil)
    if i.nil?
      raise_and_exit 'Please supply a search term. Thanks.'
    end
    @results = i
  end

  # ========================================================================= #
  # === check_on_search_term
  #
  # menu() is an alias to this method.
  # ========================================================================= #
  def check_on_search_term(i = search_term?)
    case i # case tag.
    when 'OPEN'
      open_yaml_file; exit
    when 'list?','?','help','HELP','--help'
      show_help_then_exit
    end
  end; alias menu check_on_search_term # === menu

  # ========================================================================= #
  # === show_help_then_exit
  # ========================================================================= #
  def show_help_then_exit
    show_available_videos_in_raw_format
    show_commandline_help
    exit
  end

  # ========================================================================= #
  # === show_commandline_help
  #
  # In order to invoke this, do:
  #
  #   find_video --help
  #
  # ========================================================================= #
  def show_commandline_help
    _ = MultimediaParadise.location_of_yaml_file_with_registered_videofiles
    opnn; e 'This script will attempt to find a given video, if it was registered'
    opnn; e 'in the yaml file.'
    opnn; e 'This yaml file can be found here:'
    opnn; e '  '+sfile(_)
    opnn; e 'You can also try to do a partial match. For instance, searching for'
    opnn; e '"con" will try to find "Conan" but also "L.A. Confidential".'
  end

  # ========================================================================= #
  # === report_candidates
  #
  # Output the result through this method. Since as of December 2013, we
  # will pad the result.
  # ========================================================================= #
  def report_candidates(hash)
    reset_candidates # Reset the candidates here again.
    if be_verbose?
      e N+sfancy('find_video.rb: ')+'Results for the search '\
        'term `'+simp(search_term?.to_s)+'` are as follows: ✅'
      cliner
    end
    set_results(hash.values)
    longest_video_title = hash.keys.map(&:size).max
    # e N+sfancy('find_video.rb: ')+'The longest title is: '+
    #   sfancy(longest_video_title)
    hash.each { |the_video_title, position_of_the_video_file|
      position_of_the_video_file = '%03d' % position_of_the_video_file 
      _ = '%-'+longest_video_title.to_s+'s'
      the_video_title = _ % the_video_title
      if be_verbose?
        e sfancy('  `'+the_video_title)+'` can be found at position `'+
          sfancy(position_of_the_video_file.to_s)+'`.'
      end
      # ===================================================================== #
      # We store more information in the Array @candidates,
      # including the Title.
      # ===================================================================== #
      @candidates << [the_video_title.strip, position_of_the_video_file]
    }
    cliner if be_verbose?
    return @candidates
  end

  # ========================================================================= #
  # === MultimediaParadise::FindVideo[]
  #
  # Usage example would be like:
  #
  #   MultimediaParadise::FindVideo['ninja']
  #   MultimediaParadise::FindVideo['ninja', :flatten_result]
  #
  # ========================================================================= #
  def self.[](i, flatten_result = false)
    case flatten_result
    when :do_flatten, :flatten, :flatten_result
      flatten_result = true
    end
    _ = new(i)
    _.be_silent
    _.find_video
    _ = _.candidates?
    _.flatten! if flatten_result
    return _
  end

  # ========================================================================= #
  # === show_available_videos_in_raw_format
  #
  # The purpose of this method is to show the raw dataset, via pp.
  # ========================================================================= #
  def show_available_videos_in_raw_format
    load_video_yaml_file
    pp @hash_video_collection_name.sort_by {|key, value| value }
  end

  # ========================================================================= #
  # === set_verbosity_level
  # ========================================================================= #
  def set_verbosity_level(i)
    case i.to_s
    when 'be_verbose'
      i = true
    when 'be_silent'
      i = false
    end
    @be_verbose = i
  end

  # ========================================================================= #
  # === find_video                                                 (find tag)
  #
  # I use this method to find a specific video. The first example should
  # be a string, the actual search term in question.
  #
  # Usage examples:
  #
  #   find_video Amelie
  #   find_video Tanz der Vampire
  #
  # ========================================================================= #
  def find_video(
      search_term = search_term?,
      be_verbose  = be_verbose?
    )
    if @debug
      opnn; e '(Debug Mode) The search term to find_video() method '\
              'was '+sfancy(search_term)+'.'
    end
    load_video_yaml_file
    # ======================================================================= #
    # Select all possible candidates in that hash next. The Hash will have
    # key value structures that look like this:
    #
    #   "Ice Age"=>58
    #
    # ======================================================================= #
    if (search_term =~ /^\d+$/) # If only numbers
      hash_candidates = @hash_video_collection_name.select { |key, value|
        value.to_i == search_term.to_i
      }
    else
      hash_candidates = @hash_video_collection_name.reject { |key, value|
        begin
          # =================================================================== #
          # Next match the key onto the given search term.
          # =================================================================== #
          ! (key =~ /#{search_term}/) # The /i was removed in June 2021.
        rescue Exception => error
          pp error
        end
      }
    end
    if !hash_candidates.empty?
      return report_candidates(hash_candidates)
    else
      # ===================================================================== #
      # Ok, we did not find it up to here, but perhaps we try one last 
      # strategy here, and remove all ' ' in the search string. We could 
      # then find another result. And we will also downcase the keys. But if 
      # we then still do not find a result, we report to the user.
      #
      # Since as of June 2021 we will no longer downcase, because
      # "W" is different to "w", e. g. "Watchmen" versus "watchmen".
      # ===================================================================== #
      hash_candidates = @hash_video_collection_name.reject { |key, value|
        if key
          key = ensure_main_encoding(key)
          key = key.dup if key.frozen?
          key.delete!(' ') # .downcase # .downcase was commented out in June 2021.
        end
        search_term = search_term.downcase
        ! (key =~ /#{search_term}/) # The /i part was removed in June 2021.
      }
      if hash_candidates.empty?
        if be_verbose
          opnn; ewarn 'No match for `'+sfancy(search_term)+
                       swarn('` could be found.')
        end
        e @hash_video_collection_name.keys.join ', ' if @debug
        return false
      else # else, really report that we found none.
        return report_candidates(hash_candidates)
      end
    end
  end; alias try_to_find find_video # === try_to_find
       alias find        find_video # === find

end

# =========================================================================== #
# === MultimediaParadise.find_this_video
#
# Usage examples:
#
#   result = MultimediaParadise.find_this_video('glad')
#   result = MultimediaParadise.find_this_video('24')
#
# =========================================================================== #
def self.find_this_video(
    i,
    be_silent_or_be_verbose = :be_verbose
  )
  _ = MultimediaParadise::FindVideo.new(i, be_silent_or_be_verbose)
  return _.find_video
end

# =========================================================================== #
# === MultimediaParadise.return_filename_of_this_video_file
#
# Usage example:
#
#   result = MultimediaParadise.return_filename_of_this_video_file('24') # => "024_Convoy"
#
# =========================================================================== #
def self.return_filename_of_this_video_file(i)
  _ = MultimediaParadise::FindVideo.new(i, :be_silent)
  array = _.find_video.flatten
  return array.last+'_'+array.first
end

# =========================================================================== #
# === MultimediaParadise.has_this_video_file?
#
# This method can be used to query whether there is a video registered
# with the given name.
#
# This method will return a boolean value, true/false.
#
#  o It will return true  if the video file is registered.
#  o It will return false if the video file is NOT registered.
#
# Usage examples:
#
#   result_has_this_video_file = MultimediaParadise.has_this_video_file?('Glad')
#   result_has_this_video_file = MultimediaParadise.has_this_video_file?('Gladd')
#
# =========================================================================== #
def self.has_this_video_file?(this_file)
  result = MultimediaParadise.find_this_video(this_file, :be_silent)
  if result.is_a? Hash
    if result.empty?
      result = false
    else
      result = true
    end
  elsif result.is_a? Array
    result = true unless result.empty?
  end
  return result
end

end

if __FILE__ == $PROGRAM_NAME
  require 'colours/colours_e_autoinclude.rb'
  arguments = ARGV.join(' ').strip
  _ = MultimediaParadise::FindVideo.new(arguments)
  _.enable_debug
  # _.report_missing_videos # Do we need this at all?
  _.find_video
  unless arguments.empty?
    puts 'The positions of the video files were: '+
         sfancy(_.results.inspect)
  end
end # fvid amelie
    # fvid naked
    # fvid 55
    # findvid blade runner
    # fvid dark knight
    # fvid darkknight
    # fvid pink_panther