#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::Video::MissingVideoFiles
#
# This class can report locally missing videofiles. It was written in
# August 2020 to replace the older class called FindMissingVideoFiles,
# which was not working properly anymore. All the old major functionality
# of that class has been retained.
#
# Usage example:
#
#   MultimediaParadise::Video::MissingVideoFiles.new
#
# =========================================================================== #
# require 'multimedia_paradise/video/missing_video_files/missing_video_files.rb'
# MultimediaParadise::Video::MissingVideoFiles.new(ARGV)
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

module Video

class MissingVideoFiles < ::MultimediaParadise::Base # === MultimediaParadise::Video::FindMissingVideoFiles

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      commandline_arguments = ARGV,
      run_already           = true
    )
    reset
    set_commandline_arguments(
      commandline_arguments
    )
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @work_on_this_directory
    # ======================================================================= #
    @work_on_this_directory = return_pwd
    # ======================================================================= #
    # === @video_files
    # ======================================================================= #
    @video_files = []
    # ======================================================================= #
    # === @found_n_missing_files
    #
    # This variable will keep track as to how many missing local video
    # files have been "found".
    # ======================================================================= #
    @found_n_missing_files = 0
    # ======================================================================= #
    # === @array_missing_videofiles
    #
    # This Array will store all missing (local) videofiles.
    # ======================================================================= #
    @array_missing_videofiles = []
  end

  # ========================================================================= #
  # === determine_all_local_videos
  # ========================================================================= #
  def determine_all_local_videos
    @video_files = obtain_filelisting_from_this_directory
  end

  # ========================================================================= #
  # === obtain_filelisting_from_this_directory
  #
  # Note that this method will return a sorted Array.
  # ========================================================================= #
  def obtain_filelisting_from_this_directory(
      i = @work_on_this_directory
    )
    Dir["#{i}**"].select {|entry|
      is_this_a_video_file?(entry)
    }.sort
  end; alias update_filelisting obtain_filelisting_from_this_directory # === update_filelisting
       alias update             obtain_filelisting_from_this_directory # === update (update tag)

  # ========================================================================= #
  # === video_files?
  # ========================================================================= #
  def video_files?
    @video_files
  end; alias filelisting? video_files? # === filelisting?

  # ========================================================================= #
  # === set_work_on_this_directory
  #
  # By default, the current working directory will be used (Dir.pwd, or
  # more accurately, the String returned by return_pwd()).
  # ========================================================================= #
  def set_work_on_this_directory(i = return_pwd)
    i = i.dup if i.frozen?
    i << '/' unless i.end_with? '/'
    @work_on_this_directory = i
  end; alias set_target_dir set_work_on_this_directory # === set_target_dir
       alias set_dir        set_work_on_this_directory # === set_dir

  # ========================================================================= #
  # === work_on_this_directory?
  # ========================================================================= #
  def work_on_this_directory?
    @work_on_this_directory
  end; alias target_dir? work_on_this_directory? # === target_dir?
       alias which_dir?  work_on_this_directory? # === which_dir?

  # ========================================================================= #
  # === show_help                                                  (help tag)
  # ========================================================================= #
  def show_help(
      shall_we_exit = false
    )
    shall_we_exit = true if shall_we_exit == :then_exit
    opnn
    e
    e
    e 'This class tries to show which video files are missing (locally).'
    e
    e 'A very few options are available to control '\
      'the behaviour of this class:'
    e
    ecomment ' --dir=TARGET # use this specific directory to search '\
            'for missing video files'
    e
    exit if shall_we_exit
  end

  # ========================================================================= #
  # === menu                                                       (menu tag)
  # ========================================================================= #
  def menu(
      i = commandline_arguments?
    )
    if i.is_a? Array
      i.each {|entry| menu(entry) }
    else
      case i
      # ===================================================================== #
      # === missingvideo --dir=TARGET
      #
      # Example:
      #
      #   missingvideo --dir=/tmp
      #
      # ===================================================================== #
      when /^-?-?dir=(.+)$/i
        set_target_dir($1.to_s.dup)
      # ===================================================================== #
      # === pwd
      # ===================================================================== #
      when /^-?-?pwd$/i
        set_target_dir(return_pwd)
      # ===================================================================== #
      # === missingvideo --help
      # ===================================================================== #
      when /^-?-?help$/i
        show_help :then_exit
      else
        if File.exist?(i) and File.directory?(i) # If it is a directory.
          set_dir(i)
        end
      end
    end
  end

  # ========================================================================= #
  # === determine_the_yaml_dataset
  # ========================================================================= #
  def determine_the_yaml_dataset
    @dataset = MultimediaParadise.return_all_video_files
  end

  # ========================================================================= #
  # === iterate_over_the_dataset_from_the_file_video_collection_yml
  # ========================================================================= #
  def iterate_over_the_dataset_from_the_file_video_collection_yml
    # ======================================================================= #
    # The next variable has just the filename, without any '/' characters.
    # ======================================================================= #
    shortened_video_files = @video_files.map {|entry| File.basename(entry) }
    determine_the_yaml_dataset
    @dataset.each_pair {|key, hash_dataset|
      padded_key = '%03d' % key.to_i
      if shortened_video_files.any? {|entry|
          entry.start_with? padded_key
        } # pass through in this case, as all is fine.
      else
        @found_n_missing_files += 1
        e crimson('Missing video-file')+' at position '+steelblue(key)
        local_path = target_dir?+'230_'+hash_dataset['title'].tr(' ','_')+
                     '.mp4'
        e '  The local path would be: '+
          sfile(local_path)
        @array_missing_videofiles << local_path
      end
    }
  end

  # ========================================================================= #
  # === report_happy_message_if_no_local_video_file_is_missing
  # ========================================================================= #
  def report_happy_message_if_no_local_video_file_is_missing
    if @found_n_missing_files == 0
      e 'Everything is fine - no local videofile is missing. This is good!'
    end
  end

  # ========================================================================= #
  # === notify_the_user_as_to_which_directory_will_be_processed
  #
  # Notify the user as to which directory will be scanned for
  # locally existing videofiles.
  # ========================================================================= #
  def notify_the_user_as_to_which_directory_will_be_processed
    opnn; e "Videofiles will be fetched from #{sdir(which_dir?)}."
    opnn; e 'This dataset was generated from the file '+
            '`'+sfile(MultimediaParadise.yaml_file?)+'`.'
  end

  # ========================================================================= #
  # === report_missing_videofiles
  #
  # This method will report which videofiles are missing.
  # ========================================================================= #
  def report_missing_videofiles
    unless @array_missing_videofiles.empty?
      e
      opnn; e 'The following videofiles have been determined to be missing:'
      e
      @array_missing_videofiles.each {|original_entry|
        if Object.const_defined? :Colours and use_colours?
          entry  = COLOURS.limegreen(original_entry.to_s+')')
          result = COLOURS.lavender(result.to_s)
        else
          entry = sfancy(original_entry.to_s+')')
        end
        e entry+' '+result.to_s
      }
    }
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    menu
    notify_the_user_as_to_which_directory_will_be_processed
    determine_all_local_videos
    iterate_over_the_dataset_from_the_file_video_collection_yml
    report_happy_message_if_no_local_video_file_is_missing
    report_missing_videofiles
  end

  # ========================================================================= #
  # === MultimediaParadise::Video::MissingVideoFiles[]
  # ========================================================================= #
  def self.[](i = '')
    new(i)
  end

end; end; end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::Video::MissingVideoFiles.new(ARGV)
end # missingvideofiles