#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::Video::MovieSearcher
#
# This class allows you to remotely search for movies, and return
# information based on this. We will refer to omdbapi for this.
#
# Usage example:
#
#   MultimediaParadise::Video::MovieSearcher.new(ARGV)
#
# =========================================================================== #
# require 'multimedia_paradise/video/movie_searcher.rb'
# MultimediaParadise::Video::MovieSearcher.new(ARGV)
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

module Video

class MovieSearcher < ::MultimediaParadise::Base # === MultimediaParadise::Video::MovieSearcher

  alias ee print

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === API_KEY
  #
  # The following defines a public API key. As many people may use this,
  # it is recommend to not over-use it. So please limit your search ratio
  # to only when you really need some information quickly, on the
  # commandline.
  #
  # Note that you can overrule this if you do set an environment
  # key called OMDBAPI_API_KEY.
  # ========================================================================= #
  if ENV['OMDBAPI_API_KEY']
    API_KEY = ENV['OMDBAPI_API_KEY'].to_s.dup
  else # else use a hardcoded default.
    API_KEY = '946f500a'
  end

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = nil,
      run_already = true
    )
    reset
    set_name_of_the_movie(i)
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @width_limit
    # ======================================================================= #
    @width_limit = 78
    # ======================================================================= #
    # === @main_string
    # ======================================================================= #
    @main_string = ''.dup
    # ======================================================================= #
    # === @api_key
    # ======================================================================= #
    @api_key = API_KEY
  end

  # ========================================================================= #
  # === api_key?
  # ========================================================================= #
  def api_key?
    @api_key
  end

  # ========================================================================= #
  # === set_name_of_the_movie
  # ========================================================================= #
  def set_name_of_the_movie(i = '')
    i = i.join(' ').strip if i.is_a? Array
    i = i.to_s.dup.chomp.strip
    # ======================================================================= #
    # We may also do some sanitizing. First, get rid of the file extension
    # if the input includes a '.' token.
    # ======================================================================= #
    if i.include? '.'
      i.sub!(/#{File.extname(i)}/,'')
    end
    if i.include?('_') and (i =~ /^\d+/)
      use_this_regex = /\d{1,3}_(.+)/ # See: http://rubular.com/r/9fsrGiTQi1
      i =~ use_this_regex
      i = $1.to_s.dup if $1
    end
    @name_of_the_movie = i
  end

  # ========================================================================= #
  # === name_of_the_movie?
  # ========================================================================= #
  def name_of_the_movie?
    @name_of_the_movie
  end; alias movie_name? name_of_the_movie? # === movie_name?

  # ========================================================================= #
  # === wrap                                                       (wrap tag)
  #
  # A word-wrapping method.
  # ========================================================================= #
  def wrap(i, width_limit = @width_limit)
    i.gsub(/(.{1,#{width_limit}})(\s+|\Z)/, "\\1\n| ")
  end

  # ========================================================================= #
  # === display_main_string
  # ========================================================================= #
  def display_main_string
    e @main_string
  end

  # ========================================================================= #
  # === report_that_no_movie_has_been_found_with_this_name
  # ========================================================================= #
  def report_that_no_movie_has_been_found_with_this_name(
      i = movie_name?
    )
    e
    opnn; e "No movie has been found with this name (#{sfancy(i)})."
    e
  end

  # ========================================================================= #
  # === cliner
  #
  # If input is provided, then we will use this embedded into the line.
  # ========================================================================= #
  def cliner(optional_input = nil)
    result = '=' * @width_limit
    if optional_input
      _ = " #{optional_input} "
      # ===================================================================== #
      # Also colourize the title there.
      # ===================================================================== #
      result[1,_.size] = palegoldenrod(_)
    end
    return result+N
  end

  # ========================================================================= #
  # === determine_whether_rest_client_is_available_and_exit_if_it_is_unavailable
  # ========================================================================= #
  def determine_whether_rest_client_is_available_and_exit_if_it_is_unavailable

    begin
      require 'rest-client'
    rescue LoadError
      unless Object.const_defined? :RestClient
        opnn; e "The gem #{slateblue('rest-client')}"\
                " appears to be missing. Please install it"
        opnn; e 'first, as this class depends on the namespace '\
                'RestClient.'
        exit 1
      end
    end

    begin
      require 'json'
    rescue LoadError
      unless Object.const_defined? :Json
        opnn; e "The gem #{slateblue('json')}"\
                " appears to be missing. Please install it"
        opnn; e 'first, as this class depends on json indirectly.'
        exit 1
      end
    end

  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    determine_whether_rest_client_is_available_and_exit_if_it_is_unavailable
    # ======================================================================= #
    # First, denote the names to end the program early.
    # ======================================================================= #
    _ = movie_name? # Pointer to the movie name.
    case _ # case tag
    when 'quit','exit'
      exit 0
    else
      url = "http://www.omdbapi.com/?t=#{_}&apikey=#{api_key?}"
      if debug?
        opnn; e 'Now working on the URL `'+sfancy(url)+'`.'
      end
      # ===================================================================== #
      # The constructed URl would look like this:
      #   http://www.omdbapi.com/?t=Alien&apikey=946f500a
      # ===================================================================== #
      response = RestClient.get(url)
      # ===================================================================== #
      # Next, parse the response via JSON.
      # ===================================================================== #
      info = JSON.parse(response)
      # ===================================================================== #
      # Handle exceptions, in case that we obtain an invalid response.
      # ===================================================================== #
      case info['Response'].to_s
      when 'False'
        report_that_no_movie_has_been_found_with_this_name(_)
        exit 0
      else
        # =================================================================== #
        # We have to rescue in the event that no tomato score has been
        # supplied. Additionally, we will word wrap to the plot content
        # and the actors of said movie.
        #
        # Commandline examples:
        #
        #   commandline_imdb Toy Story
        #
        # =================================================================== #
        begin
          title = info['Title']
          year  = info['Year']
          score = info['Ratings'][1]['Value']
        rescue
          score = 'No Score Found'
        end
        indent_to_use = 65
        rated    = info['Rated']
        genre    = info['Genre']
        director = info['Director']
        actors   = wrap(info['Actors'], indent_to_use)
        plot_unformatted = info['Plot']
        plot     = wrap(plot_unformatted, indent_to_use)
        @main_string << "#{N}" << cliner('Movie')
        @main_string << "| #{mediumorchid('Title')}:    #{title}"+N
        @main_string << "| #{mediumpurple('Year')}:     #{year.rjust(5)}"+N
        @main_string << "| #{skyblue('Tomato')}:   #{score.rjust(5)}"+N
        @main_string << "| #{royalblue('Rated')}:    #{rated.center(5)}"+N
        @main_string << "| #{palegreen('Genre')}:    #{genre}"+N
        @main_string << "| #{indianred('Director')}: #{director}"+N
        @main_string << "| #{springgreen('Actors')}:   #{actors}"+N
        # Add the Plot subsection next.
        # =================================================================== #
        @main_string << "| #{seagreen('Plot')}:     #{plot}"+N
        # =================================================================== #
        @main_string << cliner+N
        display_main_string
      end
    end
  end

  # ========================================================================= #
  # === MultimediaParadise::Video::MovieSearcher[]
  # ========================================================================= #
  def self.[](i = '')
    new(i)
  end

end; end; end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::Video::MovieSearcher.new(ARGV)
end # moviesearcher The naked gun