#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::MplayerWrapper
#
# This class simply wraps over mplayer, so that we can use it to play
# a bunch of Multimedia files. It is intented as a "drop-in" replacement
# for mplayer-related actions.
#
# Usage example:
#
#   MultimediaParadise::MplayerWrapper.new
#
# ============================================================================ #
# require 'multimedia_paradise/video/mplayer_wrapper.rb'
# MultimediaParadise::MplayerWrapper.new
# ============================================================================ #
require 'multimedia_paradise/base/base.rb'
require 'multimedia_paradise/toplevel_methods/return_path_to_random_simpsons_video_file.rb'

module MultimediaParadise

class MplayerWrapper < Base # === MultimediaParadise::MplayerWrapper

  require 'multimedia_paradise/audio/file_duration/file_duration.rb'
  require 'multimedia_paradise/video/the_simpsons/good_the_simpsons_episodes.rb'

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === ALSA_OPTIONS
  # ========================================================================= #
  ALSA_OPTIONS = '-ao alsa '

  # ========================================================================= #
  # === VO_X11
  # ========================================================================= #
  VO_X11 = '-vo x11' # More conservative option for mplayer.

  # ========================================================================= #
  # === APPEND_THESE_OPTIONS_TO_MPLAYER
  # ========================================================================= #
  APPEND_THESE_OPTIONS_TO_MPLAYER = ' '+VO_X11 # +ALSA_OPTIONS # Simply append these options.

  # ========================================================================= #
  # === USE_THIS_MULTIMEDIA_PLAYER
  # ========================================================================= #
  # USE_THIS_MULTIMEDIA_PLAYER = 'mpv' # Can also be mplayer.
  USE_THIS_MULTIMEDIA_PLAYER = ::MultimediaParadise.use_this_player # 'mplayer'

  # ========================================================================= #
  # === STORE_WHERE
  #
  # cat /root/mplayer_favourite_files.yml
  # ========================================================================= #
  if Dir.exist? '/Depot/Information/'
    STORE_WHERE = '/Depot/Information/mplayer_favourite_files.yml'
  else
    STORE_WHERE = "#{HOME_DIR}mplayer_favourite_files.yml"
  end

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      commandline_arguments = nil,
      run_already           = true
    )
    register_sigint
    reset
    set_commandline_arguments(
      commandline_arguments
    )
    run if run_already
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @play_these_multimedia_files
    # ======================================================================= #
    @play_these_multimedia_files = [] # Is an array. In fact, it must be always an array.
    # ======================================================================= #
    # === @use_this_multimedia_player
    # ======================================================================= #
    @use_this_multimedia_player = USE_THIS_MULTIMEDIA_PLAYER
    # ======================================================================= #
    # === @store_where
    # ======================================================================= #
    @store_where = STORE_WHERE
    # ======================================================================= #
    # === @append_this
    # ======================================================================= #
    @append_this = ''
    # ======================================================================= #
    # === @play_zoomed
    # ======================================================================= #
    @play_zoomed = PLAY_ZOOMED
    # ======================================================================= #
    # === @disable_video
    # ======================================================================= #
    @disable_video = false
    # ======================================================================= #
    # === @use_mplayer_colours
    #
    # Whether to use -msgcolor or not, for mplayer.
    # ======================================================================= #
    @use_mplayer_colours = true
    # ======================================================================= #
    # === @hash
    #
    # Next, we setup the @hash variable.
    # ======================================================================= #
    @hash = {} # This hash tells us how often a song was played.
    @hash = YAML.load_file(@store_where) if File.exist? @store_where
    # Need to ensure that @hash is a hash indeed.
    @hash = {} if @hash == false
    # ======================================================================= #
    # === @may_we_exit
    # ======================================================================= #
    @may_we_exit = true
  end

  # ========================================================================= #
  # === reset_main_string
  # ========================================================================= #
  def reset_main_string
    @_ = ''.dup # This is the sys-command to use, the mplayer string.
  end

  # ========================================================================= #
  # === verbose_exit
  # ========================================================================= #
  def verbose_exit
    opnn; e 'You did not provide a multimedia file to play.'
    opnn; e 'Thus exiting now.'
    exit_program
  end

  # ========================================================================= #
  # === show_help_then_exit
  # ========================================================================= #
  def show_help_then_exit
    esystem "#{@use_this_multimedia_player} --help"
    exit
  end

  # ========================================================================= #
  # === show_version_then_exit
  # ========================================================================= #
  def show_version_then_exit
    _ = `#{@use_this_multimedia_player} -version 2>&1`
    _ = _.split("\n")[-1]
    e(_)
    exit
  end

  # ========================================================================= #
  # === sanitize_input
  #
  # We need to be careful here.
  # ========================================================================= #
  def sanitize_input(i)
    if i.is_a? Array # Replace all '+' with '*'.
      # ===================================================================== #
      # Added Jan 2014. Might fail for certain input though.
      # ===================================================================== #
      return i.map {|entry| entry.to_s.tr('+','*') }
    else
      return i
    end
  end

  # ========================================================================= #
  # === try_to_ensure_that_the_files_exist
  #
  # Here we try to make sure that the files exist, by applying a
  # glob on it.
  # ========================================================================= #
  def try_to_ensure_that_the_files_exist
    @play_these_multimedia_files.map! {|file|
      unless File.exist? file # If the file does not exist, continue here.
        globbed = Dir[file+'*']
        file = globbed.first unless globbed.empty?
      end
      file
    }
  end

  # ========================================================================= #
  # === set_play_these_multimedia_files
  #
  # This will always return an Array.
  # ========================================================================= #
  def set_play_these_multimedia_files(
      i
    )
    verbose_exit if i.to_s.empty?
    i = i.dup if i.is_a? String
    # ======================================================================= #
    # Also includes nil.
    # ======================================================================= #
    @play_these_multimedia_files << sanitize_input(i)
    @play_these_multimedia_files.flatten!
    @play_these_multimedia_files.compact!
    try_to_ensure_that_the_files_exist
  end; alias set_use_these_files set_play_these_multimedia_files # === set_use_these_files

  # ========================================================================= #
  # === add                                                         (add tag)
  # ========================================================================= #
  def add(i)
    @play_these_multimedia_files << [i].flatten.compact
    @play_these_multimedia_files.flatten!
    @play_these_multimedia_files.uniq! # ← Added in June 2020.
  end

  # ========================================================================= #
  # === register_file
  # ========================================================================= #
  def register_file(
      i = @file_name # register the file
    )
    number = 1
    if i
      if @hash.is_a? Hash
        _ = File.basename( i.delete('"') ) # Get rid of "
        if @hash.has_key? _
          @hash[_] = @hash[_] + number 
        else
          @hash[_] = number
        end
        save_what_into_this_file
      else
        e 'Something went wrong - @hash variable is not a hash.'
        e "Make sure to check the file #{sfile(@store_where)}."
        exit
      end
    end
  end

  # ========================================================================= #
  # === save_what_into_this_file
  #
  # Use this to save the file played.
  # ========================================================================= #
  def save_what_into_this_file(
      what = @hash, where = STORE_WHERE
    )
    begin
      File.open(where, 'w+') { |f| YAML.dump(what, f) }
    rescue Exception => error
      opnn; pp error
    end
  end

  # ========================================================================= #
  # === cyan
  # ========================================================================= #
  def cyan(i)
    "#{Colours::CYAN}#{i}#{Colours::GREEN}"
  end

  # ========================================================================= #
  # === main_string?
  # ========================================================================= #
  def main_string?
    @_
  end

  # ========================================================================= #
  # === exit_program
  #
  # Use this method when exiting from this class.
  # ========================================================================= #
  def exit_program
    exit if @may_we_exit
  end

  # ========================================================================= #
  # === register_sigint
  # ========================================================================= #
  def register_sigint # Register at-exit handler.
    Signal.trap('SIGINT') { exit_program }
  end

  # ========================================================================= #
  # === consider_to_notify_the_user_if_more_than_one_video_file
  # ========================================================================= #
  def consider_to_notify_the_user_if_more_than_one_video_file
    if @play_these_multimedia_files.size > 1
      e "#{rev}More than one audio/video file was found. "\
        "We will output these now:"
      e
      @play_these_multimedia_files.each {|entry|
        e "#{sfancy('  - ')}#{entry}"
      }
      e
    end
  end

  # ========================================================================= #
  # === notify_the_user_that_these_simpsons_episodes_will_be_played
  # ========================================================================= #
  def notify_the_user_that_these_simpsons_episodes_will_be_played(i)
    e "#{rev}A total of #{i.size} different simpson episodes will be played."
    e "#{rev}The total duration of these files will be:"
    n_seconds = MultimediaParadise::FileDuration.total_time_of(i)
    e
    e sfancy('  '+n_seconds.to_s+' seconds')+
      ' ('+
      (n_seconds / 60.0).round(1).to_s+' minutes; '+
      slateblue(
        (n_seconds / 3600).round(1).to_s+' hours'
      )+
      ')'
    e
    set_use_these_files(i)
  end

  # ========================================================================= #
  # === menu                                                       (menu tag)
  # ========================================================================= #
  def menu(
      i = @commandline_arguments
    )
    if i and i.is_a?(Array)
      i.each {|entry| menu(entry) }
    else
      case i # case tag
      # ===================================================================== #
      # === mplayer_wrapper --simpsons=10
      # ===================================================================== #
      when /^-?-?simpsons=(\d+)$/i
        i = ::MultimediaParadise::Video.return_good_the_simpsons_episodes.shuffle
        n_files_to_play = $1.to_s.dup.to_i
        play_these_episodes = i[0 .. (n_files_to_play - 1)]
        notify_the_user_that_these_simpsons_episodes_will_be_played(play_these_episodes)
      # ===================================================================== #
      # === mplayer_wrapper --simpsons
      # ===================================================================== #
      when /^-?-?simpsons?\??$/i
        # =================================================================== #
        # Obtain all good simpsons episodes next.
        # =================================================================== #
        i = ::MultimediaParadise::Video.return_good_the_simpsons_episodes.shuffle # We play the default simpsons movies here.
        # =================================================================== #
        # Also add two random simpson videos, since as of 06.07.2019, if
        # we are on roebe.
        # =================================================================== #
        if is_on_roebe?
          i << MultimediaParadise.return_path_to_random_simpsons_video_file
          i << MultimediaParadise.return_path_to_random_simpsons_video_file
          i.uniq! # And make the list unique.
        end
        notify_the_user_that_these_simpsons_episodes_will_be_played(i)
      # ===================================================================== #
      # === --use-mplayer
      #
      # This entry point allows the user to specifically overrule
      # the multimedia-player in use.
      #
      # Invocation example:
      #
      #   mplaywrapper --use-mplayer --simpsons
      #
      # ===================================================================== #
      when /^-?-?use(-|_)?mplayer$/i
        @use_this_multimedia_player = 'mplayer'
      # ===================================================================== #
      # === :halloween
      #
      # This entry point specifically handles Symbols and Symbol-looking
      # Strings.
      # ===================================================================== #
      when :halloween, ':halloween',
           :predator,  ':predator'
        add map_symbol_to_locally_existing_file(i)
      # ===================================================================== #
      # === -nosound
      # ===================================================================== #
      when '-nosound',
           'silence',
           /^nosound$/i
        @append_this << '-nosound '
        i[1] = nil
      # ===================================================================== #
      # === mplayer_wrapper --version
      # ===================================================================== #
      when /^-?-?version\??$/i,
           '-v',
           '?'
        show_version_then_exit
      # ===================================================================== #
      # === --help
      # ===================================================================== #
      when /^-?-?help$/i
        show_help_then_exit
      else
        if File.exist? i
          add(i)
        else
          return i
        end
      end
    end
  end

  # ========================================================================= #
  # === play_each_multimedia_file
  #
  # Here we finally iterate over the collection and play every file.
  # ========================================================================= #
  def play_each_multimedia_file(
      i = @play_these_multimedia_files
    )
    i.each {|file|
      if File.directory? file
        opnn; e 'This is a directory. We will try to play every'
        opnn; e 'file in this directory next.'
        play_each_multimedia_file(
          Dir["#{file}/*"]
        )
      else
        register_file(file)
        reset_main_string
        _ "#{@use_this_multimedia_player} "
        _ APPEND_THESE_OPTIONS_TO_MPLAYER if APPEND_THESE_OPTIONS_TO_MPLAYER
        if @use_this_multimedia_player == 'mplayer'
          # The above check is necessary because mpv has no -zoom option.
          _ ' -fs -zoom ' if @play_zoomed
        end
        _ ' -vo null '  if @disable_video
        # =================================================================== #
        # Next add the -msgcolor flag. This works for mplayer, but may not
        # work for mpv, which insists on using -msg_color, so check for
        # this.
        # =================================================================== #
        if @use_mplayer_colours
          if main_string?.start_with? 'mpv '
            cmd = ' --msg-color '
            # ============================================================= #
            # Ok - this is mpv in this case:
            # ============================================================= #
            if does_the_video_player_support_this_commandline?
              _ cmd
            else
              _ ' -msgcolor '
            end
          else # else we assume mplayer is in use.
            _ ' -msgcolor '
          end
        end
        _ @append_this # Also append this string.
        # Starting from this point here, we will append the padded filename.
        e(@_+'"'+cyan(file)+'"') # We want the colours.
        _ '"'+file+'"'
        system(@_) # @_ is the String that holds our mplayer-command.
      end
    }
  end
 
  # ========================================================================= #
  # === _
  # ========================================================================= #
  def _(i)
    @_ << " #{i}"
    @_.squeeze!(' ')
  end; alias append _ # === append

  # ========================================================================= #
  # === run
  # ========================================================================= #
  def run
    menu
    consider_to_notify_the_user_if_more_than_one_video_file
    play_each_multimedia_file
  end

end

# =========================================================================== #
# === MultimediaParadise.play_simpsons
# =========================================================================== #
def self.play_simpsons(n_simpsons = 10)
  MultimediaParadise::MplayerWrapper.new('--simpsons='+n_simpsons.to_s)
end
          
end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::MplayerWrapper.new(ARGV)
end # mplayer Fette*