#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::PrepareVideoLecture
#
# This class will prepare a video lecture.
#
# The proper usage is as follows:
#
# You feed it some remote URLs, such as youtube URLs.
#
# Then, we will download these URLs.
#
# These files will then be renamed, and afterwards merged into a new
# file giant file.
#
# Usage example:
#
#   MultimediaParadise::PrepareVideoLecture.new

# =========================================================================== #
# require 'multimedia_paradise/video/prepare_video_lecture.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'
require 'multimedia_paradise/multimedia/merge_multimedia.rb'

module MultimediaParadise

class PrepareVideoLecture < ::MultimediaParadise::Base # === MultimediaParadise::PrepareVideoLecture

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === DEFAULT_URLS
  #
  # Which URLS to use by default.
  # ========================================================================= #
  DEFAULT_URLS = %w(
    https://www.youtube.com/watch?v=tyANFcnrM8I
    https://www.youtube.com/watch?v=4zHm_6AQ7CY
    https://www.youtube.com/watch?v=0OCnHNk2Hac
  )

  # ========================================================================= #
  # === BASE_DIRECTORY
  #
  # This is the directory we will log the output to.
  #
  # On my system, this will be at:
  #   "/Depot/Temp/MultimediaProject/video_lectures/"
  # ========================================================================= #
  BASE_DIRECTORY = "#{::MultimediaParadise.log_directory?}video_lectures/"

  # ========================================================================= #
  # === DELAY_IN_N_SECONDS
  #
  # Specify the delay that the interval-video will use.
  # ========================================================================= #
  DELAY_IN_N_SECONDS = 8

  # ========================================================================= #
  # === USE_DELAY_WHEN_MERGING_THE_VIDEO_FILES
  # ========================================================================= #
  USE_DELAY_WHEN_MERGING_THE_VIDEO_FILES = true

  # ========================================================================= #
  # === BLACK_MP4_VIDEO_FILE
  # ========================================================================= #
  BLACK_MP4_VIDEO_FILE = 'black.mp4'

  # ========================================================================= #
  # === OUTPUT_FILE
  # ========================================================================= #
  OUTPUT_FILE = 'output.mp4'

  # ========================================================================= #
  # === COMMAND
  # ========================================================================= #
  COMMAND = 'youtube-dl'

  # ========================================================================= #
  # === initialize
  #
  # The first argument is the list of remote URLs.
  # ========================================================================= #
  def initialize(
      these_urls  = nil,
      run_already = true
    )
    reset
    set_work_on_these_urls(these_urls)
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    @array_downloaded_files = []
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
  end

  # ========================================================================= #
  # === set_work_on_these_urls
  #
  # Pass here the remote URLS to use.
  # ========================================================================= #
  def set_work_on_these_urls(i = DEFAULT_URLS)
    i = [i] unless i.is_a? Array
    i = DEFAULT_URLS if i.empty?
    @work_on_these_urls = i # Must be an Array.
  end

  # ========================================================================= #
  # === work_on_these_urls?
  # ========================================================================= #
  def work_on_these_urls?
    @work_on_these_urls
  end; alias work_on_these_urls work_on_these_urls? # === work_on_these_urls
       alias urls?              work_on_these_urls? # === urls?
       alias input?             work_on_these_urls? # === input?

  # ========================================================================= #
  # === create_base_directory
  # ========================================================================= #
  def create_base_directory
    mkdir(download_dir?)
  end

  # ========================================================================= #
  # === download_dir?
  # ========================================================================= #
  def download_dir?
    BASE_DIRECTORY
  end

  # ========================================================================= #
  # === enter_base_directory
  # ========================================================================= #
  def enter_base_directory
    cd download_dir?
  end

  # ========================================================================= #
  # === create_black_image
  #
  # We will use ImageMagick convert to create a black .png file.
  # ========================================================================= #
  def create_black_image
    _ = 'convert -size 320x240 xc:black black.png'
    system _
  end

  # ========================================================================= #
  # === create_black_mp4_video_file
  # ========================================================================= #
  def create_black_mp4_video_file
    output_file = BLACK_MP4_VIDEO_FILE
    _ = 'ffmpeg -f lavfi -i color=color=black -t '+DELAY_IN_N_SECONDS.to_s+' '+output_file
    system _ unless File.exist? output_file
  end

  # ========================================================================= #
  # === show_welcome_message
  # ========================================================================= #
  def show_welcome_message
    unless input?.empty?
      opnn; e 'Welcome! We will next work on the given URLs.'
    end
  end

  # ========================================================================= #
  # === empty_base_directory
  # ========================================================================= #
  def remove_base_directory
    FileUtils.rm_rf(download_dir?) unless download_dir? == '/'
  end

  # ========================================================================= #
  # === report_everything_has_finished
  # ========================================================================= #
  def report_everything_has_finished
    result = 'Everything has now finished.'
    if File.exist? OUTPUT_FILE
      result << ' It may be available here: '+
                sfile((Dir.pwd+'/'+OUTPUT_FILE).squeeze('/'))
    end
    opnn; e result
  end

  # ========================================================================= #
  # === merge_the_videos
  # ========================================================================= #
  def merge_the_videos
    n_videos = @array_downloaded_files.size.to_s
    opnn; e 'These videos ('+simp(n_videos)+') will next be merged.'
    # ======================================================================= #
    # Sneakily insert the delay-video into the Array.
    # ======================================================================= #
    if USE_DELAY_WHEN_MERGING_THE_VIDEO_FILES
      _ = @array_downloaded_files
      array = Array.new(_.size, BLACK_MP4_VIDEO_FILE)
      _ = _.zip(array).flatten
      # ===================================================================== #
      # Last one should not be black, hence we remove it.
      # ===================================================================== #
      _.pop if _.last == BLACK_MP4_VIDEO_FILE
      @array_downloaded_files = _
    end
    pp @array_downloaded_files
    ::MultimediaParadise.merge(@array_downloaded_files)
  end

  # ========================================================================= #
  # === download_the_videos
  # ========================================================================= #
  def download_the_videos
    @array_downloaded_files = [] # Reset to empty always.
    opnn; e 'Now downloading the URLs. The download directory will be at:'+
            sdir(download_dir?)
    urls?.each {|remote_url|
      opnn; e 'Now working on the URL `'+sfancy(remote_url)+'` via '+
               simp(COMMAND)+'.'
      cmd = COMMAND+' '+remote_url
      # e cmd # Don't want to display it really.
      result = `#{cmd}` # <- This is the part where we download the file finally.
      File.delete('gmon.out') if File.exist? 'gmon.out' # We don't need gmon.out
      regex = /Merging formats into "(.*)"/
      result =~ regex
      this_file = $1.to_s.dup
      if File.exist? this_file
        new_filename = this_file.tr(' ','_').tr("'",'').tr('-','_').
                       tr('(','').tr(')','').tr('&','')
        new_filename.squeeze! '_'
        opnn; rename(this_file, new_filename)
        this_file = new_filename # Reassign.
      end
      @array_downloaded_files << this_file
      opnn; e 'The final file result is: `'+sfile(this_file)+'`.'
    }
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    show_welcome_message
    remove_base_directory
    create_base_directory
    enter_base_directory
    # create_black_image
    if USE_DELAY_WHEN_MERGING_THE_VIDEO_FILES
      create_black_mp4_video_file
    end
    download_the_videos
    merge_the_videos
    report_everything_has_finished
  end

end; end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::PrepareVideoLecture.new(ARGV)
end # prepare_videolecture https://www.youtube.com/watch?v=7TVWM45tfc0 https://www.youtube.com/watch?v=cHHOgYzudMw