#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::Video::RandomVideo
#
# This class will randomly fetch a video file from the current working
# directory, then play it. Alternatively you can provide another target
# directory, from which this script will then fetch all multimedia files
# and randomly play one.
#
# Usage example:
#
#   MultimediaParadise::Video::RandomVideo.new(ARGV)
#
# =========================================================================== #
# require 'multimedia_paradise/video/random_video.rb'
# MultimediaParadise::Video::RandomVideo.new(ARGV)
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'
require 'multimedia_paradise/toplevel_methods/is_a_multimedia_file.rb'

module MultimediaParadise

module Video

class RandomVideo < Base # === MultimediaParadise::Video::RandomVideo

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === initialize
  #
  # The first argument is usually the target directory.
  # ========================================================================= #
  def initialize(
      commandline_arguments = Dir.pwd,
      run_already           = true
    )
    reset
    set_commandline_arguments(
      commandline_arguments
    )
    set_play_from_this_directory(
      Dir.pwd # Default.
    )
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @which_video_player_to_use
    # ======================================================================= #
    @which_video_player_to_use = ::MultimediaParadise.which_video_player_to_use?
    # ======================================================================= #
    # === @from_where
    #
    # The next variable determines from where we will pick our video
    # files. If it is '*' then this means the current working
    # directory.
    # ======================================================================= #
    @from_where = '*'
  end

  # ========================================================================= #
  # === do_use_mplayer
  #
  # This method can be used to specifically use mplayer rather than, for
  # example, mpv.
  # ========================================================================= #
  def do_use_mplayer
    @which_video_player_to_use = 'mplayer'
  end

  # ========================================================================= #
  # === set_play_from_this_directory
  # ========================================================================= #
  def set_play_from_this_directory(i)
    i = i.dup if i.frozen?
    i << '/' unless i.end_with? '/'
    @play_from_this_directory = i
  end; alias set_target_directory set_play_from_this_directory # === set_target_directory

  # ========================================================================= #
  # === play_from_this_directory?
  # ========================================================================= #
  def play_from_this_directory?
    @play_from_this_directory
  end; alias target_directory? play_from_this_directory? # === target_directory? 
       alias target_dir?       play_from_this_directory? # === target_dir?

  # ========================================================================= #
  # === set_commandline_arguments
  # ========================================================================= #
  def set_commandline_arguments(i = '')
    unless i.is_a? Array
      i = [i]
    end
    i.flatten!
    @commandline_arguments = i
  end

  # ========================================================================= #
  # === commandline_arguments?
  # ========================================================================= #
  def commandline_arguments?
    @commandline_arguments
  end; alias input? commandline_arguments? # === input?

  # ========================================================================= #
  # === select_only_multimedia_files
  #
  # We can only play video or audio files. This method will take care
  # of that constraint, by selecting only files that are multimedia
  # files.
  # ========================================================================= #
  def select_only_multimedia_files
    @all_files.select! {|entry|
      is_a_multimedia_file?(entry)
    }
  end

  # ========================================================================= #
  # === is_a_multimedia_file?
  # ========================================================================= #
  def is_a_multimedia_file?(i)
    ::MultimediaParadise.is_a_multimedia_file?(i)
  end

  # ========================================================================= #
  # === sort
  # ========================================================================= #
  def sort
    @all_files.sort!
  end

  # ========================================================================= #
  # === can_not_play_any_file_as_we_did_not_find_a_multimedia_file_in_this_directory
  # ========================================================================= #
  def can_not_play_any_file_as_we_did_not_find_a_multimedia_file_in_this_directory(
      this_directory = target_dir?
    )
    opnn; e 'Can not play any file as we did not find a multimedia '\
            'file in this ('+sdir(this_directory)+') directory.'
  end

  # ========================================================================= #
  # === play_a_random_video
  # ========================================================================= #
  def play_a_random_video
    _ = @all_files
    if _.empty?
      can_not_play_any_file_as_we_did_not_find_a_multimedia_file_in_this_directory
    else
      # ===================================================================== #
      # Keep in mind that the input may be a number.
      # ===================================================================== #
      if input?.first =~ /^\d+$/ # If only numbers.
        input?.first.to_i.times {
          play_video(_.sample)
        }
      else
        play_video(_.sample)
      end
    end
  end

  # ========================================================================= #
  # === play_video
  #
  # This method is the one that will play the video-file at hand.
  # ========================================================================= #
  def play_video(i)
    if i.is_a? Array
      i.each {|this_file| play_video(this_file) }
    else
      if i.include?("'") or i.include?(' ')
        i = '"'+i+'"'
      end
      begin
        require 'roebe/classes/kde/kde_konsole/kde_konsole.rb'
        Roebe.rename_konsole(:television, :be_quiet)
      rescue; end
      _ = use_which_video_player?+' -vo x11 '+i
      _.squeeze!(' ')
      esystem _
    end
  end

  # ========================================================================= #
  # === use_which_video_player?
  # ========================================================================= #
  def use_which_video_player?
    @which_video_player_to_use
  end; alias use_this_video_player use_which_video_player? # === use_this_video_player

  # ========================================================================= #
  # === menu                                                       (menu tag)
  # ========================================================================= #
  def menu(
      i = @commandline_arguments
    )
    if i.is_a? Array
      i.each {|entry| menu(entry) }
    else
      case i # case tag
      # ===================================================================== #
      # === random_video --use-mplayer
      # ===================================================================== #
      when /-?-?use(-|_)?mplayer$/,
           /-?-?mplayer$/
        do_use_mplayer
      # ===================================================================== #
      # === random_video --from-realvids
      # ===================================================================== #
      when /-?-?from(-|_)?realvids$/
        @from_where = realvids_directory?+'*'
        set_target_directory(realvids_directory?)
      # ===================================================================== #
      # === random_video --help
      # ===================================================================== #
      when /help$/
        show_help
        exit
      end
    end
  end

  # ========================================================================= #
  # === obtain_all_files
  #
  # By default, we will grab entries from the current directory.
  #
  # This can be modified by the input.
  # ========================================================================= #
  def obtain_all_files(
      from_where = @from_where # This defaults to the current working directory.
    )
    unless target_directory?.empty?
      from_where = target_directory?.dup
    end
    from_where << '*' unless from_where.end_with? '*'
    all_files = Dir[from_where]
    @all_files = all_files
  end

  # ========================================================================= #
  # === show_help                                                  (help tag)
  #
  # Invocation example:
  #
  #   random_video --help
  #
  # ========================================================================= #
  def show_help
    help_string = <<-EOF

The following help options are available and documented:

  --use-mplayer   # use mplayer to play a random video
  --from-realvids # play from the realvids directory

EOF
    e help_string
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    menu
    obtain_all_files
    select_only_multimedia_files
    sort
    play_a_random_video
  end

  # ========================================================================= #
  # === MultimediaParadise[]
  # ========================================================================= #
  def self.[](i = '')
    new
  end

end; end

# =========================================================================== #
# === MultimediaParadise.random_video
# =========================================================================== #
def self.random_video
  ::MultimediaParadise::Video::RandomVideo.new
end

end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::Video::RandomVideo.new(ARGV)
end # random_video
    # random_video 3 # grab 3 random videos