#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::SpeedUpVideo
#
# This class can be used to speed up a video, using ffmpeg.
#
# Usage example:
#   MultimediaParadise::SpeedUpVideo.new
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

module Video

class SpeedUpVideo < MultimediaParadise::Base # require 'speed_up_video'; SpeedUpVideo.new

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = nil,
      run_already = true
    )
    reset
    set_input_file(i)
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # Define the speed-up as percentage-number for the instance
    # variable @speed_up_by_n_percent. 200 effectively means
    # "double speed".
    # ======================================================================= #
    @speed_up_by_n_percent = 200 
  end

  # ========================================================================= #
  # === set_input_file
  # ========================================================================= #
  def set_input_file(i = '')
    if i.is_a? Array
      i = i.join(' ').strip
    end
    i = i.to_s.dup
    @input_file = i
  end

  # ========================================================================= #
  # === input_file?
  # ========================================================================= #
  def input_file?
    @input_file
  end

  # ========================================================================= #
  # === return_multiplier_for_speeding_up_the_audio_rate
  # ========================================================================= #
  def return_multiplier_for_speeding_up_the_audio_rate
    result = (@speed_up_by_n_percent.to_f / 25).to_f
    return result.to_s
  end

  # ========================================================================= #
  # === return_multiplier_for_speeding_up_the_video_rate
  #
  # This will return the number for the multiplier in use for
  # videos. We can get it by using the multiplier 0.005 for
  # percentage value.
  #
  # Here is a table that shows these value:
  #
  #      50% is 2.0  #  50 /   25
  #     100% is 1.0  # 100 /  100
  #     150% is 0.75 # 150 /  200
  #     200% is 0.5  # 200 /  400
  #     400% is 0.25 # 400 / 1600
  #
  # Normalized towards 100 it is:
  #
  #      50% is 2.0  # 200 /  100
  #     100% is 1.0  # 100 /  100
  #     150% is 0.75 #  75 /  100
  #     200% is 0.5  #  50 /  100
  #     400% is 0.25 #  25 /  100
  #
  # ========================================================================= #
  def return_multiplier_for_speeding_up_the_video_rate
    (return_multiplier_for_speeding_up_the_audio_rate.to_f / 4).to_s
  end

  # ========================================================================= #
  # === return_filter_for_doubling_the_speed
  # ========================================================================= #
  def return_filter_for_doubling_the_speed
    ' -filter_complex "[0:v]setpts=0.5*PTS[v];[0:a]atempo=2.0[a]"'
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    # ======================================================================= #
    # We will speed up video and audio at the same time.
    # ======================================================================= #
    _ = 'ffmpeg'.dup
    _ << ' -i '+input_file?
    # _ << ' -filter_complex "[0:v]setpts='+return_multiplier_for_speeding_up_the_video_rate+
    #      'x*PTS[v];[0:a]atempo='+return_multiplier_for_speeding_up_the_audio_rate+'[a]"'
    _ << return_filter_for_doubling_the_speed
    _ << ' -map "[v]" -map "[a]"'
    _ << ' -y' # Overwrite it: yes. 
    _ << ' output.mkv'
    esystem _
  end

  # ========================================================================= #
  # === MultimediaParadise::Video::SpeedUpVideo[]
  # ========================================================================= #
  def self.[](i = '')
    self.new(i)
  end

end; end; end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::Video::SpeedUpVideo.new(ARGV)
end # speed_up_video