#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::SrtRegex
#
# Use this class to modify .srt files.
#
# Specific Invocation and Usage examples follow now:
#
#   rsub 166_Kagemusha_CD2.srt > X.srt
#   rsub WarriorsOfHeavenAndEarth2.srt > /X.srt;mv WarriorsOfHeavenAndEarth.srt backup2.srt;mv /X.srt WarriorsOfHeavenAndEarth.srt
#   rsub Musa_CD2.srt > /Y.srt;bl /Y.srt;mv WarriorsOfHeavenAndEarth.srt backup2.srt;mv /X.srt WarriorsOfHeavenAndEarth.srt
#   rsub /j/Catch_22_2.srt > TEST_2.srt
#   rsub 222*.srt > modified.srt;mv modified.srt 222_CurseOfTheGoldenFlower.srt
#   rsub 222*.srt > modified.srt;mv modified.srt 222_CurseOfTheGoldenFlower.srt
#   rsub 233_Daywatch.srt > modified.srt;mv modified.srt x233_Daywatch.srt;mv  x233_Daywatch.srt 233_Daywatch.srt
#   rsub *srt > test.srt
#
# =========================================================================== #
# Usage examples:
#   MultimediaParadise::SrtRegex.new
# =========================================================================== #
# require 'multimedia_paradise/video/srt_regex.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'
begin
  require 'x/time/long_format_to_milliseconds_converter.rb'
rescue LoadError; end
begin
  require 'x/time/milliseconds_to_long_format_converter'
rescue LoadError; end

module MultimediaParadise

class SrtRegex < ::MultimediaParadise::Base # === MultimediaParadie::SrtRegex.new

  alias ee print # Must be print. Colours would spam with Escape Sequences.

  # ========================================================================= #
  # === SHALL_WE_DEBUG
  # ========================================================================= #
  SHALL_WE_DEBUG = false # if true, then we debug this script.

  # ========================================================================= #
  # === ARROW
  # ========================================================================= #
  ARROW = '-->'

  # ========================================================================= #
  # === THIS_FILE
  # ========================================================================= #
  THIS_FILE = MultimediaParadise.project_base_dir+'video/srt_regex.rb'

  # ========================================================================= #
  # === MODIFIER_IN_MS
  #
  # Here, specify how many ms we shall add to our new .srt file
  # Because it is in milliseconds, append 1000 to it, in order to
  # know how many seconds you must add. So for example, 30.000
  # milliseconds means that we have 30 seconds.
  # ========================================================================= #
  MODIFIER_IN_MS = 5_057

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = nil,
      run_already = true
    )
    reset
    set_input(i)
    run if run_already
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    super()
    clear_new_dataset
    @modifier_in_ms = MODIFIER_IN_MS
    @debug = SHALL_WE_DEBUG
    @shall_we_overwrite = false
  end

  # ========================================================================= #
  # === clear_new_dataset
  # ========================================================================= #
  def clear_new_dataset
    @new_dataset = []
  end

  # ========================================================================= #
  # === set_input
  # ========================================================================= #
  def set_input(i = nil)
    if i.empty?
      i = try_to_find_srt_files_in_current_working_directory()
      if i.empty?
        e 'No arg given for '+sfile(__FILE__)+', exiting now.'
        exit
      end
    end
    if i.is_a? Array
      if i[1]
        @modifier_in_ms = i[1].to_f
      end
      @shall_we_overwrite = true if i[2] == 'OVERWRITE'
      i = i.first
    end
    case i # case tag
    # ======================================================================= #
    # === srt --help
    # ======================================================================= #
    when 'HELP','help','--help'
      show_help
      exit
    end
    @input = i
  end

  # ========================================================================= #
  # === show_help                                                  (help tag)
  # ========================================================================= #
  def show_help
    opn; e 'How to use this class?'
    e
    e '    srtt *InfernalAff*.srt 7_250'
    e '    srtt *InfernalAff*.srt 1_250 OVERWRITE'
    e
    if ENV['IS_ROEBE'].to_s == '1' # We assume that we are on my system here, so we can display more stuff.
      _ = File.readlines(THIS_FILE).select {|entry| entry.start_with? '#   rsub' }
      _.each {|line|
        line = line.gsub(/^#   rsub/, 'srtt')
        e '    '+line
      }
    end; e
  end

  # ========================================================================= #
  # === try_to_find_srt_files_in_current_working_directory
  # ========================================================================= #
  def try_to_find_srt_files_in_current_working_directory
    return Dir['*.srt']
  end

  # ========================================================================= #
  # === read_in_file
  # ========================================================================= #
  def read_in_file
    @dataset = File.readlines(@input)
  end

  # ========================================================================= #
  # === process_file
  # ========================================================================= #
  def process_file
    clear_new_dataset
    @dataset.each {|line|
      line.chomp!
      e line
      ee ">>>>>>> #{line} <<<<<<<<" if @debug
      if line.include? ARROW

        splitted = line.split(ARROW)
        _ = LongFormatToMillisecondsConverter.new(splitted[0])
        _.add @modifier_in_ms
        result1 = MillisecondsToLongFormatConverter.convert(_.ms.to_s)
        # result1 = _.ms.to_s

        _ = LongFormatToMillisecondsConverter.new(splitted[1])
        _.add @modifier_in_ms
        result2 = MillisecondsToLongFormatConverter.convert(_.ms.to_s)
        # result2 = _.ms.to_s

        line = result1+' '+ARROW+' '+result2
      end
      @new_dataset << line+N
    }
  end

  # ========================================================================= #
  # === create_new_file
  #
  # This method will create a new file, with the name being prepended
  # to "NEW_".
  # ========================================================================= #
  def create_new_file
    _ = 'NEW_'+@input
    _ = @input if @shall_we_overwrite
    e 'We will create a new file `'+sfile(_)+'`.'
    save_file(@new_dataset, _)
  end

  # ========================================================================= #
  # === run
  # ========================================================================= #
  def run
    read_in_file
    process_file
    create_new_file
  end

end; end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::SrtRegex.new(ARGV)
end # srtt
    # srtt Blade.srt 5_111