#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::Video::VideoInformation
#
# This class will give us information about a video file, by making
# use of ffmpeg.
#
# Example output might be like this:
#
# > vinfo 033_JackieBrown.mp4
#
# VideoInformation: The video stream is `h264`.
# VideoInformation: The audio stream is `aac`.
#
# Usage example:
#
#   MultimediaParadise::Video::VideoInformation.new
#
# =========================================================================== #
# require 'multimedia_paradise/video/video_information.rb'
# MultimediaParadise::Video::VideoInformation.new(ARGV)
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

module Video

class VideoInformation < ::MultimediaParadise::Base # === MultimediaParadise::Video::VideoInformation.new

  require 'multimedia_paradise/misc/long_format_to_milliseconds_converter.rb'

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === SHALL_WE_DEBUG
  # ========================================================================= #
  SHALL_WE_DEBUG = false

  # ========================================================================= #
  # === VIDEO_STREAMS
  #
  # The registered videostreams come here.
  # ========================================================================= #
  VIDEO_STREAMS = %w(
    h264 mpeg1video msmpeg4v2 mpeg4
  )

  # ========================================================================= #
  # === AUDIO_STREAMS
  # ========================================================================= #
  AUDIO_STREAMS = %w( aac wmav2 mp2 mp3 )

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = nil,
      run_already = true
    )
    reset
    set_files(i)
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    set_audio # Set it to empty first.
    # ======================================================================= #
    # === @video
    # ======================================================================= #
    @video = nil
    # ======================================================================= #
    # === @be_verbose
    # ======================================================================= #
    @be_verbose = true
    # ======================================================================= #
    # === @data
    #
    # This Array will hold video and audio.
    # ======================================================================= #
    @data = []
    # ======================================================================= #
    # === @report_the_namespace
    # ======================================================================= #
    @report_the_namespace = true
    # ======================================================================= #
    # === resolution
    #
    # This will carry the video result, such as: 1280x720
    # ======================================================================= #
    @resolution = nil
    set_duration nil # Initialize it to nil specifically.
  end

  # ========================================================================= #
  # === set_files
  # ========================================================================= #
  def set_files(i = '')
    @files = i
  end

  # ========================================================================= #
  # === files?
  # ========================================================================= #
  def files?
    @files
  end; alias input? files? # === input?
       alias file?  files? # === file?

  # ========================================================================= #
  # === name?
  # ========================================================================= #
  def name?
    _ = file?
    _ = _.first if _.is_a? Array
    _
  end

  # ========================================================================= #
  # === video?
  # ========================================================================= #
  def video?
    @video
  end

  # ========================================================================= #
  # === set_hertz
  # ========================================================================= #
  def set_hertz(i)
    @hertz = i.scan(/(\d+) Hz/).flatten.first if i.include? ' Hz'
  end

  # ========================================================================= #
  # === audio?
  # ========================================================================= #
  def audio?
    @audio
  end; alias audio audio? # === audio

  # ========================================================================= #
  # === be_verbose?
  # ========================================================================= #
  def be_verbose?
    @be_verbose
  end

  # ========================================================================= #
  # === set_audio                                                 (audio tag)
  # ========================================================================= #
  def set_audio(i = nil)
    if i.is_a? Array
      i = i.select {|entry| entry.include? 'Audio:' }
      set_hertz(i) if i.include? ' Hz'
      i = i.first unless i.empty?
    end
    if i
      i = AUDIO_STREAMS.find { |s| i.include?(s.to_s) }
      i = nil if i.is_a? Array # Then it must be empty.
    end
    @audio = i
  end

  # ========================================================================= #
  # === populate_main_data
  # ========================================================================= #
  def populate_main_data
    @data = [video?, audio?] # Build up @data.
  end

  # ========================================================================= #
  # === files?
  # ========================================================================= #
  def files?
    @files
  end

  # ========================================================================= #
  # === inspect_this_file_via_ffmpeg
  #
  # The argument given to this method should be an existing file.
  # ========================================================================= #
  def inspect_this_file_via_ffmpeg(
      the_file = files?.first
    )
    `ffmpeg -i #{the_file} 2>&1` # The input-stream.
  end

  # ========================================================================= #
  # === set_video                                                 (video tag)
  # ========================================================================= #
  def set_video(i)
    if i.is_a? Array
      i = i.select {|entry| entry.include? 'Video:' }
      i = i.first unless i.empty?
    end
    result = VIDEO_STREAMS.find { |s| i.include?(s) }
    # ======================================================================= #
    # Handle XVID next:
    # ======================================================================= #
    if result == 'mpeg4' and i.include?('XVID')
      result << ' (XVID)'
    end
    @video = result
  end

  # ========================================================================= #
  # === data?
  # ========================================================================= #
  def data?
    @data
  end; alias data data? # === data

  # ========================================================================= #
  # === report_audio_stream
  #
  # We try to report which audio stream we have, e. g. ".mp3".
  # ========================================================================= #
  def report_audio_stream
    if audio?
      opnn; e 'The audio stream is `'+sfancy(audio?)+'`.'
      if audio?.end_with? 'mp3' # Ok we have a .mp3 file. Report the audio-bitrate next.
        report_audio_bitrate
      end
    else
      opnn; e 'No audio stream could be found.'
    end
  end

  # ========================================================================= #
  # === report_audio_bitrate
  # ========================================================================= #
  def report_audio_bitrate
    regex = /bitrate: (\d+ kb\/s)/
    _ = inspect_this_file_via_ffmpeg
    _ =~ regex
    if $1
      set_bitrate($1)
      opnn; e 'The audio bitrate is:'
      opnn; e "  #{simp(bitrate?)}"
    end
  end

  # ========================================================================= #
  # === bitrate?
  # ========================================================================= #
  def bitrate?
    @bitrate
  end

  # ========================================================================= #
  # === set_bitrate
  # ========================================================================= #
  def set_bitrate(i)
    @bitrate = i.to_s
  end

  # ========================================================================= #
  # === opnn
  # ========================================================================= #
  def opnn
    super(NAMESPACE) if @report_the_namespace
  end

  # ========================================================================= #
  # === report_the_name_of_the_file
  # ========================================================================= #
  def report_the_name_of_the_file
    _ = name?
    if File.exist? _
      opnn; e "#{rev}The name of the video file is `#{simp(_)}`."
    end
  end

  # ========================================================================= #
  # === report_video_stream
  # ========================================================================= #
  def report_video_stream
    if File.exist? name?
      opnn; print "The video stream is `#{sfancy(video?)}`."
      if @resolution
        report_the_resolution
      end
      e
    end
  end

  # ========================================================================= #
  # === input_stream?
  # ========================================================================= #
  def input_stream?
    inspect_this_file_via_ffmpeg(
      files?.first
    )
  end

  # ========================================================================= #
  # === determine_duration_of_the_file
  # ========================================================================= #
  def determine_duration_of_the_file
    use_this_regex = /Duration: (\d{1,2}:\d{1,2}:\d{1,2}.\d{1,2})/ # See: http://rubular.com/r/NYoZ9GMryH
    input_stream? =~ use_this_regex
    set_duration($1.to_s.dup)
  end

  # ========================================================================= #
  # === duration?
  # ========================================================================= #
  def duration?
    @duration
  end

  # ========================================================================= #
  # === verbose_report
  #
  # This method will report the name, the video stream and the
  # audio stream.
  # ========================================================================= #
  def verbose_report(report_the_namespace = true)
    case report_the_namespace
    when :do_not_report_the_namespace
      @report_the_namespace = false
    end
    report_the_name_of_the_file
    report_video_stream
    report_audio_stream
    report_the_duration if @duration
  end; alias report verbose_report # === report

  # ========================================================================= #
  # === set_duration
  # ========================================================================= #
  def set_duration(i = nil)
    if i and i.is_a?(String) and i.include?(':')
      i = MultimediaParadise.long_format_to_milliseconds(i).to_f / 1000.0
    end
    @duration = i
  end

  # ========================================================================= #
  # === report_the_resolution
  # ========================================================================= #
  def report_the_resolution
    print ' The video-resolution is '+
          sfancy(@resolution)+'.'
  end

  # ========================================================================= #
  # === report_the_duration
  #
  # This method will report the duration of that particular multimedia
  # file.
  # ========================================================================= #
  def report_the_duration
    duration = duration?
    opnn; e 'The duration of this file is `'+
            sfancy(duration.to_s)+'` seconds. '\
            '('+seagreen(
              to_hh_mm_ss(duration)
            )+')'
  end

  # ========================================================================= #
  # === decompose
  #
  # The input will hold all information.
  # ========================================================================= #
  def decompose(i)
    i = i.split(N)
    if SHALL_WE_DEBUG
      opnn; e 'The user wants to debug, so we will output the information'
      opnn; e 'that is to be decomposed:'
      pp i
    end
    grab_all_streams = i.select {|entry|
      entry.strip.include? 'Stream #' # Grab all video and audio streams.
    }
    set_audio(grab_all_streams)
    set_video(grab_all_streams)
    set_resolution(grab_all_streams)
  end

  # ========================================================================= #
  # === set_resolution
  # ========================================================================= #
  def set_resolution(i)
    if i
      if i.is_a? Array
        selection = i.select {|entry|
          entry.include?('Video: ')
        }.first
        if selection.include? 'x'
          i = selection.scan(
                /\d{2,4}x\d{2,4}/
              ).flatten.first.to_s.strip
        end
      end
      @resolution = i
    end
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    files?.each {|the_file|
      if File.exist? the_file
        _ = inspect_this_file_via_ffmpeg(the_file)
        decompose(_)
        populate_main_data
        determine_duration_of_the_file
      else
        if be_verbose?
          opnn; e 'File at `'+sfile(the_file)+'` does not exist.'
        end
      end
    }
  end

end; end

# =========================================================================== #
# === MultimediaParadise.video_information
# =========================================================================== #
def self.video_information(i)
  _ = MultimediaParadise::Video::VideoInformation.new(i)
  _.verbose_report(:do_not_report_the_namespace)
end

end

if __FILE__ == $PROGRAM_NAME
  _ = MultimediaParadise::Video::VideoInformation.new(ARGV)
  _.verbose_report(:do_not_report_the_namespace) # Better to not report it - less verbose that way.
end # vinfo /home/x/video/Realvids/254_SleepyHollow.mp4