#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::VideoPlayer
#
# Description:
#
#   This is a little ruby-wrapper over mplayer (or mpv).
#
#   We can use this class to play a video file from a USB device, for example.
#
#   Of course this class will also work for local video files from a HDD/SDD
#   device.
#
#   We can also use this class to play sub-groups of Videos, certain genres,
#   such as horror movies, science fiction movies and so forth.
#
#   See the usage examples in the next section in order to find out how
#   to use this "video player".
#
# =========================================================================== #
# === Usage Examples Examples:
#
#    rvid naked
#    rivd branca   l   # play Branca Leone, and then continue to loop
#    rvid branca   551 # plays via -ss 551
#    rvid powerman hdd # Play all Powerman-matches, but search only locally.
#
# =========================================================================== #
# require 'multimedia_paradise/video/video_player.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'
require 'multimedia_paradise/video/find_video/find_video.rb' # bl $RSRC/VID/find_video.rb

module MultimediaParadise

class VideoPlayer < Base # === MultimediaParadise::VideoPlayer

  begin
    require 'mountpoints'
  rescue LoadError; end

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === BE_VERBOSE
  # ========================================================================= #
  BE_VERBOSE  = true

  # ========================================================================= #
  # === DEBUG
  # ========================================================================= #
  DEBUG       = false

  # ========================================================================= #
  # === DEFAULT_PLAY_MODE
  # ========================================================================= #
  DEFAULT_PLAY_MODE = 'loop'

  # ========================================================================= #
  # === USE_VO_X11
  #
  # If true then we will use '-vo x11'
  # ========================================================================= #
  USE_VO_X11  = true

  # ========================================================================= #
  # Setup the colours to use.
  # ========================================================================= #
  DEFAULT_COLOUR = 'green'
  FANCY_COLOUR   = 'cyanb'

  # ========================================================================= #
  # === COLOUR_WARN
  # ========================================================================= #
  COLOUR_WARN    = 'redb'

  # ========================================================================= #
  # This is the hash that has the colour definitions. Note that since as
  # of June 2019, we will use this Hash to update the old Hash.
  # ========================================================================= #
  hash = {
    swarn:          COLOUR_WARN.to_sym,
    sfancy:         FANCY_COLOUR.to_sym,
    default_colour: DEFAULT_COLOUR.to_sym,
    sfile:          :magenta
  }
  Colours.update_the_colour_codes(hash) # Set the new colours here.

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      commandline_options = nil,
      run_already         = true
    )
    register_sigint
    reset
    set_runmode :standalone
    set_commandline_options(
      commandline_options
    )
    case run_already
    when :dont_run_yet, :do_not_run_yet
      run_already = false
    end
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @use_wrapper_over_mplayer
    # ======================================================================= #
    set_be_verbose
    @use_wrapper_over_mplayer = false
    # ======================================================================= #
    # === @debug
    # ======================================================================= #
    @debug = false # Whether to debug or whether we won't.
    # ======================================================================= #
    # === @target_location
    #
    # The ivar @target_location will tell us where we have to search for.
    #
    # For instance, it can be '/Depot/Video/' or '/home/x/VIDEO/'.
    # ======================================================================= #
    @target_location = '/home/x/video/'.dup # We may need to change it.
    @shall_we_exit = false
    # ======================================================================= #
    # === @colour
    # ======================================================================= #
    @colour = DEFAULT_COLOUR
    # ======================================================================= #
    # === @disable_video
    # ======================================================================= #
    @disable_video = false
    # ======================================================================= #
    # @array_play_these_video_files is an Array that tells us which
    # video files will be played.
    # ======================================================================= #
    @array_play_these_video_files = []
    # ======================================================================= #
    # We keep track how many times we played a video file.
    # ======================================================================= #
    @count_how_many_times_we_played_already = 0
    @play_in_a_loop = false
    set_runmode :standalone # Set @runmode to 'standalone'
  end

  # ========================================================================= #
  # === set_runmode
  #
  # This method can be used to set the runmode.
  #
  # We currently support only two runmodes:
  #
  #   - standalone, or
  #   - connected.
  #
  # 's' is an alias to standalone. Some more aliases exist, see the
  # case menu below for details.
  # ========================================================================= #
  def set_runmode(i = 'standalone')
    i = i.to_s
    case i
    when 'conn'
      i = 'connected'
    when 's','solo','stand'
      i = 'standalone'
    end
    @runmode = i
  end

  # ========================================================================= #
  # === report_filename
  #
  # Simply prepend the filename.
  # ========================================================================= #
  def report_filename
    File.basename(__FILE__)+': '
  end

  # ========================================================================= #
  # === disable_video
  #
  # Use this method to disable video playback. You still get the audio
  # output of the video file of course.
  # ========================================================================= #
  def disable_video
    @disable_video = true
  end

  # ========================================================================= #
  # === set_be_verbose
  # ========================================================================= #
  def set_be_verbose(i = BE_VERBOSE)
    @be_verbose = i
  end

  # ========================================================================= #
  # === set_debug
  # ========================================================================= #
  def set_debug(i = DEBUG)
    @debug = DEBUG
  end

  # ========================================================================= #
  # === register_sigint
  #
  # Silently restore the prompt again here via reset, which is done via
  # a system() call.
  # ========================================================================= #
  def register_sigint
    sigint_proc = proc {
      system 'reset'
      efancy 'Requested to exit from MultimediaParadise::VideoPlayer. Exiting now.'
      @shall_we_exit = true
      exit_program
    }
    Signal.trap 'SIGINT', sigint_proc
  end

  # ========================================================================= #
  # === runmode?
  # ========================================================================= #
  def runmode?
    @runmode
  end

  # ========================================================================= #
  # === warn_then_exit
  # ========================================================================= #
  def warn_then_exit(i)
    e Colours::RED+i
    exit_program
  end

  # ========================================================================= #
  # === add
  # ========================================================================= #
  def add(i = nil)
    if i
      @array_play_these_video_files << i unless i.empty?
    end
  end

  # ========================================================================= #
  # === video_files?
  # ========================================================================= #
  def video_files?
    @array_play_these_video_files.flatten
  end; alias videos? video_files? # === videos?

  # ========================================================================= #
  # === play_in_a_loop?
  # ========================================================================= #
  def play_in_a_loop?
    @play_in_a_loop
  end

  # ========================================================================= #
  # === play_video_files
  # ========================================================================= #
  def play_video_files(
      i = @array_play_these_video_files
    )
    if i.empty?
      opnn; e 'Can not play any video file - our collection is empty.'
    else
      if play_in_a_loop?
        loop {
          play(random_video)
          break if @shall_we_exit
        }
      else
        play(video_files?)
      end
    end
  end; alias play_video_file play_video_files # === play_video_file

  # ========================================================================= #
  # === save_last_video_file_played
  #
  # This method saves which video file was last played.
  #
  # By default it stores into "$HOME/last_video_file_played".
  # ========================================================================= #
  def save_last_video_file_played(i = nil)
    if i
      store_where = HOME_DIR+'last_video_file_played'
      opnn; e 'Storing at `'+sfile(store_where)+'`.'
      save_file(i, store_where)
    end
  end

  # ========================================================================= #
  # === play                                                       (play tag)
  #
  # Use this method to play the respective video file in question.
  #
  # We will also pad the filename.
  # ========================================================================= #
  def play(i)
    e 'In play() the argument was '+sfancy(i.to_s) if @debug
    if i.is_a? Array
      i.each {|entry| play(entry) }
    else
      i.strip!
      i.gsub!(/  /,' ') if i.include? '  '
      @count_how_many_times_we_played_already += 1
      if @count_how_many_times_we_played_already > 1000
        e 'We played more than 1000 files. We will exit now.'
        @shall_we_exit = true
      end
      if @use_wrapper_over_mplayer # Use the mplayer-wrapper in this case.
        Mplayer.new(i)
      else # else just build the command.
        _ = ::MultimediaParadise.which_video_player_to_use?.dup
        if @disable_video
          _ << ' -vo null '
        else
          _ << ' -vo x11' if USE_VO_X11
          # ================================================================= #
          # mpv does no longer support - zoom, whereas mplayer does.
          # ================================================================= #
          _ << ' -zoom'   if PLAY_ZOOMED and _.include? 'mplayer'
        end
        # =================================================================== #
        # Save which video file was last played next.
        # =================================================================== #        
        save_last_video_file_played(i)
        # =================================================================== #
        # Pad the given input next.
        # =================================================================== #
        i = '"'+i+'"'
        _ << ' '+i
        e _
        system _ # Could also use: `#{_}`
      end
    end
  end

  # ========================================================================= #
  # === commandline_options?
  # ========================================================================= #
  def commandline_options?
    @commandline_options
  end; alias commandline? commandline_options? # === commandline?

  # ========================================================================= #
  # === set_commandline_options
  # ========================================================================= #
  def set_commandline_options(i)
    i = i.join(' ').strip if i.is_a? Array
    # ======================================================================= #
    # Get rid of ',' if they are part of a String.
    # ======================================================================= #
    i.delete!(',') if i.include? ','
    @commandline_options = i # Will be a String.
  end

  # ========================================================================= #
  # === obtain_all_video_files
  # ========================================================================= #
  def obtain_all_video_files
    # ======================================================================= #
    # Return only video files. And only from the target location.
    # ======================================================================= #
    files = Dir[@target_location+'**/**']
    filter_for_video_files(files)
  end

  # ========================================================================= #
  # === menu                                                       (menu tag)
  # ========================================================================= #
  def menu(
      i = commandline_options?
    )
    if i.is_a? Array
      i.each {|entry| menu(entry) }
    else
      case i # case tag
      # ===================================================================== #
      # === mma
      # ===================================================================== #
      when /^-?-?mma$/i
        @target_location << 'Mma'
      # ===================================================================== #
      # === breakdance
      # ===================================================================== #
      when /^-?-?breakdance$/i,
           /^-?-?bdance$/i
        @target_location << 'Bboying'
      # ===================================================================== #
      # === married
      #
      # Invocation example:
      #
      #   videoplayer --married
      #
      # ===================================================================== #
      when /^-?-?married$/i,
           /^-?-?married(_|-)?with(_|-)?children$/i,
           /^-?-?with(_|-)?children$/i
        @target_location << 'Comedy/MarriedWithChildren'
      # ===================================================================== #
      # === sex
      #
      # Invocation example:
      #
      #   videoplayer --sex
      #
      # ===================================================================== #
      when /^-?-?sexy?$/i,
           /^-?-?se$/i,
           /^-?-?x$/i
        @target_location << 'xx'
      # ===================================================================== #
      # === simpsons
      #
      # Invocation example:
      #
      #   videoplayer --simpsons
      #
      # ===================================================================== #
      when /^-?-?simpsons?$/i,
           /^-?-?cartoons?$/i,
           /^-?-?s$/i,
           /^-?-?simp$/i
        @target_location << 'Cartoons/Simpsons' # Purposely has no trailing /
      # ===================================================================== #
      # === --help
      #
      # Invocation example:
      #
      #   videoplayer --help
      #
      # ===================================================================== #
      when 'HELP','--help','help?','help','hel','h','h?','?'
        show_help_then_exit
      # ===================================================================== #
      # === THIS_DIR
      # ===================================================================== #
      when 'THIS_DIR','Dir.pwd','pwd','here','HERE'
        @target_location << rds('/'+Dir.pwd+'/')
      # ===================================================================== #
      # === listing
      # ===================================================================== #
      when 'listing','listing?','videos?','list',
           'videos','' # Feedback available videos.
        add_all_videos
        feedback_available_videos
        exit
      # ===================================================================== #
      # === science
      # ===================================================================== #
      when 'science'
        @target_location << 'Science'
      when 'loop','l','rvid','realvid','realvids' # Loop.
        @target_location << 'Realvids/'
      end
      add_all_videos
    end
  end; alias check_against_menu menu # === check_against_menu

  # ========================================================================= #
  # === add_all_videos
  # ========================================================================= #
  def add_all_videos
    add(obtain_all_video_files.sort)
  end

  # ========================================================================= #
  # === feedback_available_videos                               (listing tag)
  #
  # Use this method to feedback all available video files.
  # ========================================================================= #
  def feedback_available_videos
    videos?.uniq.sort_by {|line| File.basename(line)}.each {|entry|
      file = File.basename(entry)
      dir  = File.dirname(entry)
      e '  '+sdir(dir)+'/'+sfile(file)
    }
  end

  # ========================================================================= #
  # === exit_program
  #
  # Use only this method when you want to exit from this program.
  # ========================================================================= #
  def exit_program
    case runmode?
    when 'standalone'
      exit
    else # else pass through.
    end
  end

  # ========================================================================= #
  # === show_help_then_exit                                        (help tag)
  # ========================================================================= #
  def show_help_then_exit
    show_help
    exit
  end

  # ========================================================================= #
  # === show_help                                                  (help tag)
  #
  # This method will tell us which options are available for this class.
  #
  # It can be invoked via these keywords:
  #
  #   help? help hel h h? ?
  #
  # rvid --help
  # ========================================================================= #
  def show_help(
      prepend_this = 'Help requested. '
    )
    e prepend_this+N+N unless prepend_this.empty?
    e 'These options are available:'
    e
    e simp('  loop')
    e simp('  simpsons')
    e simp('  science')
    e simp('  married')
    e simp('  mma')
    e simp('  pwd')
    e
  end

  # ========================================================================= #
  # === random_video
  #
  # This will return a random video.
  #
  # It assumes that the @matches array exists; otherwise you will have
  # to provide an array with all videos to this method.
  # ========================================================================= #
  def random_video(i = @matches)
    i[rand(i.size)]
  end; alias return_random_video random_video # === return_random_video
       alias random_video?       random_video # === random_video?

  # ========================================================================= #
  # === run
  # ========================================================================= #
  def run
    check_against_menu
    play_video_files
  end

end

# =========================================================================== #
# === MultimediaParadise.play_video_file
# =========================================================================== #
def self.play_video_file(i)
  if i.is_a? Array
    i.each {|entry| MultimediaParadise.play_video_file(entry) }
  else
    MultimediaParadise::VideoPlayer.new(i)
  end
end

end # rvid --help
# __END__
# mplayer # vlc