#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::Video::VideoRenamer
#
# You will typically use this class on input that has no number as part
# of its filename. Some filenames may be convoluted, e. g. if you did
# download these files. The class here will try to rename a given file
# to match to the set of registered videofiles.
#
# Usage example:
#
#   MultimediaParadise::Video::VideoRenamer.new(ARGV)
#
# =========================================================================== #
# require 'multimedia_paradise/video/video_renamer.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'
  
module MultimediaParadise

module Video

class VideoRenamer < ::MultimediaParadise::Base # === MultimediaParadise::Video::VideoRenamer

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = ARGV,
      run_already = true
    )
    reset
    set_input(i)
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    infer_the_namespace
  end

  # ========================================================================= #
  # === set_input
  # ========================================================================= #
  def set_input(i = '')
    i = i.first if i.is_a? Array
    i = i.to_s.dup
    @input = i
  end

  # ========================================================================= #
  # === input?
  # ========================================================================= #
  def input?
    @input
  end

  # ========================================================================= #
  # === dataset?
  # ========================================================================= #
  def dataset?
    MultimediaParadise.return_all_video_files
  end

  # ========================================================================= #
  # === all_titles?
  # ========================================================================= #
  def all_titles?
    dataset?.map {|key, hash|
      hash['title']
    }
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    # ======================================================================= #
    # First, determine whether we have a match.
    # ======================================================================= #
    _ = first_word?
    # ======================================================================= #
    # If a '.' is there, we will assume a file extension such as '.mp4'
    # which is not necessary, thus it will be removed..
    # ======================================================================= #
    if _.include? '.'
      _.sub!(/#{File.extname(_)}$/,'')
    end
    video_is_registered = all_titles?.any? {|entry|
      entry.include? _
    }
    if video_is_registered
      rename_this_video_file
    else
      opnn; e 'It appears as if we do not have a video for this '\
              'input ('+simp(_)+').'
      # ===================================================================== #
      # === Check for numbers only
      # ===================================================================== #
      if _.to_s =~ /^\d+$/
        title = dataset?[_.to_i]['title']
        e 'Only numbers were given. We will thus check for the given file'
        e 'position at number '+simp(_)+' which is `'+simp(title)+'`.'
      end
    end
  end

  # ========================================================================= #
  # === rename_this_video_file
  #
  # Only call this method when you are sure that we have an input file.
  #
  # It will then proceed to rename the video file.
  # ========================================================================= #
  def rename_this_video_file(i = input?)
    matching_dataset = dataset?.select {|key, hash|
      downcased_word = first_word?.downcase
      hash['title'].downcase.include? downcased_word
    }
    if matching_dataset
      file_extension = File.extname(i).delete('.')
      the_key = matching_dataset.keys.first
      new_filename = the_key.to_s+'_'
      # ===================================================================== #
      # Next, append the real name.
      # ===================================================================== #
      new_filename << matching_dataset[the_key]['title'].delete(' ')+'.'+file_extension
      opnn; e 'Now renaming `'+sfile(input?)+'` to `'+sfile(new_filename)+'`.'
      rename_file(i, new_filename)
    end
  end

  # ========================================================================= #
  # === first_word?
  #
  # Obtain the first word of the input file.
  # ========================================================================= #
  def first_word?
    _ = input?
    _ = _.split('_').first if _.include? '_'
    return _
  end

  # ========================================================================= #
  # === MultimediaParadise::Video::VideoRenamer.new[]
  # ========================================================================= #
  def self.[](i = ARGV)
    new(i)
  end

end; end; end

if __FILE__ == $PROGRAM_NAME
  MultimediaParadise::Video::VideoRenamer.new(ARGV)
end # vrenamer