#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === MultimediaParadise::Watermark
#
# This class will watermark a video, using ffmpeg.
#
# It will create a new output file, with the name having the prefix
# "watermarked_video_file.avi".
#
# Usage examples:
#
#   MultimediaParadise::Watermark.new()
#   MultimediaParadise.watermark()
#
# Commandline example:
#
#   vwatermark /Depot/Temp/MultimediaProject/JosefHader_DerRobert.mp4
#   vmwark $JJJJ/output_264_PlanetOfTheApes.avi
#
# ============================================================================ #
# require 'multimedia_paradise/video/watermark.rb' 
# ============================================================================ #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

module Video

class Watermark < ::MultimediaParadise::Base # === MultimediaParadise::Watermark

  begin
    require 'cyberweb/requires/require_web_images.rb' # A non-critical require here. Not everyone needs to have WebImages.
  rescue LoadError; end

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === USE_THIS_IMAGE_AS_WATERMARK
  #
  # Which image to use a watermark. This will become the ivar
  # @image_to_use_as_watermark.
  # ========================================================================= #
  # USE_THIS_IMAGE_AS_WATERMARK = '/home/x/DATA/IMG/NJOY/RandomBeachGal.jpg'
  USE_THIS_IMAGE_AS_WATERMARK =
    '/home/x/data/images/STD/YIN_YANG.png'

  # ========================================================================= #
  # === OVERLAY_POSITION
  #
  # This may specify where the overlayed image is situated.
  # ========================================================================= #
  OVERLAY_POSITION = '8:12'

  # ========================================================================= #
  # === initialize
  #
  # The first input is the file that will be modified via a watermark.
  #
  # The second input is the image we use as watermark.
  # ========================================================================= #
  def initialize(
      commandline_arguments       = ARGV,
      run_already                 = true
    )
    reset
    set_commandline_arguments(
      commandline_arguments
    )
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @videofile
    # ======================================================================= #
    @videofile = nil
    # ======================================================================= #
    # Provide a default watermark-logo next.
    # ======================================================================= #
    set_image_to_use_as_watermark(
      USE_THIS_IMAGE_AS_WATERMARK
    )
    set_name_of_output_file
  end

  # ========================================================================= #
  # === videofile?
  # ========================================================================= #
  def videofile?
    _ = @videofile
    _ = _.first if _.is_a? Array
    _.to_s # Must return a String.
  end; alias input_file? videofile? # === input_file?
       alias video_file? videofile? # === video_file?
       alias v?          videofile? # === v?
       alias input?      videofile? # === input?

  # ========================================================================= #
  # === set_videofile
  # ========================================================================= #
  def set_videofile(i = '')
    i = i.join(' ').strip if i.is_a? Array
    i = i.to_s.dup
    case i
    # ======================================================================= #
    # === watermark --help
    # ======================================================================= #
    when /^-?-?help$/
      show_help; exit
    # ======================================================================= #
    # === image=:austria
    # ======================================================================= #
    when /--image=(\S+)/ # This will match to "--image=:austria 273_Rambo.avi", see http://rubular.com/r/SwkpuftyFT
      set_image($1.to_s.dup)
      i.gsub!(/(--image=\S+)/, '') # Get rid of this part here.
    end
    @videofile = [i]
    set_name_of_output_file
  end

  # ========================================================================= #
  # === show_help
  # ========================================================================= #
  def show_help
    opnn; e 'Currently these options are supported:'
    e
    e "  #{simp('--image=')} # use this image, as in \"--image=austria\" "\
      "to use the austrian flag"
    e
  end

  # ========================================================================= #
  # === sanitize_name_of_output_file
  # ========================================================================= #
  def sanitize_name_of_output_file
    @name_of_output_file << File.extname(@videofile.first)
  end

  # ========================================================================= #
  # === report
  # ========================================================================= #
  def report
    opnn; e "We stored at the location `#{sfile(output?)}`."
  end

  # ========================================================================= #
  # === output_file?                                             (output tag)
  # ========================================================================= #
  def output_file?
    @name_of_output_file
  end; alias output? output_file? # === output?

  # ========================================================================= #
  # === set_name_of_output_file
  #
  # This method-call must be after the input has been determined.
  # ========================================================================= #
  def set_name_of_output_file(
      i = 'watermarked_video_file_'+
          File.basename(
            input?.gsub(/#{File.extname(input_file?)}/,'')
          )
    )
    i = i.first if i.is_a? Array
    i = i.to_s.dup
    unless i.include? '/' # In this case, add the full path.
      i.prepend (Dir.pwd+'/').squeeze '/'
    end
    @name_of_output_file = i
  end

  # ========================================================================= #
  # === set_image_to_use_as_watermark
  #
  # Set the image you want to use for watermarking here. 
  # ========================================================================= #
  def set_image_to_use_as_watermark(i = :rand)
    i = i.to_s
    if i.include? '--'
      i = i.scan(/-?-?image=(.+)/).flatten.first.to_s
    end
    case i
    # ======================================================================= #
    # === rand
    # ======================================================================= #
    when /^rand/
      i = Dir[File.dirname(USE_THIS_IMAGE_AS_WATERMARK)+'/*'].sample
    else
      if Cyberweb.web_images_does_include?(i)
        i = Cyberweb.get_webimage(i)
      end
    end
    # ======================================================================= #
    # Next, assign to @image_to_use_as_watermark. We will use
    # this as path to the image file in question.
    # ======================================================================= #
    @image_to_use_as_watermark = i
  end; alias set_image set_image_to_use_as_watermark # === set_image

  # ========================================================================= #
  # === process_each_video_now
  # ========================================================================= #
  def process_each_video_now
    @videofile.each {|entry|
      _  = "ffmpeg -i #{entry}".dup
      _ << ' -strict -2'
      _ << ' -y' # We will always overwrite, hence this option.
      _ << ' -vf "movie='+@image_to_use_as_watermark+
           ' [watermark]; [in][watermark] overlay='+OVERLAY_POSITION+' [out]" '
      _ << output_file?
      e simp(_) # Better to have colours.
      system(_)
    }
  end

  # ========================================================================= #
  # === set_commandline_arguments
  # ========================================================================= #
  def set_commandline_arguments(
      i
    )
    @commandline_arguments = [i].flatten.compact
  end

  # ========================================================================= #
  # === determine_which_image_to_use_and_which_videofile_to_use
  # ========================================================================= #
  def determine_which_image_to_use_and_which_videofile_to_use
    first_argument  = @commandline_arguments.first
    second_argument = @commandline_arguments[1]
    set_videofile(first_argument)
    set_image(second_argument)
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    determine_which_image_to_use_and_which_videofile_to_use
    sanitize_name_of_output_file
    process_each_video_now
  end

end; end

  # ========================================================================= #
  # === MultimediaParadise.watermark
  #
  # The first argument shall be the video file that we wish to watermark.
  #
  # The second argument shall be the path to the image that we will use
  # for the watermark in question.
  #
  # To use this method, try this:
  #
  #   MultimediaParadise.watermark('/Depot/Video/053_Muttertag.mp4')
  # ========================================================================= #
  def self.watermark(
      videofile                   = nil,
      use_this_image_as_watermark =
        MultimediaParadise::Video::Watermark::USE_THIS_IMAGE_AS_WATERMARK
    )
    MultimediaParadise::Video::Watermark.new(
      videofile, use_this_image_as_watermark
    )
  end

end

if __FILE__ == $PROGRAM_NAME
  _ = MultimediaParadise::Video::Watermark.new(ARGV)
  _.report
end # vwatermark *.flv
    # watermark --image=:austria 273_Rambo.avi
    # watermark --image=:austria /Depot/j/foobar.mp4
    # watermark --image=/home/x/data/images/NJOY/SmilingLady.jpg /Depot/j/foobar.mp4
    # watermark FORUM.png foobar.mkv