#!/usr/bin/ruby -w
# Encoding: UTF-8
# =========================================================================== #
# === MultimediaParadise::YoutubeEmbedder
#
# This may allow you to embed videos for youtube-display.
#
# Usage example:
#
#   yedder https://www.youtube.com/watch?v=y6GZEvwkMi8
# =========================================================================== #
# require 'multimedia_paradise/video/youtube_embedder.rb'
# =========================================================================== #
require 'multimedia_paradise/base/base.rb'

module MultimediaParadise

class YoutubeEmbedder < ::MultimediaParadise::Base # === MultimediaParadise::YoutubeEmbedder

  require 'open-uri'

  # ========================================================================= #
  # Define four setters and getters - id, uri, provider, params.
  # ========================================================================= #
  attr_writer   :id # The getter is set in this class, somewhere below.
  attr_accessor :uri
  attr_accessor :provider
  attr_accessor :params

  # ========================================================================= #
  # === DEFAULT_URL
  # ========================================================================= #
  DEFAULT_URL = 'https://www.youtube.com/watch?v=8wqxmlUXPAs'

  # ========================================================================= #
  # === initialize
  #
  # The first argument should be the link to youtube.
  # ========================================================================= #
  def initialize(link, run_already = true)
    reset
    set_link(link)
    case run_already
    when :dont_run_yet
      run_already = false
    end
    run if run_already
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    super()
    @embed = ''
    set_width('560')
    set_height('315')
  end

  # ========================================================================= #
  # === embed
  # ========================================================================= #
  def embed(
      dimensions = @width+'x'+@height,
      options = {}
    )
    options[:load] = true if options[:load].nil?
    if dimensions.include? 'x'
      h, w = dimensions.split('x')
    else
      h = w = dimensions # Assume '50' as input actually means '50x50'.
    end
    if youtube?
      @embed = %(
        <iframe width="#{h}" height="#{w}" #{options[:load] ? 'src' : 'data-src'}="http://www.youtube.com/embed/#{id}?html5=1"
      )
      # Process the @options next.
      options.each {| key, value|
        next if key == :load
        @embed << %( data-#{key}="#{value}" )
      }
      @embed << %( frameborder="0" allowfullscreen></iframe>)
    else
      @embed << 'NOT FOUND ANYTHING.<br>'
    end
  end

  # ========================================================================= #
  # === embed?
  # ========================================================================= #
  def embed?
    @embed
  end; alias embedded_string? embed?

  # ========================================================================= #
  # === set_link
  # ========================================================================= #
  def set_link(i = nil)
    i = i.first if i.is_a? Array
    i = DEFAULT_URL if i.nil?
    i = i.to_s.dup
    @link = i
  end

  # ========================================================================= #
  # === determine_params
  #
  # We must be careful here - the input should be a string.
  # ========================================================================= #
  def determine_params(i = @uri)
    @params = Hash[*i.query.split('&').map {|entry|
      splitted = entry.split('=')
      splitted[0] = splitted.first.to_sym # First entry becomes a symbol
      splitted
    }.flatten]
  end

  # ========================================================================= #
  # === determine_provider
  #
  # We determine the provider here.
  # ========================================================================= #
  def determine_provider(i = @uri)
    @provider = case i.host
                when /youtube.com/ then :youtube
                end
  end

  # ========================================================================= #
  # === parse_uri
  #
  # This will set @uri.
  # ========================================================================= #
  def parse_uri(i = @link)
    @uri = URI(i)
  end

  # ========================================================================= #
  # === id?
  # ========================================================================= #
  def id?
    params[:v] if youtube?
  end; alias id id? # === id

  # ========================================================================= #
  # === youtube?
  # ========================================================================= #
  def youtube?
    provider == :youtube # This sets the @provider ivar.
  end

  # ========================================================================= #
  # === feedback
  # ========================================================================= #
  def feedback
    lpad = 32
    e 'Now testing some components for '+simp(url?)+':'
    e ('The id is: ').ljust(lpad)+sfancy(id)
    e ('The provider is: ').ljust(lpad)+sfancy(provider.to_s)
    e ('The thumbnail is: ').ljust(lpad)+sfancy(thumbnail)
    e
    e 'And to embed this into your webpage, use this code here:'+N+N
    e '  '+embed
    e N+N
  end

  # ========================================================================= #
  # === set_width
  # ========================================================================= #
  def set_width(i)
    if i.is_a? Hash and i.has_key?(:width)
      i = i.delete :width
    end
    @width = i.to_s
  end

  # ========================================================================= #
  # === set_height
  # ========================================================================= #
  def set_height(i)
    if i.is_a? Hash and i.has_key?(:height)
      i = i.delete :height
    end
    @height = i.to_s
  end

  # ========================================================================= #
  # === url?
  # ========================================================================= #
  def url?
    @uri.to_s
  end
  
  # ========================================================================= #
  # === thumbnail
  #
  # Give back a thumbnail of the youtube video.
  #
  # Each YouTube video has 4 generated images.
  #
  # They are predictably formatted as follows:
  # 
  #   http://img.youtube.com/vi/<insert-youtube-video-id-here>/0.jpg
  #   http://img.youtube.com/vi/<insert-youtube-video-id-here>/1.jpg
  #   http://img.youtube.com/vi/<insert-youtube-video-id-here>/2.jpg
  #   http://img.youtube.com/vi/<insert-youtube-video-id-here>/3.jpg
  # ========================================================================= #
  def thumbnail(size = :default)
    if youtube?
      "http://img.youtube.com/vi/#{id?}/0.jpg"
    end
  end

  # ========================================================================= #
  # === run
  # ========================================================================= #
  def run
    parse_uri
    determine_provider
    determine_params
    embed
    return @embed
  end

end; end

if __FILE__ == $PROGRAM_NAME
  _ = MultimediaParadise::YoutubeEmbedder.new(ARGV, :dont_run_yet)
  _.set_width  width: 550
  _.set_height height: 20
  _.run
  pp _
  puts; _.feedback
end # yedder
    # yedder https://www.youtube.com/watch?v=y6GZEvwkMi8